/*
 * Copyright (c) 2021 Bitdefender
 * SPDX-License-Identifier: Apache-2.0
 */
//! Mnemonics.

use super::decode_error::DecodeError;
use core::convert::TryFrom;

/// Uniquely identifies an instruction.
#[derive(Copy, Clone, Eq, PartialEq, Hash, Debug)]
#[allow(non_camel_case_types)]
pub enum Mnemonic {
    AAA,
    AAD,
    AADD,
    AAM,
    AAND,
    AAS,
    ADC,
    ADCX,
    ADD,
    ADDPD,
    ADDPS,
    ADDSD,
    ADDSS,
    ADDSUBPD,
    ADDSUBPS,
    ADOX,
    AESDEC,
    AESDEC128KL,
    AESDEC256KL,
    AESDECLAST,
    AESDECWIDE128KL,
    AESDECWIDE256KL,
    AESENC,
    AESENC128KL,
    AESENC256KL,
    AESENCLAST,
    AESENCWIDE128KL,
    AESENCWIDE256KL,
    AESIMC,
    AESKEYGENASSIST,
    AND,
    ANDN,
    ANDNPD,
    ANDNPS,
    ANDPD,
    ANDPS,
    AOR,
    ARPL,
    AXOR,
    BEXTR,
    BLCFILL,
    BLCI,
    BLCIC,
    BLCMSK,
    BLCS,
    BLENDPD,
    BLENDPS,
    BLENDVPD,
    BLENDVPS,
    BLSFILL,
    BLSI,
    BLSIC,
    BLSMSK,
    BLSR,
    BNDCL,
    BNDCN,
    BNDCU,
    BNDLDX,
    BNDMK,
    BNDMOV,
    BNDSTX,
    BOUND,
    BSF,
    BSR,
    BSWAP,
    BT,
    BTC,
    BTR,
    BTS,
    BZHI,
    CALLFD,
    CALLFI,
    CALLNI,
    CALLNR,
    CBW,
    CCMP,
    CDQ,
    CDQE,
    CFCMOV,
    CLAC,
    CLC,
    CLD,
    CLDEMOTE,
    CLEVICT0,
    CLEVICT1,
    CLFLUSH,
    CLFLUSHOPT,
    CLGI,
    CLI,
    CLRSSBSY,
    CLTS,
    CLUI,
    CLWB,
    CLZERO,
    CMC,
    CMOVcc,
    CMP,
    CMPBEXADD,
    CMPCXADD,
    CMPLEXADD,
    CMPLXADD,
    CMPNBEXADD,
    CMPNCXADD,
    CMPNLEXADD,
    CMPNLXADD,
    CMPNOXADD,
    CMPNPXADD,
    CMPNSXADD,
    CMPNZXADD,
    CMPOXADD,
    CMPPD,
    CMPPS,
    CMPPXADD,
    CMPS,
    CMPSD,
    CMPSS,
    CMPSXADD,
    CMPXCHG,
    CMPXCHG16B,
    CMPXCHG8B,
    CMPZXADD,
    COMISD,
    COMISS,
    CPUID,
    CQO,
    CRC32,
    CTEST,
    CVTDQ2PD,
    CVTDQ2PS,
    CVTPD2DQ,
    CVTPD2PI,
    CVTPD2PS,
    CVTPI2PD,
    CVTPI2PS,
    CVTPS2DQ,
    CVTPS2PD,
    CVTPS2PI,
    CVTSD2SI,
    CVTSD2SS,
    CVTSI2SD,
    CVTSI2SS,
    CVTSS2SD,
    CVTSS2SI,
    CVTTPD2DQ,
    CVTTPD2PI,
    CVTTPS2DQ,
    CVTTPS2PI,
    CVTTSD2SI,
    CVTTSS2SI,
    CWD,
    CWDE,
    DAA,
    DAS,
    DEC,
    DELAY,
    DIV,
    DIVPD,
    DIVPS,
    DIVSD,
    DIVSS,
    DPPD,
    DPPS,
    EMMS,
    ENCLS,
    ENCLU,
    ENCLV,
    ENCODEKEY128,
    ENCODEKEY256,
    ENDBR,
    ENQCMD,
    ENQCMDS,
    ENTER,
    ERETS,
    ERETU,
    EXTRACTPS,
    EXTRQ,
    F2XM1,
    FABS,
    FADD,
    FADDP,
    FBLD,
    FBSTP,
    FCHS,
    FCMOVB,
    FCMOVBE,
    FCMOVE,
    FCMOVNB,
    FCMOVNBE,
    FCMOVNE,
    FCMOVNU,
    FCMOVU,
    FCOM,
    FCOMI,
    FCOMIP,
    FCOMP,
    FCOMPP,
    FCOS,
    FDECSTP,
    FDIV,
    FDIVP,
    FDIVR,
    FDIVRP,
    FEMMS,
    FFREE,
    FFREEP,
    FIADD,
    FICOM,
    FICOMP,
    FIDIV,
    FIDIVR,
    FILD,
    FIMUL,
    FINCSTP,
    FIST,
    FISTP,
    FISTTP,
    FISUB,
    FISUBR,
    FLD,
    FLD1,
    FLDCW,
    FLDENV,
    FLDL2E,
    FLDL2T,
    FLDLG2,
    FLDLN2,
    FLDPI,
    FLDZ,
    FMUL,
    FMULP,
    FNCLEX,
    FNDISI,
    FNINIT,
    FNOP,
    FNSAVE,
    FNSTCW,
    FNSTENV,
    FNSTSW,
    FPATAN,
    FPREM,
    FPREM1,
    FPTAN,
    FRINEAR,
    FRNDINT,
    FRSTOR,
    FSCALE,
    FSIN,
    FSINCOS,
    FSQRT,
    FST,
    FSTDW,
    FSTP,
    FSTPNCE,
    FSTSG,
    FSUB,
    FSUBP,
    FSUBR,
    FSUBRP,
    FTST,
    FUCOM,
    FUCOMI,
    FUCOMIP,
    FUCOMP,
    FUCOMPP,
    FXAM,
    FXCH,
    FXRSTOR,
    FXRSTOR64,
    FXSAVE,
    FXSAVE64,
    FXTRACT,
    FYL2X,
    FYL2XP1,
    GETSEC,
    GF2P8AFFINEINVQB,
    GF2P8AFFINEQB,
    GF2P8MULB,
    HADDPD,
    HADDPS,
    HLT,
    HRESET,
    HSUBPD,
    HSUBPS,
    IDIV,
    IMUL,
    IN,
    INC,
    INCSSP,
    INS,
    INSERTPS,
    INSERTQ,
    INT,
    INT1,
    INT3,
    INTO,
    INVD,
    INVEPT,
    INVLPG,
    INVLPGA,
    INVLPGB,
    INVPCID,
    INVVPID,
    IRET,
    JMPABS,
    JMPE,
    JMPFD,
    JMPFI,
    JMPNI,
    JMPNR,
    Jcc,
    JrCXZ,
    KADD,
    KAND,
    KANDN,
    KMERGE2L1H,
    KMERGE2L1L,
    KMOV,
    KNOT,
    KOR,
    KORTEST,
    KSHIFTL,
    KSHIFTR,
    KTEST,
    KUNPCKBW,
    KUNPCKDQ,
    KUNPCKWD,
    KXNOR,
    KXOR,
    LAHF,
    LAR,
    LDDQU,
    LDMXCSR,
    LDS,
    LDTILECFG,
    LEA,
    LEAVE,
    LES,
    LFENCE,
    LFS,
    LGDT,
    LGS,
    LIDT,
    LKGS,
    LLDT,
    LLWPCB,
    LMSW,
    LOADIWKEY,
    LODS,
    LOOP,
    LOOPNZ,
    LOOPZ,
    LSL,
    LSS,
    LTR,
    LWPINS,
    LWPVAL,
    LZCNT,
    MASKMOVDQU,
    MASKMOVQ,
    MAXPD,
    MAXPS,
    MAXSD,
    MAXSS,
    MCOMMIT,
    MFENCE,
    MINPD,
    MINPS,
    MINSD,
    MINSS,
    MONITOR,
    MONITORX,
    MOV,
    MOVAPD,
    MOVAPS,
    MOVBE,
    MOVD,
    MOVDDUP,
    MOVDIR64B,
    MOVDIRI,
    MOVDQ2Q,
    MOVDQA,
    MOVDQU,
    MOVHLPS,
    MOVHPD,
    MOVHPS,
    MOVLHPS,
    MOVLPD,
    MOVLPS,
    MOVMSKPD,
    MOVMSKPS,
    MOVNTDQ,
    MOVNTDQA,
    MOVNTI,
    MOVNTPD,
    MOVNTPS,
    MOVNTQ,
    MOVNTSD,
    MOVNTSS,
    MOVQ,
    MOVQ2DQ,
    MOVS,
    MOVSD,
    MOVSHDUP,
    MOVSLDUP,
    MOVSS,
    MOVSX,
    MOVSXD,
    MOVUPD,
    MOVUPS,
    MOVZX,
    MOV_CR,
    MOV_DR,
    MOV_TR,
    MPSADBW,
    MUL,
    MULPD,
    MULPS,
    MULSD,
    MULSS,
    MULX,
    MWAIT,
    MWAITX,
    NEG,
    NOP,
    NOT,
    OR,
    ORPD,
    ORPS,
    OUT,
    OUTS,
    PABSB,
    PABSD,
    PABSW,
    PACKSSDW,
    PACKSSWB,
    PACKUSDW,
    PACKUSWB,
    PADDB,
    PADDD,
    PADDQ,
    PADDSB,
    PADDSW,
    PADDUSB,
    PADDUSW,
    PADDW,
    PALIGNR,
    PAND,
    PANDN,
    PAUSE,
    PAVGB,
    PAVGUSB,
    PAVGW,
    PBLENDVB,
    PBLENDW,
    PBNDKB,
    PCLMULQDQ,
    PCMPEQB,
    PCMPEQD,
    PCMPEQQ,
    PCMPEQW,
    PCMPESTRI,
    PCMPESTRM,
    PCMPGTB,
    PCMPGTD,
    PCMPGTQ,
    PCMPGTW,
    PCMPISTRI,
    PCMPISTRM,
    PCONFIG,
    PDEP,
    PEXT,
    PEXTRB,
    PEXTRD,
    PEXTRQ,
    PEXTRW,
    PF2ID,
    PF2IW,
    PFACC,
    PFADD,
    PFCMPEQ,
    PFCMPGE,
    PFCMPGT,
    PFMAX,
    PFMIN,
    PFMUL,
    PFNACC,
    PFPNACC,
    PFRCP,
    PFRCPIT1,
    PFRCPIT2,
    PFRCPV,
    PFRSQIT1,
    PFRSQRT,
    PFRSQRTV,
    PFSUB,
    PFSUBR,
    PHADDD,
    PHADDSW,
    PHADDW,
    PHMINPOSUW,
    PHSUBD,
    PHSUBSW,
    PHSUBW,
    PI2FD,
    PI2FW,
    PINSRB,
    PINSRD,
    PINSRQ,
    PINSRW,
    PMADDUBSW,
    PMADDWD,
    PMAXSB,
    PMAXSD,
    PMAXSW,
    PMAXUB,
    PMAXUD,
    PMAXUW,
    PMINSB,
    PMINSD,
    PMINSW,
    PMINUB,
    PMINUD,
    PMINUW,
    PMOVMSKB,
    PMOVSXBD,
    PMOVSXBQ,
    PMOVSXBW,
    PMOVSXDQ,
    PMOVSXWD,
    PMOVSXWQ,
    PMOVZXBD,
    PMOVZXBQ,
    PMOVZXBW,
    PMOVZXDQ,
    PMOVZXWD,
    PMOVZXWQ,
    PMULDQ,
    PMULHRSW,
    PMULHRW,
    PMULHUW,
    PMULHW,
    PMULLD,
    PMULLW,
    PMULUDQ,
    POP,
    POP2,
    POP2P,
    POPA,
    POPAD,
    POPCNT,
    POPF,
    POPP,
    POR,
    PREFETCH,
    PREFETCHE,
    PREFETCHIT0,
    PREFETCHIT1,
    PREFETCHM,
    PREFETCHNTA,
    PREFETCHT0,
    PREFETCHT1,
    PREFETCHT2,
    PREFETCHW,
    PREFETCHWT1,
    PSADBW,
    PSHUFB,
    PSHUFD,
    PSHUFHW,
    PSHUFLW,
    PSHUFW,
    PSIGNB,
    PSIGND,
    PSIGNW,
    PSLLD,
    PSLLDQ,
    PSLLQ,
    PSLLW,
    PSMASH,
    PSRAD,
    PSRAW,
    PSRLD,
    PSRLDQ,
    PSRLQ,
    PSRLW,
    PSUBB,
    PSUBD,
    PSUBQ,
    PSUBSB,
    PSUBSW,
    PSUBUSB,
    PSUBUSW,
    PSUBW,
    PSWAPD,
    PTEST,
    PTWRITE,
    PUNPCKHBW,
    PUNPCKHDQ,
    PUNPCKHQDQ,
    PUNPCKHWD,
    PUNPCKLBW,
    PUNPCKLDQ,
    PUNPCKLQDQ,
    PUNPCKLWD,
    PUSH,
    PUSH2,
    PUSH2P,
    PUSHA,
    PUSHAD,
    PUSHF,
    PUSHP,
    PVALIDATE,
    PXOR,
    RCL,
    RCPPS,
    RCPSS,
    RCR,
    RDFSBASE,
    RDGSBASE,
    RDMSR,
    RDMSRLIST,
    RDPID,
    RDPKRU,
    RDPMC,
    RDPRU,
    RDRAND,
    RDSEED,
    RDTSC,
    RDTSCP,
    RETF,
    RETN,
    RMPADJUST,
    RMPQUERY,
    RMPREAD,
    RMPUPDATE,
    ROL,
    ROR,
    RORX,
    ROUNDPD,
    ROUNDPS,
    ROUNDSD,
    ROUNDSS,
    RSM,
    RSQRTPS,
    RSQRTSS,
    RSSSP,
    RSTORSSP,
    SAHF,
    SAL,
    SALC,
    SAR,
    SARX,
    SAVEPREVSSP,
    SBB,
    SCAS,
    SEAMCALL,
    SEAMOPS,
    SEAMRET,
    SENDUIPI,
    SERIALIZE,
    SETSSBSY,
    SETcc,
    SFENCE,
    SGDT,
    SHA1MSG1,
    SHA1MSG2,
    SHA1NEXTE,
    SHA1RNDS4,
    SHA256MSG1,
    SHA256MSG2,
    SHA256RNDS2,
    SHL,
    SHLD,
    SHLX,
    SHR,
    SHRD,
    SHRX,
    SHUFPD,
    SHUFPS,
    SIDT,
    SKINIT,
    SLDT,
    SLWPCB,
    SMSW,
    SPFLT,
    SQRTPD,
    SQRTPS,
    SQRTSD,
    SQRTSS,
    STAC,
    STC,
    STD,
    STGI,
    STI,
    STMXCSR,
    STOS,
    STR,
    STTILECFG,
    STUI,
    SUB,
    SUBPD,
    SUBPS,
    SUBSD,
    SUBSS,
    SWAPGS,
    SYSCALL,
    SYSENTER,
    SYSEXIT,
    SYSRET,
    T1MSKC,
    TCMMIMFP16PS,
    TCMMRLFP16PS,
    TDCALL,
    TDPBF16PS,
    TDPBSSD,
    TDPBSUD,
    TDPBUSD,
    TDPBUUD,
    TDPFP16PS,
    TEST,
    TESTUI,
    TILELOADD,
    TILELOADDT1,
    TILERELEASE,
    TILESTORED,
    TILEZERO,
    TLBSYNC,
    TPAUSE,
    TZCNT,
    TZMSK,
    UCOMISD,
    UCOMISS,
    UD0,
    UD1,
    UD2,
    UIRET,
    UMONITOR,
    UMWAIT,
    UNPCKHPD,
    UNPCKHPS,
    UNPCKLPD,
    UNPCKLPS,
    URDMSR,
    UWRMSR,
    V4FMADDPS,
    V4FMADDSS,
    V4FNMADDPS,
    V4FNMADDSS,
    VADDNEPBF16,
    VADDPD,
    VADDPH,
    VADDPS,
    VADDSD,
    VADDSH,
    VADDSS,
    VADDSUBPD,
    VADDSUBPS,
    VAESDEC,
    VAESDECLAST,
    VAESENC,
    VAESENCLAST,
    VAESIMC,
    VAESKEYGENASSIST,
    VALIGND,
    VALIGNQ,
    VANDNPD,
    VANDNPS,
    VANDPD,
    VANDPS,
    VBCSTNEBF162PS,
    VBCSTNESH2PS,
    VBLENDMPD,
    VBLENDMPS,
    VBLENDPD,
    VBLENDPS,
    VBLENDVPD,
    VBLENDVPS,
    VBROADCASTF128,
    VBROADCASTF32X2,
    VBROADCASTF32X4,
    VBROADCASTF32X8,
    VBROADCASTF64X2,
    VBROADCASTF64X4,
    VBROADCASTI128,
    VBROADCASTI32X2,
    VBROADCASTI32X4,
    VBROADCASTI32X8,
    VBROADCASTI64X2,
    VBROADCASTI64X4,
    VBROADCASTSD,
    VBROADCASTSS,
    VCMPPBF16,
    VCMPPD,
    VCMPPH,
    VCMPPS,
    VCMPSD,
    VCMPSH,
    VCMPSS,
    VCOMISD,
    VCOMISH,
    VCOMISS,
    VCOMPRESSPD,
    VCOMPRESSPS,
    VCOMSBF16,
    VCOMXSD,
    VCOMXSH,
    VCOMXSS,
    VCVT2PS2PHX,
    VCVTBIASPH2BF8,
    VCVTBIASPH2BF8S,
    VCVTBIASPH2HF8,
    VCVTBIASPH2HF8S,
    VCVTDQ2PD,
    VCVTDQ2PH,
    VCVTDQ2PS,
    VCVTHF82PH,
    VCVTNE2PH2BF8,
    VCVTNE2PH2BF8S,
    VCVTNE2PH2HF8,
    VCVTNE2PH2HF8S,
    VCVTNE2PS2BF16,
    VCVTNEBF162IBS,
    VCVTNEBF162IUBS,
    VCVTNEEBF162PS,
    VCVTNEEPH2PS,
    VCVTNEOBF162PS,
    VCVTNEOPH2PS,
    VCVTNEPH2BF8,
    VCVTNEPH2BF8S,
    VCVTNEPH2HF8,
    VCVTNEPH2HF8S,
    VCVTNEPS2BF16,
    VCVTPD2DQ,
    VCVTPD2PH,
    VCVTPD2PS,
    VCVTPD2QQ,
    VCVTPD2UDQ,
    VCVTPD2UQQ,
    VCVTPH2DQ,
    VCVTPH2IBS,
    VCVTPH2IUBS,
    VCVTPH2PD,
    VCVTPH2PS,
    VCVTPH2PSX,
    VCVTPH2QQ,
    VCVTPH2UDQ,
    VCVTPH2UQQ,
    VCVTPH2UW,
    VCVTPH2W,
    VCVTPS2DQ,
    VCVTPS2IBS,
    VCVTPS2IUBS,
    VCVTPS2PD,
    VCVTPS2PH,
    VCVTPS2PHX,
    VCVTPS2QQ,
    VCVTPS2UDQ,
    VCVTPS2UQQ,
    VCVTQQ2PD,
    VCVTQQ2PH,
    VCVTQQ2PS,
    VCVTSD2SH,
    VCVTSD2SI,
    VCVTSD2SS,
    VCVTSD2USI,
    VCVTSH2SD,
    VCVTSH2SI,
    VCVTSH2SS,
    VCVTSH2USI,
    VCVTSI2SD,
    VCVTSI2SH,
    VCVTSI2SS,
    VCVTSS2SD,
    VCVTSS2SH,
    VCVTSS2SI,
    VCVTSS2USI,
    VCVTTNEBF162IBS,
    VCVTTNEBF162IUBS,
    VCVTTPD2DQ,
    VCVTTPD2DQS,
    VCVTTPD2QQ,
    VCVTTPD2QQS,
    VCVTTPD2UDQ,
    VCVTTPD2UDQS,
    VCVTTPD2UQQ,
    VCVTTPD2UQQS,
    VCVTTPH2DQ,
    VCVTTPH2IBS,
    VCVTTPH2IUBS,
    VCVTTPH2QQ,
    VCVTTPH2UDQ,
    VCVTTPH2UQQ,
    VCVTTPH2UW,
    VCVTTPH2W,
    VCVTTPS2DQ,
    VCVTTPS2DQS,
    VCVTTPS2IBS,
    VCVTTPS2IUBS,
    VCVTTPS2QQ,
    VCVTTPS2QQS,
    VCVTTPS2UDQ,
    VCVTTPS2UDQS,
    VCVTTPS2UQQ,
    VCVTTPS2UQQS,
    VCVTTSD2SI,
    VCVTTSD2SIS,
    VCVTTSD2USI,
    VCVTTSD2USIS,
    VCVTTSH2SI,
    VCVTTSH2USI,
    VCVTTSS2SI,
    VCVTTSS2SIS,
    VCVTTSS2USI,
    VCVTTSS2USIS,
    VCVTUDQ2PD,
    VCVTUDQ2PH,
    VCVTUDQ2PS,
    VCVTUQQ2PD,
    VCVTUQQ2PH,
    VCVTUQQ2PS,
    VCVTUSI2SD,
    VCVTUSI2SH,
    VCVTUSI2SS,
    VCVTUW2PH,
    VCVTW2PH,
    VDBPSADBW,
    VDIVNEPBF16,
    VDIVPD,
    VDIVPH,
    VDIVPS,
    VDIVSD,
    VDIVSH,
    VDIVSS,
    VDPBF16PS,
    VDPPD,
    VDPPHPS,
    VDPPS,
    VERR,
    VERW,
    VEXP2PD,
    VEXP2PS,
    VEXPANDPD,
    VEXPANDPS,
    VEXTRACTF128,
    VEXTRACTF32X4,
    VEXTRACTF32X8,
    VEXTRACTF64X2,
    VEXTRACTF64X4,
    VEXTRACTI128,
    VEXTRACTI32X4,
    VEXTRACTI32X8,
    VEXTRACTI64X2,
    VEXTRACTI64X4,
    VEXTRACTPS,
    VFCMADDCPH,
    VFCMADDCSH,
    VFCMULCPH,
    VFCMULCSH,
    VFIXUPIMMPD,
    VFIXUPIMMPS,
    VFIXUPIMMSD,
    VFIXUPIMMSS,
    VFMADD132NEPBF16,
    VFMADD132PD,
    VFMADD132PH,
    VFMADD132PS,
    VFMADD132SD,
    VFMADD132SH,
    VFMADD132SS,
    VFMADD213NEPBF16,
    VFMADD213PD,
    VFMADD213PH,
    VFMADD213PS,
    VFMADD213SD,
    VFMADD213SH,
    VFMADD213SS,
    VFMADD231NEPBF16,
    VFMADD231PD,
    VFMADD231PH,
    VFMADD231PS,
    VFMADD231SD,
    VFMADD231SH,
    VFMADD231SS,
    VFMADDCPH,
    VFMADDCSH,
    VFMADDPD,
    VFMADDPS,
    VFMADDSD,
    VFMADDSS,
    VFMADDSUB132PD,
    VFMADDSUB132PH,
    VFMADDSUB132PS,
    VFMADDSUB213PD,
    VFMADDSUB213PH,
    VFMADDSUB213PS,
    VFMADDSUB231PD,
    VFMADDSUB231PH,
    VFMADDSUB231PS,
    VFMADDSUBPD,
    VFMADDSUBPS,
    VFMSUB132NEPBF16,
    VFMSUB132PD,
    VFMSUB132PH,
    VFMSUB132PS,
    VFMSUB132SD,
    VFMSUB132SH,
    VFMSUB132SS,
    VFMSUB213NEPBF16,
    VFMSUB213PD,
    VFMSUB213PH,
    VFMSUB213PS,
    VFMSUB213SD,
    VFMSUB213SH,
    VFMSUB213SS,
    VFMSUB231NEPBF16,
    VFMSUB231PD,
    VFMSUB231PH,
    VFMSUB231PS,
    VFMSUB231SD,
    VFMSUB231SH,
    VFMSUB231SS,
    VFMSUBADD132PD,
    VFMSUBADD132PH,
    VFMSUBADD132PS,
    VFMSUBADD213PD,
    VFMSUBADD213PH,
    VFMSUBADD213PS,
    VFMSUBADD231PD,
    VFMSUBADD231PH,
    VFMSUBADD231PS,
    VFMSUBADDPD,
    VFMSUBADDPS,
    VFMSUBPD,
    VFMSUBPS,
    VFMSUBSD,
    VFMSUBSS,
    VFMULCPH,
    VFMULCSH,
    VFNMADD132NEPBF16,
    VFNMADD132PD,
    VFNMADD132PH,
    VFNMADD132PS,
    VFNMADD132SD,
    VFNMADD132SH,
    VFNMADD132SS,
    VFNMADD213NEPBF16,
    VFNMADD213PD,
    VFNMADD213PH,
    VFNMADD213PS,
    VFNMADD213SD,
    VFNMADD213SH,
    VFNMADD213SS,
    VFNMADD231NEPBF16,
    VFNMADD231PD,
    VFNMADD231PH,
    VFNMADD231PS,
    VFNMADD231SD,
    VFNMADD231SH,
    VFNMADD231SS,
    VFNMADDPD,
    VFNMADDPS,
    VFNMADDSD,
    VFNMADDSS,
    VFNMSUB132NEPBF16,
    VFNMSUB132PD,
    VFNMSUB132PH,
    VFNMSUB132PS,
    VFNMSUB132SD,
    VFNMSUB132SH,
    VFNMSUB132SS,
    VFNMSUB213NEPBF16,
    VFNMSUB213PD,
    VFNMSUB213PH,
    VFNMSUB213PS,
    VFNMSUB213SD,
    VFNMSUB213SH,
    VFNMSUB213SS,
    VFNMSUB231NEPBF16,
    VFNMSUB231PD,
    VFNMSUB231PH,
    VFNMSUB231PS,
    VFNMSUB231SD,
    VFNMSUB231SH,
    VFNMSUB231SS,
    VFNMSUBPD,
    VFNMSUBPS,
    VFNMSUBSD,
    VFNMSUBSS,
    VFPCLASSPBF16,
    VFPCLASSPD,
    VFPCLASSPH,
    VFPCLASSPS,
    VFPCLASSSD,
    VFPCLASSSH,
    VFPCLASSSS,
    VFRCZPD,
    VFRCZPS,
    VFRCZSD,
    VFRCZSS,
    VGATHERDPD,
    VGATHERDPS,
    VGATHERPF0DPD,
    VGATHERPF0DPS,
    VGATHERPF0QPD,
    VGATHERPF0QPS,
    VGATHERPF1DPD,
    VGATHERPF1DPS,
    VGATHERPF1QPD,
    VGATHERPF1QPS,
    VGATHERQPD,
    VGATHERQPS,
    VGETEXPPBF16,
    VGETEXPPD,
    VGETEXPPH,
    VGETEXPPS,
    VGETEXPSD,
    VGETEXPSH,
    VGETEXPSS,
    VGETMANTPBF16,
    VGETMANTPD,
    VGETMANTPH,
    VGETMANTPS,
    VGETMANTSD,
    VGETMANTSH,
    VGETMANTSS,
    VGF2P8AFFINEINVQB,
    VGF2P8AFFINEQB,
    VGF2P8MULB,
    VHADDPD,
    VHADDPS,
    VHSUBPD,
    VHSUBPS,
    VINSERTF128,
    VINSERTF32X4,
    VINSERTF32X8,
    VINSERTF64X2,
    VINSERTF64X4,
    VINSERTI128,
    VINSERTI32X4,
    VINSERTI32X8,
    VINSERTI64X2,
    VINSERTI64X4,
    VINSERTPS,
    VLDDQU,
    VLDMXCSR,
    VMASKMOVDQU,
    VMASKMOVPD,
    VMASKMOVPS,
    VMAXPBF16,
    VMAXPD,
    VMAXPH,
    VMAXPS,
    VMAXSD,
    VMAXSH,
    VMAXSS,
    VMCALL,
    VMCLEAR,
    VMFUNC,
    VMGEXIT,
    VMINMAXNEPBF16,
    VMINMAXPD,
    VMINMAXPH,
    VMINMAXPS,
    VMINMAXSD,
    VMINMAXSH,
    VMINMAXSS,
    VMINPBF16,
    VMINPD,
    VMINPH,
    VMINPS,
    VMINSD,
    VMINSH,
    VMINSS,
    VMLAUNCH,
    VMLOAD,
    VMMCALL,
    VMOVAPD,
    VMOVAPS,
    VMOVD,
    VMOVDDUP,
    VMOVDQA,
    VMOVDQA32,
    VMOVDQA64,
    VMOVDQU,
    VMOVDQU16,
    VMOVDQU32,
    VMOVDQU64,
    VMOVDQU8,
    VMOVHLPS,
    VMOVHPD,
    VMOVHPS,
    VMOVLHPS,
    VMOVLPD,
    VMOVLPS,
    VMOVMSKPD,
    VMOVMSKPS,
    VMOVNTDQ,
    VMOVNTDQA,
    VMOVNTPD,
    VMOVNTPS,
    VMOVQ,
    VMOVSD,
    VMOVSH,
    VMOVSHDUP,
    VMOVSLDUP,
    VMOVSS,
    VMOVUPD,
    VMOVUPS,
    VMOVW,
    VMPSADBW,
    VMPTRLD,
    VMPTRST,
    VMREAD,
    VMRESUME,
    VMRUN,
    VMSAVE,
    VMULNEPBF16,
    VMULPD,
    VMULPH,
    VMULPS,
    VMULSD,
    VMULSH,
    VMULSS,
    VMWRITE,
    VMXOFF,
    VMXON,
    VORPD,
    VORPS,
    VP2INTERSECTD,
    VP2INTERSECTQ,
    VP4DPWSSD,
    VP4DPWSSDS,
    VPABSB,
    VPABSD,
    VPABSQ,
    VPABSW,
    VPACKSSDW,
    VPACKSSWB,
    VPACKUSDW,
    VPACKUSWB,
    VPADDB,
    VPADDD,
    VPADDQ,
    VPADDSB,
    VPADDSW,
    VPADDUSB,
    VPADDUSW,
    VPADDW,
    VPALIGNR,
    VPAND,
    VPANDD,
    VPANDN,
    VPANDND,
    VPANDNQ,
    VPANDQ,
    VPAVGB,
    VPAVGW,
    VPBLENDD,
    VPBLENDMB,
    VPBLENDMD,
    VPBLENDMQ,
    VPBLENDMW,
    VPBLENDVB,
    VPBLENDW,
    VPBROADCASTB,
    VPBROADCASTD,
    VPBROADCASTMB2Q,
    VPBROADCASTMW2D,
    VPBROADCASTQ,
    VPBROADCASTW,
    VPCLMULQDQ,
    VPCMOV,
    VPCMPB,
    VPCMPD,
    VPCMPEQB,
    VPCMPEQD,
    VPCMPEQQ,
    VPCMPEQW,
    VPCMPESTRI,
    VPCMPESTRM,
    VPCMPGTB,
    VPCMPGTD,
    VPCMPGTQ,
    VPCMPGTW,
    VPCMPISTRI,
    VPCMPISTRM,
    VPCMPQ,
    VPCMPUB,
    VPCMPUD,
    VPCMPUQ,
    VPCMPUW,
    VPCMPW,
    VPCOMB,
    VPCOMD,
    VPCOMPRESSB,
    VPCOMPRESSD,
    VPCOMPRESSQ,
    VPCOMPRESSW,
    VPCOMQ,
    VPCOMUB,
    VPCOMUD,
    VPCOMUQ,
    VPCOMUW,
    VPCOMW,
    VPCONFLICTD,
    VPCONFLICTQ,
    VPDPBSSD,
    VPDPBSSDS,
    VPDPBSUD,
    VPDPBSUDS,
    VPDPBUSD,
    VPDPBUSDS,
    VPDPBUUD,
    VPDPBUUDS,
    VPDPWSSD,
    VPDPWSSDS,
    VPDPWSUD,
    VPDPWSUDS,
    VPDPWUSD,
    VPDPWUSDS,
    VPDPWUUD,
    VPDPWUUDS,
    VPERM2F128,
    VPERM2I128,
    VPERMB,
    VPERMD,
    VPERMI2B,
    VPERMI2D,
    VPERMI2PD,
    VPERMI2PS,
    VPERMI2Q,
    VPERMI2W,
    VPERMIL2PD,
    VPERMIL2PS,
    VPERMILPD,
    VPERMILPS,
    VPERMPD,
    VPERMPS,
    VPERMQ,
    VPERMT2B,
    VPERMT2D,
    VPERMT2PD,
    VPERMT2PS,
    VPERMT2Q,
    VPERMT2W,
    VPERMW,
    VPEXPANDB,
    VPEXPANDD,
    VPEXPANDQ,
    VPEXPANDW,
    VPEXTRB,
    VPEXTRD,
    VPEXTRQ,
    VPEXTRW,
    VPGATHERDD,
    VPGATHERDQ,
    VPGATHERQD,
    VPGATHERQQ,
    VPHADDBD,
    VPHADDBQ,
    VPHADDBW,
    VPHADDD,
    VPHADDDQ,
    VPHADDSW,
    VPHADDUBD,
    VPHADDUBQ,
    VPHADDUBW,
    VPHADDUDQ,
    VPHADDUWD,
    VPHADDUWQ,
    VPHADDW,
    VPHADDWD,
    VPHADDWQ,
    VPHMINPOSUW,
    VPHSUBBW,
    VPHSUBD,
    VPHSUBDQ,
    VPHSUBSW,
    VPHSUBW,
    VPHSUBWD,
    VPINSRB,
    VPINSRD,
    VPINSRQ,
    VPINSRW,
    VPLZCNTD,
    VPLZCNTQ,
    VPMACSDD,
    VPMACSDQH,
    VPMACSDQL,
    VPMACSSDD,
    VPMACSSDQH,
    VPMACSSDQL,
    VPMACSSWD,
    VPMACSSWW,
    VPMACSWD,
    VPMACSWW,
    VPMADCSSWD,
    VPMADCSWD,
    VPMADD52HUQ,
    VPMADD52LUQ,
    VPMADDUBSW,
    VPMADDWD,
    VPMASKMOVD,
    VPMASKMOVQ,
    VPMAXSB,
    VPMAXSD,
    VPMAXSQ,
    VPMAXSW,
    VPMAXUB,
    VPMAXUD,
    VPMAXUQ,
    VPMAXUW,
    VPMINSB,
    VPMINSD,
    VPMINSQ,
    VPMINSW,
    VPMINUB,
    VPMINUD,
    VPMINUQ,
    VPMINUW,
    VPMOVB2M,
    VPMOVD2M,
    VPMOVDB,
    VPMOVDW,
    VPMOVM2B,
    VPMOVM2D,
    VPMOVM2Q,
    VPMOVM2W,
    VPMOVMSKB,
    VPMOVQ2M,
    VPMOVQB,
    VPMOVQD,
    VPMOVQW,
    VPMOVSDB,
    VPMOVSDW,
    VPMOVSQB,
    VPMOVSQD,
    VPMOVSQW,
    VPMOVSWB,
    VPMOVSXBD,
    VPMOVSXBQ,
    VPMOVSXBW,
    VPMOVSXDQ,
    VPMOVSXWD,
    VPMOVSXWQ,
    VPMOVUSDB,
    VPMOVUSDW,
    VPMOVUSQB,
    VPMOVUSQD,
    VPMOVUSQW,
    VPMOVUSWB,
    VPMOVW2M,
    VPMOVWB,
    VPMOVZXBD,
    VPMOVZXBQ,
    VPMOVZXBW,
    VPMOVZXDQ,
    VPMOVZXWD,
    VPMOVZXWQ,
    VPMULDQ,
    VPMULHRSW,
    VPMULHUW,
    VPMULHW,
    VPMULLD,
    VPMULLQ,
    VPMULLW,
    VPMULTISHIFTQB,
    VPMULUDQ,
    VPOPCNTB,
    VPOPCNTD,
    VPOPCNTQ,
    VPOPCNTW,
    VPOR,
    VPORD,
    VPORQ,
    VPPERM,
    VPROLD,
    VPROLQ,
    VPROLVD,
    VPROLVQ,
    VPRORD,
    VPRORQ,
    VPRORVD,
    VPRORVQ,
    VPROTB,
    VPROTD,
    VPROTQ,
    VPROTW,
    VPSADBW,
    VPSCATTERDD,
    VPSCATTERDQ,
    VPSCATTERQD,
    VPSCATTERQQ,
    VPSHAB,
    VPSHAD,
    VPSHAQ,
    VPSHAW,
    VPSHLB,
    VPSHLD,
    VPSHLDD,
    VPSHLDQ,
    VPSHLDVD,
    VPSHLDVQ,
    VPSHLDVW,
    VPSHLDW,
    VPSHLQ,
    VPSHLW,
    VPSHRDD,
    VPSHRDQ,
    VPSHRDVD,
    VPSHRDVQ,
    VPSHRDVW,
    VPSHRDW,
    VPSHUFB,
    VPSHUFBITQMB,
    VPSHUFD,
    VPSHUFHW,
    VPSHUFLW,
    VPSIGNB,
    VPSIGND,
    VPSIGNW,
    VPSLLD,
    VPSLLDQ,
    VPSLLQ,
    VPSLLVD,
    VPSLLVQ,
    VPSLLVW,
    VPSLLW,
    VPSRAD,
    VPSRAQ,
    VPSRAVD,
    VPSRAVQ,
    VPSRAVW,
    VPSRAW,
    VPSRLD,
    VPSRLDQ,
    VPSRLQ,
    VPSRLVD,
    VPSRLVQ,
    VPSRLVW,
    VPSRLW,
    VPSUBB,
    VPSUBD,
    VPSUBQ,
    VPSUBSB,
    VPSUBSW,
    VPSUBUSB,
    VPSUBUSW,
    VPSUBW,
    VPTERNLOGD,
    VPTERNLOGQ,
    VPTEST,
    VPTESTMB,
    VPTESTMD,
    VPTESTMQ,
    VPTESTMW,
    VPTESTNMB,
    VPTESTNMD,
    VPTESTNMQ,
    VPTESTNMW,
    VPUNPCKHBW,
    VPUNPCKHDQ,
    VPUNPCKHQDQ,
    VPUNPCKHWD,
    VPUNPCKLBW,
    VPUNPCKLDQ,
    VPUNPCKLQDQ,
    VPUNPCKLWD,
    VPXOR,
    VPXORD,
    VPXORQ,
    VRANGEPD,
    VRANGEPS,
    VRANGESD,
    VRANGESS,
    VRCP14PD,
    VRCP14PS,
    VRCP14SD,
    VRCP14SS,
    VRCP28PD,
    VRCP28PS,
    VRCP28SD,
    VRCP28SS,
    VRCPPBF16,
    VRCPPH,
    VRCPPS,
    VRCPSH,
    VRCPSS,
    VREDUCENEPBF16,
    VREDUCEPD,
    VREDUCEPH,
    VREDUCEPS,
    VREDUCESD,
    VREDUCESH,
    VREDUCESS,
    VRNDSCALENEPBF16,
    VRNDSCALEPD,
    VRNDSCALEPH,
    VRNDSCALEPS,
    VRNDSCALESD,
    VRNDSCALESH,
    VRNDSCALESS,
    VROUNDPD,
    VROUNDPS,
    VROUNDSD,
    VROUNDSS,
    VRSQRT14PD,
    VRSQRT14PS,
    VRSQRT14SD,
    VRSQRT14SS,
    VRSQRT28PD,
    VRSQRT28PS,
    VRSQRT28SD,
    VRSQRT28SS,
    VRSQRTPBF16,
    VRSQRTPH,
    VRSQRTPS,
    VRSQRTSH,
    VRSQRTSS,
    VSCALEFPBF16,
    VSCALEFPD,
    VSCALEFPH,
    VSCALEFPS,
    VSCALEFSD,
    VSCALEFSH,
    VSCALEFSS,
    VSCATTERDPD,
    VSCATTERDPS,
    VSCATTERPF0DPD,
    VSCATTERPF0DPS,
    VSCATTERPF0QPD,
    VSCATTERPF0QPS,
    VSCATTERPF1DPD,
    VSCATTERPF1DPS,
    VSCATTERPF1QPD,
    VSCATTERPF1QPS,
    VSCATTERQPD,
    VSCATTERQPS,
    VSHA512MSG1,
    VSHA512MSG2,
    VSHA512RNDS2,
    VSHUFF32X4,
    VSHUFF64X2,
    VSHUFI32X4,
    VSHUFI64X2,
    VSHUFPD,
    VSHUFPS,
    VSM3MSG1,
    VSM3MSG2,
    VSM3RNDS2,
    VSM4KEY4,
    VSM4RNDS4,
    VSQRTNEPBF16,
    VSQRTPD,
    VSQRTPH,
    VSQRTPS,
    VSQRTSD,
    VSQRTSH,
    VSQRTSS,
    VSTMXCSR,
    VSUBNEPBF16,
    VSUBPD,
    VSUBPH,
    VSUBPS,
    VSUBSD,
    VSUBSH,
    VSUBSS,
    VTESTPD,
    VTESTPS,
    VUCOMISD,
    VUCOMISH,
    VUCOMISS,
    VUCOMXSD,
    VUCOMXSH,
    VUCOMXSS,
    VUNPCKHPD,
    VUNPCKHPS,
    VUNPCKLPD,
    VUNPCKLPS,
    VXORPD,
    VXORPS,
    VZEROALL,
    VZEROUPPER,
    WAIT,
    WBINVD,
    WBNOINVD,
    WRFSBASE,
    WRGSBASE,
    WRMSR,
    WRMSRLIST,
    WRMSRNS,
    WRPKRU,
    WRSS,
    WRUSS,
    XABORT,
    XADD,
    XBEGIN,
    XCHG,
    XEND,
    XGETBV,
    XLATB,
    XOR,
    XORPD,
    XORPS,
    XRESLDTRK,
    XRSTOR,
    XRSTORS,
    XSAVE,
    XSAVEC,
    XSAVEOPT,
    XSAVES,
    XSETBV,
    XSUSLDTRK,
    XTEST,
}

#[doc(hidden)]
impl TryFrom<ffi::ND_INS_CLASS> for Mnemonic {
    type Error = DecodeError;

    #[allow(clippy::too_many_lines)]
    fn try_from(value: ffi::ND_INS_CLASS) -> Result<Self, Self::Error> {
        match value {
            ffi::_ND_INS_CLASS::ND_INS_INVALID => Err(DecodeError::InternalError(value as u64)),
            ffi::_ND_INS_CLASS::ND_INS_AAA => Ok(Mnemonic::AAA),
            ffi::_ND_INS_CLASS::ND_INS_AAD => Ok(Mnemonic::AAD),
            ffi::_ND_INS_CLASS::ND_INS_AADD => Ok(Mnemonic::AADD),
            ffi::_ND_INS_CLASS::ND_INS_AAM => Ok(Mnemonic::AAM),
            ffi::_ND_INS_CLASS::ND_INS_AAND => Ok(Mnemonic::AAND),
            ffi::_ND_INS_CLASS::ND_INS_AAS => Ok(Mnemonic::AAS),
            ffi::_ND_INS_CLASS::ND_INS_ADC => Ok(Mnemonic::ADC),
            ffi::_ND_INS_CLASS::ND_INS_ADCX => Ok(Mnemonic::ADCX),
            ffi::_ND_INS_CLASS::ND_INS_ADD => Ok(Mnemonic::ADD),
            ffi::_ND_INS_CLASS::ND_INS_ADDPD => Ok(Mnemonic::ADDPD),
            ffi::_ND_INS_CLASS::ND_INS_ADDPS => Ok(Mnemonic::ADDPS),
            ffi::_ND_INS_CLASS::ND_INS_ADDSD => Ok(Mnemonic::ADDSD),
            ffi::_ND_INS_CLASS::ND_INS_ADDSS => Ok(Mnemonic::ADDSS),
            ffi::_ND_INS_CLASS::ND_INS_ADDSUBPD => Ok(Mnemonic::ADDSUBPD),
            ffi::_ND_INS_CLASS::ND_INS_ADDSUBPS => Ok(Mnemonic::ADDSUBPS),
            ffi::_ND_INS_CLASS::ND_INS_ADOX => Ok(Mnemonic::ADOX),
            ffi::_ND_INS_CLASS::ND_INS_AESDEC => Ok(Mnemonic::AESDEC),
            ffi::_ND_INS_CLASS::ND_INS_AESDEC128KL => Ok(Mnemonic::AESDEC128KL),
            ffi::_ND_INS_CLASS::ND_INS_AESDEC256KL => Ok(Mnemonic::AESDEC256KL),
            ffi::_ND_INS_CLASS::ND_INS_AESDECLAST => Ok(Mnemonic::AESDECLAST),
            ffi::_ND_INS_CLASS::ND_INS_AESDECWIDE128KL => Ok(Mnemonic::AESDECWIDE128KL),
            ffi::_ND_INS_CLASS::ND_INS_AESDECWIDE256KL => Ok(Mnemonic::AESDECWIDE256KL),
            ffi::_ND_INS_CLASS::ND_INS_AESENC => Ok(Mnemonic::AESENC),
            ffi::_ND_INS_CLASS::ND_INS_AESENC128KL => Ok(Mnemonic::AESENC128KL),
            ffi::_ND_INS_CLASS::ND_INS_AESENC256KL => Ok(Mnemonic::AESENC256KL),
            ffi::_ND_INS_CLASS::ND_INS_AESENCLAST => Ok(Mnemonic::AESENCLAST),
            ffi::_ND_INS_CLASS::ND_INS_AESENCWIDE128KL => Ok(Mnemonic::AESENCWIDE128KL),
            ffi::_ND_INS_CLASS::ND_INS_AESENCWIDE256KL => Ok(Mnemonic::AESENCWIDE256KL),
            ffi::_ND_INS_CLASS::ND_INS_AESIMC => Ok(Mnemonic::AESIMC),
            ffi::_ND_INS_CLASS::ND_INS_AESKEYGENASSIST => Ok(Mnemonic::AESKEYGENASSIST),
            ffi::_ND_INS_CLASS::ND_INS_AND => Ok(Mnemonic::AND),
            ffi::_ND_INS_CLASS::ND_INS_ANDN => Ok(Mnemonic::ANDN),
            ffi::_ND_INS_CLASS::ND_INS_ANDNPD => Ok(Mnemonic::ANDNPD),
            ffi::_ND_INS_CLASS::ND_INS_ANDNPS => Ok(Mnemonic::ANDNPS),
            ffi::_ND_INS_CLASS::ND_INS_ANDPD => Ok(Mnemonic::ANDPD),
            ffi::_ND_INS_CLASS::ND_INS_ANDPS => Ok(Mnemonic::ANDPS),
            ffi::_ND_INS_CLASS::ND_INS_AOR => Ok(Mnemonic::AOR),
            ffi::_ND_INS_CLASS::ND_INS_ARPL => Ok(Mnemonic::ARPL),
            ffi::_ND_INS_CLASS::ND_INS_AXOR => Ok(Mnemonic::AXOR),
            ffi::_ND_INS_CLASS::ND_INS_BEXTR => Ok(Mnemonic::BEXTR),
            ffi::_ND_INS_CLASS::ND_INS_BLCFILL => Ok(Mnemonic::BLCFILL),
            ffi::_ND_INS_CLASS::ND_INS_BLCI => Ok(Mnemonic::BLCI),
            ffi::_ND_INS_CLASS::ND_INS_BLCIC => Ok(Mnemonic::BLCIC),
            ffi::_ND_INS_CLASS::ND_INS_BLCMSK => Ok(Mnemonic::BLCMSK),
            ffi::_ND_INS_CLASS::ND_INS_BLCS => Ok(Mnemonic::BLCS),
            ffi::_ND_INS_CLASS::ND_INS_BLENDPD => Ok(Mnemonic::BLENDPD),
            ffi::_ND_INS_CLASS::ND_INS_BLENDPS => Ok(Mnemonic::BLENDPS),
            ffi::_ND_INS_CLASS::ND_INS_BLENDVPD => Ok(Mnemonic::BLENDVPD),
            ffi::_ND_INS_CLASS::ND_INS_BLENDVPS => Ok(Mnemonic::BLENDVPS),
            ffi::_ND_INS_CLASS::ND_INS_BLSFILL => Ok(Mnemonic::BLSFILL),
            ffi::_ND_INS_CLASS::ND_INS_BLSI => Ok(Mnemonic::BLSI),
            ffi::_ND_INS_CLASS::ND_INS_BLSIC => Ok(Mnemonic::BLSIC),
            ffi::_ND_INS_CLASS::ND_INS_BLSMSK => Ok(Mnemonic::BLSMSK),
            ffi::_ND_INS_CLASS::ND_INS_BLSR => Ok(Mnemonic::BLSR),
            ffi::_ND_INS_CLASS::ND_INS_BNDCL => Ok(Mnemonic::BNDCL),
            ffi::_ND_INS_CLASS::ND_INS_BNDCN => Ok(Mnemonic::BNDCN),
            ffi::_ND_INS_CLASS::ND_INS_BNDCU => Ok(Mnemonic::BNDCU),
            ffi::_ND_INS_CLASS::ND_INS_BNDLDX => Ok(Mnemonic::BNDLDX),
            ffi::_ND_INS_CLASS::ND_INS_BNDMK => Ok(Mnemonic::BNDMK),
            ffi::_ND_INS_CLASS::ND_INS_BNDMOV => Ok(Mnemonic::BNDMOV),
            ffi::_ND_INS_CLASS::ND_INS_BNDSTX => Ok(Mnemonic::BNDSTX),
            ffi::_ND_INS_CLASS::ND_INS_BOUND => Ok(Mnemonic::BOUND),
            ffi::_ND_INS_CLASS::ND_INS_BSF => Ok(Mnemonic::BSF),
            ffi::_ND_INS_CLASS::ND_INS_BSR => Ok(Mnemonic::BSR),
            ffi::_ND_INS_CLASS::ND_INS_BSWAP => Ok(Mnemonic::BSWAP),
            ffi::_ND_INS_CLASS::ND_INS_BT => Ok(Mnemonic::BT),
            ffi::_ND_INS_CLASS::ND_INS_BTC => Ok(Mnemonic::BTC),
            ffi::_ND_INS_CLASS::ND_INS_BTR => Ok(Mnemonic::BTR),
            ffi::_ND_INS_CLASS::ND_INS_BTS => Ok(Mnemonic::BTS),
            ffi::_ND_INS_CLASS::ND_INS_BZHI => Ok(Mnemonic::BZHI),
            ffi::_ND_INS_CLASS::ND_INS_CALLFD => Ok(Mnemonic::CALLFD),
            ffi::_ND_INS_CLASS::ND_INS_CALLFI => Ok(Mnemonic::CALLFI),
            ffi::_ND_INS_CLASS::ND_INS_CALLNI => Ok(Mnemonic::CALLNI),
            ffi::_ND_INS_CLASS::ND_INS_CALLNR => Ok(Mnemonic::CALLNR),
            ffi::_ND_INS_CLASS::ND_INS_CBW => Ok(Mnemonic::CBW),
            ffi::_ND_INS_CLASS::ND_INS_CCMP => Ok(Mnemonic::CCMP),
            ffi::_ND_INS_CLASS::ND_INS_CDQ => Ok(Mnemonic::CDQ),
            ffi::_ND_INS_CLASS::ND_INS_CDQE => Ok(Mnemonic::CDQE),
            ffi::_ND_INS_CLASS::ND_INS_CFCMOV => Ok(Mnemonic::CFCMOV),
            ffi::_ND_INS_CLASS::ND_INS_CLAC => Ok(Mnemonic::CLAC),
            ffi::_ND_INS_CLASS::ND_INS_CLC => Ok(Mnemonic::CLC),
            ffi::_ND_INS_CLASS::ND_INS_CLD => Ok(Mnemonic::CLD),
            ffi::_ND_INS_CLASS::ND_INS_CLDEMOTE => Ok(Mnemonic::CLDEMOTE),
            ffi::_ND_INS_CLASS::ND_INS_CLEVICT0 => Ok(Mnemonic::CLEVICT0),
            ffi::_ND_INS_CLASS::ND_INS_CLEVICT1 => Ok(Mnemonic::CLEVICT1),
            ffi::_ND_INS_CLASS::ND_INS_CLFLUSH => Ok(Mnemonic::CLFLUSH),
            ffi::_ND_INS_CLASS::ND_INS_CLFLUSHOPT => Ok(Mnemonic::CLFLUSHOPT),
            ffi::_ND_INS_CLASS::ND_INS_CLGI => Ok(Mnemonic::CLGI),
            ffi::_ND_INS_CLASS::ND_INS_CLI => Ok(Mnemonic::CLI),
            ffi::_ND_INS_CLASS::ND_INS_CLRSSBSY => Ok(Mnemonic::CLRSSBSY),
            ffi::_ND_INS_CLASS::ND_INS_CLTS => Ok(Mnemonic::CLTS),
            ffi::_ND_INS_CLASS::ND_INS_CLUI => Ok(Mnemonic::CLUI),
            ffi::_ND_INS_CLASS::ND_INS_CLWB => Ok(Mnemonic::CLWB),
            ffi::_ND_INS_CLASS::ND_INS_CLZERO => Ok(Mnemonic::CLZERO),
            ffi::_ND_INS_CLASS::ND_INS_CMC => Ok(Mnemonic::CMC),
            ffi::_ND_INS_CLASS::ND_INS_CMOVcc => Ok(Mnemonic::CMOVcc),
            ffi::_ND_INS_CLASS::ND_INS_CMP => Ok(Mnemonic::CMP),
            ffi::_ND_INS_CLASS::ND_INS_CMPBEXADD => Ok(Mnemonic::CMPBEXADD),
            ffi::_ND_INS_CLASS::ND_INS_CMPCXADD => Ok(Mnemonic::CMPCXADD),
            ffi::_ND_INS_CLASS::ND_INS_CMPLEXADD => Ok(Mnemonic::CMPLEXADD),
            ffi::_ND_INS_CLASS::ND_INS_CMPLXADD => Ok(Mnemonic::CMPLXADD),
            ffi::_ND_INS_CLASS::ND_INS_CMPNBEXADD => Ok(Mnemonic::CMPNBEXADD),
            ffi::_ND_INS_CLASS::ND_INS_CMPNCXADD => Ok(Mnemonic::CMPNCXADD),
            ffi::_ND_INS_CLASS::ND_INS_CMPNLEXADD => Ok(Mnemonic::CMPNLEXADD),
            ffi::_ND_INS_CLASS::ND_INS_CMPNLXADD => Ok(Mnemonic::CMPNLXADD),
            ffi::_ND_INS_CLASS::ND_INS_CMPNOXADD => Ok(Mnemonic::CMPNOXADD),
            ffi::_ND_INS_CLASS::ND_INS_CMPNPXADD => Ok(Mnemonic::CMPNPXADD),
            ffi::_ND_INS_CLASS::ND_INS_CMPNSXADD => Ok(Mnemonic::CMPNSXADD),
            ffi::_ND_INS_CLASS::ND_INS_CMPNZXADD => Ok(Mnemonic::CMPNZXADD),
            ffi::_ND_INS_CLASS::ND_INS_CMPOXADD => Ok(Mnemonic::CMPOXADD),
            ffi::_ND_INS_CLASS::ND_INS_CMPPD => Ok(Mnemonic::CMPPD),
            ffi::_ND_INS_CLASS::ND_INS_CMPPS => Ok(Mnemonic::CMPPS),
            ffi::_ND_INS_CLASS::ND_INS_CMPPXADD => Ok(Mnemonic::CMPPXADD),
            ffi::_ND_INS_CLASS::ND_INS_CMPS => Ok(Mnemonic::CMPS),
            ffi::_ND_INS_CLASS::ND_INS_CMPSD => Ok(Mnemonic::CMPSD),
            ffi::_ND_INS_CLASS::ND_INS_CMPSS => Ok(Mnemonic::CMPSS),
            ffi::_ND_INS_CLASS::ND_INS_CMPSXADD => Ok(Mnemonic::CMPSXADD),
            ffi::_ND_INS_CLASS::ND_INS_CMPXCHG => Ok(Mnemonic::CMPXCHG),
            ffi::_ND_INS_CLASS::ND_INS_CMPXCHG16B => Ok(Mnemonic::CMPXCHG16B),
            ffi::_ND_INS_CLASS::ND_INS_CMPXCHG8B => Ok(Mnemonic::CMPXCHG8B),
            ffi::_ND_INS_CLASS::ND_INS_CMPZXADD => Ok(Mnemonic::CMPZXADD),
            ffi::_ND_INS_CLASS::ND_INS_COMISD => Ok(Mnemonic::COMISD),
            ffi::_ND_INS_CLASS::ND_INS_COMISS => Ok(Mnemonic::COMISS),
            ffi::_ND_INS_CLASS::ND_INS_CPUID => Ok(Mnemonic::CPUID),
            ffi::_ND_INS_CLASS::ND_INS_CQO => Ok(Mnemonic::CQO),
            ffi::_ND_INS_CLASS::ND_INS_CRC32 => Ok(Mnemonic::CRC32),
            ffi::_ND_INS_CLASS::ND_INS_CTEST => Ok(Mnemonic::CTEST),
            ffi::_ND_INS_CLASS::ND_INS_CVTDQ2PD => Ok(Mnemonic::CVTDQ2PD),
            ffi::_ND_INS_CLASS::ND_INS_CVTDQ2PS => Ok(Mnemonic::CVTDQ2PS),
            ffi::_ND_INS_CLASS::ND_INS_CVTPD2DQ => Ok(Mnemonic::CVTPD2DQ),
            ffi::_ND_INS_CLASS::ND_INS_CVTPD2PI => Ok(Mnemonic::CVTPD2PI),
            ffi::_ND_INS_CLASS::ND_INS_CVTPD2PS => Ok(Mnemonic::CVTPD2PS),
            ffi::_ND_INS_CLASS::ND_INS_CVTPI2PD => Ok(Mnemonic::CVTPI2PD),
            ffi::_ND_INS_CLASS::ND_INS_CVTPI2PS => Ok(Mnemonic::CVTPI2PS),
            ffi::_ND_INS_CLASS::ND_INS_CVTPS2DQ => Ok(Mnemonic::CVTPS2DQ),
            ffi::_ND_INS_CLASS::ND_INS_CVTPS2PD => Ok(Mnemonic::CVTPS2PD),
            ffi::_ND_INS_CLASS::ND_INS_CVTPS2PI => Ok(Mnemonic::CVTPS2PI),
            ffi::_ND_INS_CLASS::ND_INS_CVTSD2SI => Ok(Mnemonic::CVTSD2SI),
            ffi::_ND_INS_CLASS::ND_INS_CVTSD2SS => Ok(Mnemonic::CVTSD2SS),
            ffi::_ND_INS_CLASS::ND_INS_CVTSI2SD => Ok(Mnemonic::CVTSI2SD),
            ffi::_ND_INS_CLASS::ND_INS_CVTSI2SS => Ok(Mnemonic::CVTSI2SS),
            ffi::_ND_INS_CLASS::ND_INS_CVTSS2SD => Ok(Mnemonic::CVTSS2SD),
            ffi::_ND_INS_CLASS::ND_INS_CVTSS2SI => Ok(Mnemonic::CVTSS2SI),
            ffi::_ND_INS_CLASS::ND_INS_CVTTPD2DQ => Ok(Mnemonic::CVTTPD2DQ),
            ffi::_ND_INS_CLASS::ND_INS_CVTTPD2PI => Ok(Mnemonic::CVTTPD2PI),
            ffi::_ND_INS_CLASS::ND_INS_CVTTPS2DQ => Ok(Mnemonic::CVTTPS2DQ),
            ffi::_ND_INS_CLASS::ND_INS_CVTTPS2PI => Ok(Mnemonic::CVTTPS2PI),
            ffi::_ND_INS_CLASS::ND_INS_CVTTSD2SI => Ok(Mnemonic::CVTTSD2SI),
            ffi::_ND_INS_CLASS::ND_INS_CVTTSS2SI => Ok(Mnemonic::CVTTSS2SI),
            ffi::_ND_INS_CLASS::ND_INS_CWD => Ok(Mnemonic::CWD),
            ffi::_ND_INS_CLASS::ND_INS_CWDE => Ok(Mnemonic::CWDE),
            ffi::_ND_INS_CLASS::ND_INS_DAA => Ok(Mnemonic::DAA),
            ffi::_ND_INS_CLASS::ND_INS_DAS => Ok(Mnemonic::DAS),
            ffi::_ND_INS_CLASS::ND_INS_DEC => Ok(Mnemonic::DEC),
            ffi::_ND_INS_CLASS::ND_INS_DELAY => Ok(Mnemonic::DELAY),
            ffi::_ND_INS_CLASS::ND_INS_DIV => Ok(Mnemonic::DIV),
            ffi::_ND_INS_CLASS::ND_INS_DIVPD => Ok(Mnemonic::DIVPD),
            ffi::_ND_INS_CLASS::ND_INS_DIVPS => Ok(Mnemonic::DIVPS),
            ffi::_ND_INS_CLASS::ND_INS_DIVSD => Ok(Mnemonic::DIVSD),
            ffi::_ND_INS_CLASS::ND_INS_DIVSS => Ok(Mnemonic::DIVSS),
            ffi::_ND_INS_CLASS::ND_INS_DPPD => Ok(Mnemonic::DPPD),
            ffi::_ND_INS_CLASS::ND_INS_DPPS => Ok(Mnemonic::DPPS),
            ffi::_ND_INS_CLASS::ND_INS_EMMS => Ok(Mnemonic::EMMS),
            ffi::_ND_INS_CLASS::ND_INS_ENCLS => Ok(Mnemonic::ENCLS),
            ffi::_ND_INS_CLASS::ND_INS_ENCLU => Ok(Mnemonic::ENCLU),
            ffi::_ND_INS_CLASS::ND_INS_ENCLV => Ok(Mnemonic::ENCLV),
            ffi::_ND_INS_CLASS::ND_INS_ENCODEKEY128 => Ok(Mnemonic::ENCODEKEY128),
            ffi::_ND_INS_CLASS::ND_INS_ENCODEKEY256 => Ok(Mnemonic::ENCODEKEY256),
            ffi::_ND_INS_CLASS::ND_INS_ENDBR => Ok(Mnemonic::ENDBR),
            ffi::_ND_INS_CLASS::ND_INS_ENQCMD => Ok(Mnemonic::ENQCMD),
            ffi::_ND_INS_CLASS::ND_INS_ENQCMDS => Ok(Mnemonic::ENQCMDS),
            ffi::_ND_INS_CLASS::ND_INS_ENTER => Ok(Mnemonic::ENTER),
            ffi::_ND_INS_CLASS::ND_INS_ERETS => Ok(Mnemonic::ERETS),
            ffi::_ND_INS_CLASS::ND_INS_ERETU => Ok(Mnemonic::ERETU),
            ffi::_ND_INS_CLASS::ND_INS_EXTRACTPS => Ok(Mnemonic::EXTRACTPS),
            ffi::_ND_INS_CLASS::ND_INS_EXTRQ => Ok(Mnemonic::EXTRQ),
            ffi::_ND_INS_CLASS::ND_INS_F2XM1 => Ok(Mnemonic::F2XM1),
            ffi::_ND_INS_CLASS::ND_INS_FABS => Ok(Mnemonic::FABS),
            ffi::_ND_INS_CLASS::ND_INS_FADD => Ok(Mnemonic::FADD),
            ffi::_ND_INS_CLASS::ND_INS_FADDP => Ok(Mnemonic::FADDP),
            ffi::_ND_INS_CLASS::ND_INS_FBLD => Ok(Mnemonic::FBLD),
            ffi::_ND_INS_CLASS::ND_INS_FBSTP => Ok(Mnemonic::FBSTP),
            ffi::_ND_INS_CLASS::ND_INS_FCHS => Ok(Mnemonic::FCHS),
            ffi::_ND_INS_CLASS::ND_INS_FCMOVB => Ok(Mnemonic::FCMOVB),
            ffi::_ND_INS_CLASS::ND_INS_FCMOVBE => Ok(Mnemonic::FCMOVBE),
            ffi::_ND_INS_CLASS::ND_INS_FCMOVE => Ok(Mnemonic::FCMOVE),
            ffi::_ND_INS_CLASS::ND_INS_FCMOVNB => Ok(Mnemonic::FCMOVNB),
            ffi::_ND_INS_CLASS::ND_INS_FCMOVNBE => Ok(Mnemonic::FCMOVNBE),
            ffi::_ND_INS_CLASS::ND_INS_FCMOVNE => Ok(Mnemonic::FCMOVNE),
            ffi::_ND_INS_CLASS::ND_INS_FCMOVNU => Ok(Mnemonic::FCMOVNU),
            ffi::_ND_INS_CLASS::ND_INS_FCMOVU => Ok(Mnemonic::FCMOVU),
            ffi::_ND_INS_CLASS::ND_INS_FCOM => Ok(Mnemonic::FCOM),
            ffi::_ND_INS_CLASS::ND_INS_FCOMI => Ok(Mnemonic::FCOMI),
            ffi::_ND_INS_CLASS::ND_INS_FCOMIP => Ok(Mnemonic::FCOMIP),
            ffi::_ND_INS_CLASS::ND_INS_FCOMP => Ok(Mnemonic::FCOMP),
            ffi::_ND_INS_CLASS::ND_INS_FCOMPP => Ok(Mnemonic::FCOMPP),
            ffi::_ND_INS_CLASS::ND_INS_FCOS => Ok(Mnemonic::FCOS),
            ffi::_ND_INS_CLASS::ND_INS_FDECSTP => Ok(Mnemonic::FDECSTP),
            ffi::_ND_INS_CLASS::ND_INS_FDIV => Ok(Mnemonic::FDIV),
            ffi::_ND_INS_CLASS::ND_INS_FDIVP => Ok(Mnemonic::FDIVP),
            ffi::_ND_INS_CLASS::ND_INS_FDIVR => Ok(Mnemonic::FDIVR),
            ffi::_ND_INS_CLASS::ND_INS_FDIVRP => Ok(Mnemonic::FDIVRP),
            ffi::_ND_INS_CLASS::ND_INS_FEMMS => Ok(Mnemonic::FEMMS),
            ffi::_ND_INS_CLASS::ND_INS_FFREE => Ok(Mnemonic::FFREE),
            ffi::_ND_INS_CLASS::ND_INS_FFREEP => Ok(Mnemonic::FFREEP),
            ffi::_ND_INS_CLASS::ND_INS_FIADD => Ok(Mnemonic::FIADD),
            ffi::_ND_INS_CLASS::ND_INS_FICOM => Ok(Mnemonic::FICOM),
            ffi::_ND_INS_CLASS::ND_INS_FICOMP => Ok(Mnemonic::FICOMP),
            ffi::_ND_INS_CLASS::ND_INS_FIDIV => Ok(Mnemonic::FIDIV),
            ffi::_ND_INS_CLASS::ND_INS_FIDIVR => Ok(Mnemonic::FIDIVR),
            ffi::_ND_INS_CLASS::ND_INS_FILD => Ok(Mnemonic::FILD),
            ffi::_ND_INS_CLASS::ND_INS_FIMUL => Ok(Mnemonic::FIMUL),
            ffi::_ND_INS_CLASS::ND_INS_FINCSTP => Ok(Mnemonic::FINCSTP),
            ffi::_ND_INS_CLASS::ND_INS_FIST => Ok(Mnemonic::FIST),
            ffi::_ND_INS_CLASS::ND_INS_FISTP => Ok(Mnemonic::FISTP),
            ffi::_ND_INS_CLASS::ND_INS_FISTTP => Ok(Mnemonic::FISTTP),
            ffi::_ND_INS_CLASS::ND_INS_FISUB => Ok(Mnemonic::FISUB),
            ffi::_ND_INS_CLASS::ND_INS_FISUBR => Ok(Mnemonic::FISUBR),
            ffi::_ND_INS_CLASS::ND_INS_FLD => Ok(Mnemonic::FLD),
            ffi::_ND_INS_CLASS::ND_INS_FLD1 => Ok(Mnemonic::FLD1),
            ffi::_ND_INS_CLASS::ND_INS_FLDCW => Ok(Mnemonic::FLDCW),
            ffi::_ND_INS_CLASS::ND_INS_FLDENV => Ok(Mnemonic::FLDENV),
            ffi::_ND_INS_CLASS::ND_INS_FLDL2E => Ok(Mnemonic::FLDL2E),
            ffi::_ND_INS_CLASS::ND_INS_FLDL2T => Ok(Mnemonic::FLDL2T),
            ffi::_ND_INS_CLASS::ND_INS_FLDLG2 => Ok(Mnemonic::FLDLG2),
            ffi::_ND_INS_CLASS::ND_INS_FLDLN2 => Ok(Mnemonic::FLDLN2),
            ffi::_ND_INS_CLASS::ND_INS_FLDPI => Ok(Mnemonic::FLDPI),
            ffi::_ND_INS_CLASS::ND_INS_FLDZ => Ok(Mnemonic::FLDZ),
            ffi::_ND_INS_CLASS::ND_INS_FMUL => Ok(Mnemonic::FMUL),
            ffi::_ND_INS_CLASS::ND_INS_FMULP => Ok(Mnemonic::FMULP),
            ffi::_ND_INS_CLASS::ND_INS_FNCLEX => Ok(Mnemonic::FNCLEX),
            ffi::_ND_INS_CLASS::ND_INS_FNDISI => Ok(Mnemonic::FNDISI),
            ffi::_ND_INS_CLASS::ND_INS_FNINIT => Ok(Mnemonic::FNINIT),
            ffi::_ND_INS_CLASS::ND_INS_FNOP => Ok(Mnemonic::FNOP),
            ffi::_ND_INS_CLASS::ND_INS_FNSAVE => Ok(Mnemonic::FNSAVE),
            ffi::_ND_INS_CLASS::ND_INS_FNSTCW => Ok(Mnemonic::FNSTCW),
            ffi::_ND_INS_CLASS::ND_INS_FNSTENV => Ok(Mnemonic::FNSTENV),
            ffi::_ND_INS_CLASS::ND_INS_FNSTSW => Ok(Mnemonic::FNSTSW),
            ffi::_ND_INS_CLASS::ND_INS_FPATAN => Ok(Mnemonic::FPATAN),
            ffi::_ND_INS_CLASS::ND_INS_FPREM => Ok(Mnemonic::FPREM),
            ffi::_ND_INS_CLASS::ND_INS_FPREM1 => Ok(Mnemonic::FPREM1),
            ffi::_ND_INS_CLASS::ND_INS_FPTAN => Ok(Mnemonic::FPTAN),
            ffi::_ND_INS_CLASS::ND_INS_FRINEAR => Ok(Mnemonic::FRINEAR),
            ffi::_ND_INS_CLASS::ND_INS_FRNDINT => Ok(Mnemonic::FRNDINT),
            ffi::_ND_INS_CLASS::ND_INS_FRSTOR => Ok(Mnemonic::FRSTOR),
            ffi::_ND_INS_CLASS::ND_INS_FSCALE => Ok(Mnemonic::FSCALE),
            ffi::_ND_INS_CLASS::ND_INS_FSIN => Ok(Mnemonic::FSIN),
            ffi::_ND_INS_CLASS::ND_INS_FSINCOS => Ok(Mnemonic::FSINCOS),
            ffi::_ND_INS_CLASS::ND_INS_FSQRT => Ok(Mnemonic::FSQRT),
            ffi::_ND_INS_CLASS::ND_INS_FST => Ok(Mnemonic::FST),
            ffi::_ND_INS_CLASS::ND_INS_FSTDW => Ok(Mnemonic::FSTDW),
            ffi::_ND_INS_CLASS::ND_INS_FSTP => Ok(Mnemonic::FSTP),
            ffi::_ND_INS_CLASS::ND_INS_FSTPNCE => Ok(Mnemonic::FSTPNCE),
            ffi::_ND_INS_CLASS::ND_INS_FSTSG => Ok(Mnemonic::FSTSG),
            ffi::_ND_INS_CLASS::ND_INS_FSUB => Ok(Mnemonic::FSUB),
            ffi::_ND_INS_CLASS::ND_INS_FSUBP => Ok(Mnemonic::FSUBP),
            ffi::_ND_INS_CLASS::ND_INS_FSUBR => Ok(Mnemonic::FSUBR),
            ffi::_ND_INS_CLASS::ND_INS_FSUBRP => Ok(Mnemonic::FSUBRP),
            ffi::_ND_INS_CLASS::ND_INS_FTST => Ok(Mnemonic::FTST),
            ffi::_ND_INS_CLASS::ND_INS_FUCOM => Ok(Mnemonic::FUCOM),
            ffi::_ND_INS_CLASS::ND_INS_FUCOMI => Ok(Mnemonic::FUCOMI),
            ffi::_ND_INS_CLASS::ND_INS_FUCOMIP => Ok(Mnemonic::FUCOMIP),
            ffi::_ND_INS_CLASS::ND_INS_FUCOMP => Ok(Mnemonic::FUCOMP),
            ffi::_ND_INS_CLASS::ND_INS_FUCOMPP => Ok(Mnemonic::FUCOMPP),
            ffi::_ND_INS_CLASS::ND_INS_FXAM => Ok(Mnemonic::FXAM),
            ffi::_ND_INS_CLASS::ND_INS_FXCH => Ok(Mnemonic::FXCH),
            ffi::_ND_INS_CLASS::ND_INS_FXRSTOR => Ok(Mnemonic::FXRSTOR),
            ffi::_ND_INS_CLASS::ND_INS_FXRSTOR64 => Ok(Mnemonic::FXRSTOR64),
            ffi::_ND_INS_CLASS::ND_INS_FXSAVE => Ok(Mnemonic::FXSAVE),
            ffi::_ND_INS_CLASS::ND_INS_FXSAVE64 => Ok(Mnemonic::FXSAVE64),
            ffi::_ND_INS_CLASS::ND_INS_FXTRACT => Ok(Mnemonic::FXTRACT),
            ffi::_ND_INS_CLASS::ND_INS_FYL2X => Ok(Mnemonic::FYL2X),
            ffi::_ND_INS_CLASS::ND_INS_FYL2XP1 => Ok(Mnemonic::FYL2XP1),
            ffi::_ND_INS_CLASS::ND_INS_GETSEC => Ok(Mnemonic::GETSEC),
            ffi::_ND_INS_CLASS::ND_INS_GF2P8AFFINEINVQB => Ok(Mnemonic::GF2P8AFFINEINVQB),
            ffi::_ND_INS_CLASS::ND_INS_GF2P8AFFINEQB => Ok(Mnemonic::GF2P8AFFINEQB),
            ffi::_ND_INS_CLASS::ND_INS_GF2P8MULB => Ok(Mnemonic::GF2P8MULB),
            ffi::_ND_INS_CLASS::ND_INS_HADDPD => Ok(Mnemonic::HADDPD),
            ffi::_ND_INS_CLASS::ND_INS_HADDPS => Ok(Mnemonic::HADDPS),
            ffi::_ND_INS_CLASS::ND_INS_HLT => Ok(Mnemonic::HLT),
            ffi::_ND_INS_CLASS::ND_INS_HRESET => Ok(Mnemonic::HRESET),
            ffi::_ND_INS_CLASS::ND_INS_HSUBPD => Ok(Mnemonic::HSUBPD),
            ffi::_ND_INS_CLASS::ND_INS_HSUBPS => Ok(Mnemonic::HSUBPS),
            ffi::_ND_INS_CLASS::ND_INS_IDIV => Ok(Mnemonic::IDIV),
            ffi::_ND_INS_CLASS::ND_INS_IMUL => Ok(Mnemonic::IMUL),
            ffi::_ND_INS_CLASS::ND_INS_IN => Ok(Mnemonic::IN),
            ffi::_ND_INS_CLASS::ND_INS_INC => Ok(Mnemonic::INC),
            ffi::_ND_INS_CLASS::ND_INS_INCSSP => Ok(Mnemonic::INCSSP),
            ffi::_ND_INS_CLASS::ND_INS_INS => Ok(Mnemonic::INS),
            ffi::_ND_INS_CLASS::ND_INS_INSERTPS => Ok(Mnemonic::INSERTPS),
            ffi::_ND_INS_CLASS::ND_INS_INSERTQ => Ok(Mnemonic::INSERTQ),
            ffi::_ND_INS_CLASS::ND_INS_INT => Ok(Mnemonic::INT),
            ffi::_ND_INS_CLASS::ND_INS_INT1 => Ok(Mnemonic::INT1),
            ffi::_ND_INS_CLASS::ND_INS_INT3 => Ok(Mnemonic::INT3),
            ffi::_ND_INS_CLASS::ND_INS_INTO => Ok(Mnemonic::INTO),
            ffi::_ND_INS_CLASS::ND_INS_INVD => Ok(Mnemonic::INVD),
            ffi::_ND_INS_CLASS::ND_INS_INVEPT => Ok(Mnemonic::INVEPT),
            ffi::_ND_INS_CLASS::ND_INS_INVLPG => Ok(Mnemonic::INVLPG),
            ffi::_ND_INS_CLASS::ND_INS_INVLPGA => Ok(Mnemonic::INVLPGA),
            ffi::_ND_INS_CLASS::ND_INS_INVLPGB => Ok(Mnemonic::INVLPGB),
            ffi::_ND_INS_CLASS::ND_INS_INVPCID => Ok(Mnemonic::INVPCID),
            ffi::_ND_INS_CLASS::ND_INS_INVVPID => Ok(Mnemonic::INVVPID),
            ffi::_ND_INS_CLASS::ND_INS_IRET => Ok(Mnemonic::IRET),
            ffi::_ND_INS_CLASS::ND_INS_JMPABS => Ok(Mnemonic::JMPABS),
            ffi::_ND_INS_CLASS::ND_INS_JMPE => Ok(Mnemonic::JMPE),
            ffi::_ND_INS_CLASS::ND_INS_JMPFD => Ok(Mnemonic::JMPFD),
            ffi::_ND_INS_CLASS::ND_INS_JMPFI => Ok(Mnemonic::JMPFI),
            ffi::_ND_INS_CLASS::ND_INS_JMPNI => Ok(Mnemonic::JMPNI),
            ffi::_ND_INS_CLASS::ND_INS_JMPNR => Ok(Mnemonic::JMPNR),
            ffi::_ND_INS_CLASS::ND_INS_Jcc => Ok(Mnemonic::Jcc),
            ffi::_ND_INS_CLASS::ND_INS_JrCXZ => Ok(Mnemonic::JrCXZ),
            ffi::_ND_INS_CLASS::ND_INS_KADD => Ok(Mnemonic::KADD),
            ffi::_ND_INS_CLASS::ND_INS_KAND => Ok(Mnemonic::KAND),
            ffi::_ND_INS_CLASS::ND_INS_KANDN => Ok(Mnemonic::KANDN),
            ffi::_ND_INS_CLASS::ND_INS_KMERGE2L1H => Ok(Mnemonic::KMERGE2L1H),
            ffi::_ND_INS_CLASS::ND_INS_KMERGE2L1L => Ok(Mnemonic::KMERGE2L1L),
            ffi::_ND_INS_CLASS::ND_INS_KMOV => Ok(Mnemonic::KMOV),
            ffi::_ND_INS_CLASS::ND_INS_KNOT => Ok(Mnemonic::KNOT),
            ffi::_ND_INS_CLASS::ND_INS_KOR => Ok(Mnemonic::KOR),
            ffi::_ND_INS_CLASS::ND_INS_KORTEST => Ok(Mnemonic::KORTEST),
            ffi::_ND_INS_CLASS::ND_INS_KSHIFTL => Ok(Mnemonic::KSHIFTL),
            ffi::_ND_INS_CLASS::ND_INS_KSHIFTR => Ok(Mnemonic::KSHIFTR),
            ffi::_ND_INS_CLASS::ND_INS_KTEST => Ok(Mnemonic::KTEST),
            ffi::_ND_INS_CLASS::ND_INS_KUNPCKBW => Ok(Mnemonic::KUNPCKBW),
            ffi::_ND_INS_CLASS::ND_INS_KUNPCKDQ => Ok(Mnemonic::KUNPCKDQ),
            ffi::_ND_INS_CLASS::ND_INS_KUNPCKWD => Ok(Mnemonic::KUNPCKWD),
            ffi::_ND_INS_CLASS::ND_INS_KXNOR => Ok(Mnemonic::KXNOR),
            ffi::_ND_INS_CLASS::ND_INS_KXOR => Ok(Mnemonic::KXOR),
            ffi::_ND_INS_CLASS::ND_INS_LAHF => Ok(Mnemonic::LAHF),
            ffi::_ND_INS_CLASS::ND_INS_LAR => Ok(Mnemonic::LAR),
            ffi::_ND_INS_CLASS::ND_INS_LDDQU => Ok(Mnemonic::LDDQU),
            ffi::_ND_INS_CLASS::ND_INS_LDMXCSR => Ok(Mnemonic::LDMXCSR),
            ffi::_ND_INS_CLASS::ND_INS_LDS => Ok(Mnemonic::LDS),
            ffi::_ND_INS_CLASS::ND_INS_LDTILECFG => Ok(Mnemonic::LDTILECFG),
            ffi::_ND_INS_CLASS::ND_INS_LEA => Ok(Mnemonic::LEA),
            ffi::_ND_INS_CLASS::ND_INS_LEAVE => Ok(Mnemonic::LEAVE),
            ffi::_ND_INS_CLASS::ND_INS_LES => Ok(Mnemonic::LES),
            ffi::_ND_INS_CLASS::ND_INS_LFENCE => Ok(Mnemonic::LFENCE),
            ffi::_ND_INS_CLASS::ND_INS_LFS => Ok(Mnemonic::LFS),
            ffi::_ND_INS_CLASS::ND_INS_LGDT => Ok(Mnemonic::LGDT),
            ffi::_ND_INS_CLASS::ND_INS_LGS => Ok(Mnemonic::LGS),
            ffi::_ND_INS_CLASS::ND_INS_LIDT => Ok(Mnemonic::LIDT),
            ffi::_ND_INS_CLASS::ND_INS_LKGS => Ok(Mnemonic::LKGS),
            ffi::_ND_INS_CLASS::ND_INS_LLDT => Ok(Mnemonic::LLDT),
            ffi::_ND_INS_CLASS::ND_INS_LLWPCB => Ok(Mnemonic::LLWPCB),
            ffi::_ND_INS_CLASS::ND_INS_LMSW => Ok(Mnemonic::LMSW),
            ffi::_ND_INS_CLASS::ND_INS_LOADIWKEY => Ok(Mnemonic::LOADIWKEY),
            ffi::_ND_INS_CLASS::ND_INS_LODS => Ok(Mnemonic::LODS),
            ffi::_ND_INS_CLASS::ND_INS_LOOP => Ok(Mnemonic::LOOP),
            ffi::_ND_INS_CLASS::ND_INS_LOOPNZ => Ok(Mnemonic::LOOPNZ),
            ffi::_ND_INS_CLASS::ND_INS_LOOPZ => Ok(Mnemonic::LOOPZ),
            ffi::_ND_INS_CLASS::ND_INS_LSL => Ok(Mnemonic::LSL),
            ffi::_ND_INS_CLASS::ND_INS_LSS => Ok(Mnemonic::LSS),
            ffi::_ND_INS_CLASS::ND_INS_LTR => Ok(Mnemonic::LTR),
            ffi::_ND_INS_CLASS::ND_INS_LWPINS => Ok(Mnemonic::LWPINS),
            ffi::_ND_INS_CLASS::ND_INS_LWPVAL => Ok(Mnemonic::LWPVAL),
            ffi::_ND_INS_CLASS::ND_INS_LZCNT => Ok(Mnemonic::LZCNT),
            ffi::_ND_INS_CLASS::ND_INS_MASKMOVDQU => Ok(Mnemonic::MASKMOVDQU),
            ffi::_ND_INS_CLASS::ND_INS_MASKMOVQ => Ok(Mnemonic::MASKMOVQ),
            ffi::_ND_INS_CLASS::ND_INS_MAXPD => Ok(Mnemonic::MAXPD),
            ffi::_ND_INS_CLASS::ND_INS_MAXPS => Ok(Mnemonic::MAXPS),
            ffi::_ND_INS_CLASS::ND_INS_MAXSD => Ok(Mnemonic::MAXSD),
            ffi::_ND_INS_CLASS::ND_INS_MAXSS => Ok(Mnemonic::MAXSS),
            ffi::_ND_INS_CLASS::ND_INS_MCOMMIT => Ok(Mnemonic::MCOMMIT),
            ffi::_ND_INS_CLASS::ND_INS_MFENCE => Ok(Mnemonic::MFENCE),
            ffi::_ND_INS_CLASS::ND_INS_MINPD => Ok(Mnemonic::MINPD),
            ffi::_ND_INS_CLASS::ND_INS_MINPS => Ok(Mnemonic::MINPS),
            ffi::_ND_INS_CLASS::ND_INS_MINSD => Ok(Mnemonic::MINSD),
            ffi::_ND_INS_CLASS::ND_INS_MINSS => Ok(Mnemonic::MINSS),
            ffi::_ND_INS_CLASS::ND_INS_MONITOR => Ok(Mnemonic::MONITOR),
            ffi::_ND_INS_CLASS::ND_INS_MONITORX => Ok(Mnemonic::MONITORX),
            ffi::_ND_INS_CLASS::ND_INS_MOV => Ok(Mnemonic::MOV),
            ffi::_ND_INS_CLASS::ND_INS_MOVAPD => Ok(Mnemonic::MOVAPD),
            ffi::_ND_INS_CLASS::ND_INS_MOVAPS => Ok(Mnemonic::MOVAPS),
            ffi::_ND_INS_CLASS::ND_INS_MOVBE => Ok(Mnemonic::MOVBE),
            ffi::_ND_INS_CLASS::ND_INS_MOVD => Ok(Mnemonic::MOVD),
            ffi::_ND_INS_CLASS::ND_INS_MOVDDUP => Ok(Mnemonic::MOVDDUP),
            ffi::_ND_INS_CLASS::ND_INS_MOVDIR64B => Ok(Mnemonic::MOVDIR64B),
            ffi::_ND_INS_CLASS::ND_INS_MOVDIRI => Ok(Mnemonic::MOVDIRI),
            ffi::_ND_INS_CLASS::ND_INS_MOVDQ2Q => Ok(Mnemonic::MOVDQ2Q),
            ffi::_ND_INS_CLASS::ND_INS_MOVDQA => Ok(Mnemonic::MOVDQA),
            ffi::_ND_INS_CLASS::ND_INS_MOVDQU => Ok(Mnemonic::MOVDQU),
            ffi::_ND_INS_CLASS::ND_INS_MOVHLPS => Ok(Mnemonic::MOVHLPS),
            ffi::_ND_INS_CLASS::ND_INS_MOVHPD => Ok(Mnemonic::MOVHPD),
            ffi::_ND_INS_CLASS::ND_INS_MOVHPS => Ok(Mnemonic::MOVHPS),
            ffi::_ND_INS_CLASS::ND_INS_MOVLHPS => Ok(Mnemonic::MOVLHPS),
            ffi::_ND_INS_CLASS::ND_INS_MOVLPD => Ok(Mnemonic::MOVLPD),
            ffi::_ND_INS_CLASS::ND_INS_MOVLPS => Ok(Mnemonic::MOVLPS),
            ffi::_ND_INS_CLASS::ND_INS_MOVMSKPD => Ok(Mnemonic::MOVMSKPD),
            ffi::_ND_INS_CLASS::ND_INS_MOVMSKPS => Ok(Mnemonic::MOVMSKPS),
            ffi::_ND_INS_CLASS::ND_INS_MOVNTDQ => Ok(Mnemonic::MOVNTDQ),
            ffi::_ND_INS_CLASS::ND_INS_MOVNTDQA => Ok(Mnemonic::MOVNTDQA),
            ffi::_ND_INS_CLASS::ND_INS_MOVNTI => Ok(Mnemonic::MOVNTI),
            ffi::_ND_INS_CLASS::ND_INS_MOVNTPD => Ok(Mnemonic::MOVNTPD),
            ffi::_ND_INS_CLASS::ND_INS_MOVNTPS => Ok(Mnemonic::MOVNTPS),
            ffi::_ND_INS_CLASS::ND_INS_MOVNTQ => Ok(Mnemonic::MOVNTQ),
            ffi::_ND_INS_CLASS::ND_INS_MOVNTSD => Ok(Mnemonic::MOVNTSD),
            ffi::_ND_INS_CLASS::ND_INS_MOVNTSS => Ok(Mnemonic::MOVNTSS),
            ffi::_ND_INS_CLASS::ND_INS_MOVQ => Ok(Mnemonic::MOVQ),
            ffi::_ND_INS_CLASS::ND_INS_MOVQ2DQ => Ok(Mnemonic::MOVQ2DQ),
            ffi::_ND_INS_CLASS::ND_INS_MOVS => Ok(Mnemonic::MOVS),
            ffi::_ND_INS_CLASS::ND_INS_MOVSD => Ok(Mnemonic::MOVSD),
            ffi::_ND_INS_CLASS::ND_INS_MOVSHDUP => Ok(Mnemonic::MOVSHDUP),
            ffi::_ND_INS_CLASS::ND_INS_MOVSLDUP => Ok(Mnemonic::MOVSLDUP),
            ffi::_ND_INS_CLASS::ND_INS_MOVSS => Ok(Mnemonic::MOVSS),
            ffi::_ND_INS_CLASS::ND_INS_MOVSX => Ok(Mnemonic::MOVSX),
            ffi::_ND_INS_CLASS::ND_INS_MOVSXD => Ok(Mnemonic::MOVSXD),
            ffi::_ND_INS_CLASS::ND_INS_MOVUPD => Ok(Mnemonic::MOVUPD),
            ffi::_ND_INS_CLASS::ND_INS_MOVUPS => Ok(Mnemonic::MOVUPS),
            ffi::_ND_INS_CLASS::ND_INS_MOVZX => Ok(Mnemonic::MOVZX),
            ffi::_ND_INS_CLASS::ND_INS_MOV_CR => Ok(Mnemonic::MOV_CR),
            ffi::_ND_INS_CLASS::ND_INS_MOV_DR => Ok(Mnemonic::MOV_DR),
            ffi::_ND_INS_CLASS::ND_INS_MOV_TR => Ok(Mnemonic::MOV_TR),
            ffi::_ND_INS_CLASS::ND_INS_MPSADBW => Ok(Mnemonic::MPSADBW),
            ffi::_ND_INS_CLASS::ND_INS_MUL => Ok(Mnemonic::MUL),
            ffi::_ND_INS_CLASS::ND_INS_MULPD => Ok(Mnemonic::MULPD),
            ffi::_ND_INS_CLASS::ND_INS_MULPS => Ok(Mnemonic::MULPS),
            ffi::_ND_INS_CLASS::ND_INS_MULSD => Ok(Mnemonic::MULSD),
            ffi::_ND_INS_CLASS::ND_INS_MULSS => Ok(Mnemonic::MULSS),
            ffi::_ND_INS_CLASS::ND_INS_MULX => Ok(Mnemonic::MULX),
            ffi::_ND_INS_CLASS::ND_INS_MWAIT => Ok(Mnemonic::MWAIT),
            ffi::_ND_INS_CLASS::ND_INS_MWAITX => Ok(Mnemonic::MWAITX),
            ffi::_ND_INS_CLASS::ND_INS_NEG => Ok(Mnemonic::NEG),
            ffi::_ND_INS_CLASS::ND_INS_NOP => Ok(Mnemonic::NOP),
            ffi::_ND_INS_CLASS::ND_INS_NOT => Ok(Mnemonic::NOT),
            ffi::_ND_INS_CLASS::ND_INS_OR => Ok(Mnemonic::OR),
            ffi::_ND_INS_CLASS::ND_INS_ORPD => Ok(Mnemonic::ORPD),
            ffi::_ND_INS_CLASS::ND_INS_ORPS => Ok(Mnemonic::ORPS),
            ffi::_ND_INS_CLASS::ND_INS_OUT => Ok(Mnemonic::OUT),
            ffi::_ND_INS_CLASS::ND_INS_OUTS => Ok(Mnemonic::OUTS),
            ffi::_ND_INS_CLASS::ND_INS_PABSB => Ok(Mnemonic::PABSB),
            ffi::_ND_INS_CLASS::ND_INS_PABSD => Ok(Mnemonic::PABSD),
            ffi::_ND_INS_CLASS::ND_INS_PABSW => Ok(Mnemonic::PABSW),
            ffi::_ND_INS_CLASS::ND_INS_PACKSSDW => Ok(Mnemonic::PACKSSDW),
            ffi::_ND_INS_CLASS::ND_INS_PACKSSWB => Ok(Mnemonic::PACKSSWB),
            ffi::_ND_INS_CLASS::ND_INS_PACKUSDW => Ok(Mnemonic::PACKUSDW),
            ffi::_ND_INS_CLASS::ND_INS_PACKUSWB => Ok(Mnemonic::PACKUSWB),
            ffi::_ND_INS_CLASS::ND_INS_PADDB => Ok(Mnemonic::PADDB),
            ffi::_ND_INS_CLASS::ND_INS_PADDD => Ok(Mnemonic::PADDD),
            ffi::_ND_INS_CLASS::ND_INS_PADDQ => Ok(Mnemonic::PADDQ),
            ffi::_ND_INS_CLASS::ND_INS_PADDSB => Ok(Mnemonic::PADDSB),
            ffi::_ND_INS_CLASS::ND_INS_PADDSW => Ok(Mnemonic::PADDSW),
            ffi::_ND_INS_CLASS::ND_INS_PADDUSB => Ok(Mnemonic::PADDUSB),
            ffi::_ND_INS_CLASS::ND_INS_PADDUSW => Ok(Mnemonic::PADDUSW),
            ffi::_ND_INS_CLASS::ND_INS_PADDW => Ok(Mnemonic::PADDW),
            ffi::_ND_INS_CLASS::ND_INS_PALIGNR => Ok(Mnemonic::PALIGNR),
            ffi::_ND_INS_CLASS::ND_INS_PAND => Ok(Mnemonic::PAND),
            ffi::_ND_INS_CLASS::ND_INS_PANDN => Ok(Mnemonic::PANDN),
            ffi::_ND_INS_CLASS::ND_INS_PAUSE => Ok(Mnemonic::PAUSE),
            ffi::_ND_INS_CLASS::ND_INS_PAVGB => Ok(Mnemonic::PAVGB),
            ffi::_ND_INS_CLASS::ND_INS_PAVGUSB => Ok(Mnemonic::PAVGUSB),
            ffi::_ND_INS_CLASS::ND_INS_PAVGW => Ok(Mnemonic::PAVGW),
            ffi::_ND_INS_CLASS::ND_INS_PBLENDVB => Ok(Mnemonic::PBLENDVB),
            ffi::_ND_INS_CLASS::ND_INS_PBLENDW => Ok(Mnemonic::PBLENDW),
            ffi::_ND_INS_CLASS::ND_INS_PBNDKB => Ok(Mnemonic::PBNDKB),
            ffi::_ND_INS_CLASS::ND_INS_PCLMULQDQ => Ok(Mnemonic::PCLMULQDQ),
            ffi::_ND_INS_CLASS::ND_INS_PCMPEQB => Ok(Mnemonic::PCMPEQB),
            ffi::_ND_INS_CLASS::ND_INS_PCMPEQD => Ok(Mnemonic::PCMPEQD),
            ffi::_ND_INS_CLASS::ND_INS_PCMPEQQ => Ok(Mnemonic::PCMPEQQ),
            ffi::_ND_INS_CLASS::ND_INS_PCMPEQW => Ok(Mnemonic::PCMPEQW),
            ffi::_ND_INS_CLASS::ND_INS_PCMPESTRI => Ok(Mnemonic::PCMPESTRI),
            ffi::_ND_INS_CLASS::ND_INS_PCMPESTRM => Ok(Mnemonic::PCMPESTRM),
            ffi::_ND_INS_CLASS::ND_INS_PCMPGTB => Ok(Mnemonic::PCMPGTB),
            ffi::_ND_INS_CLASS::ND_INS_PCMPGTD => Ok(Mnemonic::PCMPGTD),
            ffi::_ND_INS_CLASS::ND_INS_PCMPGTQ => Ok(Mnemonic::PCMPGTQ),
            ffi::_ND_INS_CLASS::ND_INS_PCMPGTW => Ok(Mnemonic::PCMPGTW),
            ffi::_ND_INS_CLASS::ND_INS_PCMPISTRI => Ok(Mnemonic::PCMPISTRI),
            ffi::_ND_INS_CLASS::ND_INS_PCMPISTRM => Ok(Mnemonic::PCMPISTRM),
            ffi::_ND_INS_CLASS::ND_INS_PCONFIG => Ok(Mnemonic::PCONFIG),
            ffi::_ND_INS_CLASS::ND_INS_PDEP => Ok(Mnemonic::PDEP),
            ffi::_ND_INS_CLASS::ND_INS_PEXT => Ok(Mnemonic::PEXT),
            ffi::_ND_INS_CLASS::ND_INS_PEXTRB => Ok(Mnemonic::PEXTRB),
            ffi::_ND_INS_CLASS::ND_INS_PEXTRD => Ok(Mnemonic::PEXTRD),
            ffi::_ND_INS_CLASS::ND_INS_PEXTRQ => Ok(Mnemonic::PEXTRQ),
            ffi::_ND_INS_CLASS::ND_INS_PEXTRW => Ok(Mnemonic::PEXTRW),
            ffi::_ND_INS_CLASS::ND_INS_PF2ID => Ok(Mnemonic::PF2ID),
            ffi::_ND_INS_CLASS::ND_INS_PF2IW => Ok(Mnemonic::PF2IW),
            ffi::_ND_INS_CLASS::ND_INS_PFACC => Ok(Mnemonic::PFACC),
            ffi::_ND_INS_CLASS::ND_INS_PFADD => Ok(Mnemonic::PFADD),
            ffi::_ND_INS_CLASS::ND_INS_PFCMPEQ => Ok(Mnemonic::PFCMPEQ),
            ffi::_ND_INS_CLASS::ND_INS_PFCMPGE => Ok(Mnemonic::PFCMPGE),
            ffi::_ND_INS_CLASS::ND_INS_PFCMPGT => Ok(Mnemonic::PFCMPGT),
            ffi::_ND_INS_CLASS::ND_INS_PFMAX => Ok(Mnemonic::PFMAX),
            ffi::_ND_INS_CLASS::ND_INS_PFMIN => Ok(Mnemonic::PFMIN),
            ffi::_ND_INS_CLASS::ND_INS_PFMUL => Ok(Mnemonic::PFMUL),
            ffi::_ND_INS_CLASS::ND_INS_PFNACC => Ok(Mnemonic::PFNACC),
            ffi::_ND_INS_CLASS::ND_INS_PFPNACC => Ok(Mnemonic::PFPNACC),
            ffi::_ND_INS_CLASS::ND_INS_PFRCP => Ok(Mnemonic::PFRCP),
            ffi::_ND_INS_CLASS::ND_INS_PFRCPIT1 => Ok(Mnemonic::PFRCPIT1),
            ffi::_ND_INS_CLASS::ND_INS_PFRCPIT2 => Ok(Mnemonic::PFRCPIT2),
            ffi::_ND_INS_CLASS::ND_INS_PFRCPV => Ok(Mnemonic::PFRCPV),
            ffi::_ND_INS_CLASS::ND_INS_PFRSQIT1 => Ok(Mnemonic::PFRSQIT1),
            ffi::_ND_INS_CLASS::ND_INS_PFRSQRT => Ok(Mnemonic::PFRSQRT),
            ffi::_ND_INS_CLASS::ND_INS_PFRSQRTV => Ok(Mnemonic::PFRSQRTV),
            ffi::_ND_INS_CLASS::ND_INS_PFSUB => Ok(Mnemonic::PFSUB),
            ffi::_ND_INS_CLASS::ND_INS_PFSUBR => Ok(Mnemonic::PFSUBR),
            ffi::_ND_INS_CLASS::ND_INS_PHADDD => Ok(Mnemonic::PHADDD),
            ffi::_ND_INS_CLASS::ND_INS_PHADDSW => Ok(Mnemonic::PHADDSW),
            ffi::_ND_INS_CLASS::ND_INS_PHADDW => Ok(Mnemonic::PHADDW),
            ffi::_ND_INS_CLASS::ND_INS_PHMINPOSUW => Ok(Mnemonic::PHMINPOSUW),
            ffi::_ND_INS_CLASS::ND_INS_PHSUBD => Ok(Mnemonic::PHSUBD),
            ffi::_ND_INS_CLASS::ND_INS_PHSUBSW => Ok(Mnemonic::PHSUBSW),
            ffi::_ND_INS_CLASS::ND_INS_PHSUBW => Ok(Mnemonic::PHSUBW),
            ffi::_ND_INS_CLASS::ND_INS_PI2FD => Ok(Mnemonic::PI2FD),
            ffi::_ND_INS_CLASS::ND_INS_PI2FW => Ok(Mnemonic::PI2FW),
            ffi::_ND_INS_CLASS::ND_INS_PINSRB => Ok(Mnemonic::PINSRB),
            ffi::_ND_INS_CLASS::ND_INS_PINSRD => Ok(Mnemonic::PINSRD),
            ffi::_ND_INS_CLASS::ND_INS_PINSRQ => Ok(Mnemonic::PINSRQ),
            ffi::_ND_INS_CLASS::ND_INS_PINSRW => Ok(Mnemonic::PINSRW),
            ffi::_ND_INS_CLASS::ND_INS_PMADDUBSW => Ok(Mnemonic::PMADDUBSW),
            ffi::_ND_INS_CLASS::ND_INS_PMADDWD => Ok(Mnemonic::PMADDWD),
            ffi::_ND_INS_CLASS::ND_INS_PMAXSB => Ok(Mnemonic::PMAXSB),
            ffi::_ND_INS_CLASS::ND_INS_PMAXSD => Ok(Mnemonic::PMAXSD),
            ffi::_ND_INS_CLASS::ND_INS_PMAXSW => Ok(Mnemonic::PMAXSW),
            ffi::_ND_INS_CLASS::ND_INS_PMAXUB => Ok(Mnemonic::PMAXUB),
            ffi::_ND_INS_CLASS::ND_INS_PMAXUD => Ok(Mnemonic::PMAXUD),
            ffi::_ND_INS_CLASS::ND_INS_PMAXUW => Ok(Mnemonic::PMAXUW),
            ffi::_ND_INS_CLASS::ND_INS_PMINSB => Ok(Mnemonic::PMINSB),
            ffi::_ND_INS_CLASS::ND_INS_PMINSD => Ok(Mnemonic::PMINSD),
            ffi::_ND_INS_CLASS::ND_INS_PMINSW => Ok(Mnemonic::PMINSW),
            ffi::_ND_INS_CLASS::ND_INS_PMINUB => Ok(Mnemonic::PMINUB),
            ffi::_ND_INS_CLASS::ND_INS_PMINUD => Ok(Mnemonic::PMINUD),
            ffi::_ND_INS_CLASS::ND_INS_PMINUW => Ok(Mnemonic::PMINUW),
            ffi::_ND_INS_CLASS::ND_INS_PMOVMSKB => Ok(Mnemonic::PMOVMSKB),
            ffi::_ND_INS_CLASS::ND_INS_PMOVSXBD => Ok(Mnemonic::PMOVSXBD),
            ffi::_ND_INS_CLASS::ND_INS_PMOVSXBQ => Ok(Mnemonic::PMOVSXBQ),
            ffi::_ND_INS_CLASS::ND_INS_PMOVSXBW => Ok(Mnemonic::PMOVSXBW),
            ffi::_ND_INS_CLASS::ND_INS_PMOVSXDQ => Ok(Mnemonic::PMOVSXDQ),
            ffi::_ND_INS_CLASS::ND_INS_PMOVSXWD => Ok(Mnemonic::PMOVSXWD),
            ffi::_ND_INS_CLASS::ND_INS_PMOVSXWQ => Ok(Mnemonic::PMOVSXWQ),
            ffi::_ND_INS_CLASS::ND_INS_PMOVZXBD => Ok(Mnemonic::PMOVZXBD),
            ffi::_ND_INS_CLASS::ND_INS_PMOVZXBQ => Ok(Mnemonic::PMOVZXBQ),
            ffi::_ND_INS_CLASS::ND_INS_PMOVZXBW => Ok(Mnemonic::PMOVZXBW),
            ffi::_ND_INS_CLASS::ND_INS_PMOVZXDQ => Ok(Mnemonic::PMOVZXDQ),
            ffi::_ND_INS_CLASS::ND_INS_PMOVZXWD => Ok(Mnemonic::PMOVZXWD),
            ffi::_ND_INS_CLASS::ND_INS_PMOVZXWQ => Ok(Mnemonic::PMOVZXWQ),
            ffi::_ND_INS_CLASS::ND_INS_PMULDQ => Ok(Mnemonic::PMULDQ),
            ffi::_ND_INS_CLASS::ND_INS_PMULHRSW => Ok(Mnemonic::PMULHRSW),
            ffi::_ND_INS_CLASS::ND_INS_PMULHRW => Ok(Mnemonic::PMULHRW),
            ffi::_ND_INS_CLASS::ND_INS_PMULHUW => Ok(Mnemonic::PMULHUW),
            ffi::_ND_INS_CLASS::ND_INS_PMULHW => Ok(Mnemonic::PMULHW),
            ffi::_ND_INS_CLASS::ND_INS_PMULLD => Ok(Mnemonic::PMULLD),
            ffi::_ND_INS_CLASS::ND_INS_PMULLW => Ok(Mnemonic::PMULLW),
            ffi::_ND_INS_CLASS::ND_INS_PMULUDQ => Ok(Mnemonic::PMULUDQ),
            ffi::_ND_INS_CLASS::ND_INS_POP => Ok(Mnemonic::POP),
            ffi::_ND_INS_CLASS::ND_INS_POP2 => Ok(Mnemonic::POP2),
            ffi::_ND_INS_CLASS::ND_INS_POP2P => Ok(Mnemonic::POP2P),
            ffi::_ND_INS_CLASS::ND_INS_POPA => Ok(Mnemonic::POPA),
            ffi::_ND_INS_CLASS::ND_INS_POPAD => Ok(Mnemonic::POPAD),
            ffi::_ND_INS_CLASS::ND_INS_POPCNT => Ok(Mnemonic::POPCNT),
            ffi::_ND_INS_CLASS::ND_INS_POPF => Ok(Mnemonic::POPF),
            ffi::_ND_INS_CLASS::ND_INS_POPP => Ok(Mnemonic::POPP),
            ffi::_ND_INS_CLASS::ND_INS_POR => Ok(Mnemonic::POR),
            ffi::_ND_INS_CLASS::ND_INS_PREFETCH => Ok(Mnemonic::PREFETCH),
            ffi::_ND_INS_CLASS::ND_INS_PREFETCHE => Ok(Mnemonic::PREFETCHE),
            ffi::_ND_INS_CLASS::ND_INS_PREFETCHIT0 => Ok(Mnemonic::PREFETCHIT0),
            ffi::_ND_INS_CLASS::ND_INS_PREFETCHIT1 => Ok(Mnemonic::PREFETCHIT1),
            ffi::_ND_INS_CLASS::ND_INS_PREFETCHM => Ok(Mnemonic::PREFETCHM),
            ffi::_ND_INS_CLASS::ND_INS_PREFETCHNTA => Ok(Mnemonic::PREFETCHNTA),
            ffi::_ND_INS_CLASS::ND_INS_PREFETCHT0 => Ok(Mnemonic::PREFETCHT0),
            ffi::_ND_INS_CLASS::ND_INS_PREFETCHT1 => Ok(Mnemonic::PREFETCHT1),
            ffi::_ND_INS_CLASS::ND_INS_PREFETCHT2 => Ok(Mnemonic::PREFETCHT2),
            ffi::_ND_INS_CLASS::ND_INS_PREFETCHW => Ok(Mnemonic::PREFETCHW),
            ffi::_ND_INS_CLASS::ND_INS_PREFETCHWT1 => Ok(Mnemonic::PREFETCHWT1),
            ffi::_ND_INS_CLASS::ND_INS_PSADBW => Ok(Mnemonic::PSADBW),
            ffi::_ND_INS_CLASS::ND_INS_PSHUFB => Ok(Mnemonic::PSHUFB),
            ffi::_ND_INS_CLASS::ND_INS_PSHUFD => Ok(Mnemonic::PSHUFD),
            ffi::_ND_INS_CLASS::ND_INS_PSHUFHW => Ok(Mnemonic::PSHUFHW),
            ffi::_ND_INS_CLASS::ND_INS_PSHUFLW => Ok(Mnemonic::PSHUFLW),
            ffi::_ND_INS_CLASS::ND_INS_PSHUFW => Ok(Mnemonic::PSHUFW),
            ffi::_ND_INS_CLASS::ND_INS_PSIGNB => Ok(Mnemonic::PSIGNB),
            ffi::_ND_INS_CLASS::ND_INS_PSIGND => Ok(Mnemonic::PSIGND),
            ffi::_ND_INS_CLASS::ND_INS_PSIGNW => Ok(Mnemonic::PSIGNW),
            ffi::_ND_INS_CLASS::ND_INS_PSLLD => Ok(Mnemonic::PSLLD),
            ffi::_ND_INS_CLASS::ND_INS_PSLLDQ => Ok(Mnemonic::PSLLDQ),
            ffi::_ND_INS_CLASS::ND_INS_PSLLQ => Ok(Mnemonic::PSLLQ),
            ffi::_ND_INS_CLASS::ND_INS_PSLLW => Ok(Mnemonic::PSLLW),
            ffi::_ND_INS_CLASS::ND_INS_PSMASH => Ok(Mnemonic::PSMASH),
            ffi::_ND_INS_CLASS::ND_INS_PSRAD => Ok(Mnemonic::PSRAD),
            ffi::_ND_INS_CLASS::ND_INS_PSRAW => Ok(Mnemonic::PSRAW),
            ffi::_ND_INS_CLASS::ND_INS_PSRLD => Ok(Mnemonic::PSRLD),
            ffi::_ND_INS_CLASS::ND_INS_PSRLDQ => Ok(Mnemonic::PSRLDQ),
            ffi::_ND_INS_CLASS::ND_INS_PSRLQ => Ok(Mnemonic::PSRLQ),
            ffi::_ND_INS_CLASS::ND_INS_PSRLW => Ok(Mnemonic::PSRLW),
            ffi::_ND_INS_CLASS::ND_INS_PSUBB => Ok(Mnemonic::PSUBB),
            ffi::_ND_INS_CLASS::ND_INS_PSUBD => Ok(Mnemonic::PSUBD),
            ffi::_ND_INS_CLASS::ND_INS_PSUBQ => Ok(Mnemonic::PSUBQ),
            ffi::_ND_INS_CLASS::ND_INS_PSUBSB => Ok(Mnemonic::PSUBSB),
            ffi::_ND_INS_CLASS::ND_INS_PSUBSW => Ok(Mnemonic::PSUBSW),
            ffi::_ND_INS_CLASS::ND_INS_PSUBUSB => Ok(Mnemonic::PSUBUSB),
            ffi::_ND_INS_CLASS::ND_INS_PSUBUSW => Ok(Mnemonic::PSUBUSW),
            ffi::_ND_INS_CLASS::ND_INS_PSUBW => Ok(Mnemonic::PSUBW),
            ffi::_ND_INS_CLASS::ND_INS_PSWAPD => Ok(Mnemonic::PSWAPD),
            ffi::_ND_INS_CLASS::ND_INS_PTEST => Ok(Mnemonic::PTEST),
            ffi::_ND_INS_CLASS::ND_INS_PTWRITE => Ok(Mnemonic::PTWRITE),
            ffi::_ND_INS_CLASS::ND_INS_PUNPCKHBW => Ok(Mnemonic::PUNPCKHBW),
            ffi::_ND_INS_CLASS::ND_INS_PUNPCKHDQ => Ok(Mnemonic::PUNPCKHDQ),
            ffi::_ND_INS_CLASS::ND_INS_PUNPCKHQDQ => Ok(Mnemonic::PUNPCKHQDQ),
            ffi::_ND_INS_CLASS::ND_INS_PUNPCKHWD => Ok(Mnemonic::PUNPCKHWD),
            ffi::_ND_INS_CLASS::ND_INS_PUNPCKLBW => Ok(Mnemonic::PUNPCKLBW),
            ffi::_ND_INS_CLASS::ND_INS_PUNPCKLDQ => Ok(Mnemonic::PUNPCKLDQ),
            ffi::_ND_INS_CLASS::ND_INS_PUNPCKLQDQ => Ok(Mnemonic::PUNPCKLQDQ),
            ffi::_ND_INS_CLASS::ND_INS_PUNPCKLWD => Ok(Mnemonic::PUNPCKLWD),
            ffi::_ND_INS_CLASS::ND_INS_PUSH => Ok(Mnemonic::PUSH),
            ffi::_ND_INS_CLASS::ND_INS_PUSH2 => Ok(Mnemonic::PUSH2),
            ffi::_ND_INS_CLASS::ND_INS_PUSH2P => Ok(Mnemonic::PUSH2P),
            ffi::_ND_INS_CLASS::ND_INS_PUSHA => Ok(Mnemonic::PUSHA),
            ffi::_ND_INS_CLASS::ND_INS_PUSHAD => Ok(Mnemonic::PUSHAD),
            ffi::_ND_INS_CLASS::ND_INS_PUSHF => Ok(Mnemonic::PUSHF),
            ffi::_ND_INS_CLASS::ND_INS_PUSHP => Ok(Mnemonic::PUSHP),
            ffi::_ND_INS_CLASS::ND_INS_PVALIDATE => Ok(Mnemonic::PVALIDATE),
            ffi::_ND_INS_CLASS::ND_INS_PXOR => Ok(Mnemonic::PXOR),
            ffi::_ND_INS_CLASS::ND_INS_RCL => Ok(Mnemonic::RCL),
            ffi::_ND_INS_CLASS::ND_INS_RCPPS => Ok(Mnemonic::RCPPS),
            ffi::_ND_INS_CLASS::ND_INS_RCPSS => Ok(Mnemonic::RCPSS),
            ffi::_ND_INS_CLASS::ND_INS_RCR => Ok(Mnemonic::RCR),
            ffi::_ND_INS_CLASS::ND_INS_RDFSBASE => Ok(Mnemonic::RDFSBASE),
            ffi::_ND_INS_CLASS::ND_INS_RDGSBASE => Ok(Mnemonic::RDGSBASE),
            ffi::_ND_INS_CLASS::ND_INS_RDMSR => Ok(Mnemonic::RDMSR),
            ffi::_ND_INS_CLASS::ND_INS_RDMSRLIST => Ok(Mnemonic::RDMSRLIST),
            ffi::_ND_INS_CLASS::ND_INS_RDPID => Ok(Mnemonic::RDPID),
            ffi::_ND_INS_CLASS::ND_INS_RDPKRU => Ok(Mnemonic::RDPKRU),
            ffi::_ND_INS_CLASS::ND_INS_RDPMC => Ok(Mnemonic::RDPMC),
            ffi::_ND_INS_CLASS::ND_INS_RDPRU => Ok(Mnemonic::RDPRU),
            ffi::_ND_INS_CLASS::ND_INS_RDRAND => Ok(Mnemonic::RDRAND),
            ffi::_ND_INS_CLASS::ND_INS_RDSEED => Ok(Mnemonic::RDSEED),
            ffi::_ND_INS_CLASS::ND_INS_RDTSC => Ok(Mnemonic::RDTSC),
            ffi::_ND_INS_CLASS::ND_INS_RDTSCP => Ok(Mnemonic::RDTSCP),
            ffi::_ND_INS_CLASS::ND_INS_RETF => Ok(Mnemonic::RETF),
            ffi::_ND_INS_CLASS::ND_INS_RETN => Ok(Mnemonic::RETN),
            ffi::_ND_INS_CLASS::ND_INS_RMPADJUST => Ok(Mnemonic::RMPADJUST),
            ffi::_ND_INS_CLASS::ND_INS_RMPQUERY => Ok(Mnemonic::RMPQUERY),
            ffi::_ND_INS_CLASS::ND_INS_RMPREAD => Ok(Mnemonic::RMPREAD),
            ffi::_ND_INS_CLASS::ND_INS_RMPUPDATE => Ok(Mnemonic::RMPUPDATE),
            ffi::_ND_INS_CLASS::ND_INS_ROL => Ok(Mnemonic::ROL),
            ffi::_ND_INS_CLASS::ND_INS_ROR => Ok(Mnemonic::ROR),
            ffi::_ND_INS_CLASS::ND_INS_RORX => Ok(Mnemonic::RORX),
            ffi::_ND_INS_CLASS::ND_INS_ROUNDPD => Ok(Mnemonic::ROUNDPD),
            ffi::_ND_INS_CLASS::ND_INS_ROUNDPS => Ok(Mnemonic::ROUNDPS),
            ffi::_ND_INS_CLASS::ND_INS_ROUNDSD => Ok(Mnemonic::ROUNDSD),
            ffi::_ND_INS_CLASS::ND_INS_ROUNDSS => Ok(Mnemonic::ROUNDSS),
            ffi::_ND_INS_CLASS::ND_INS_RSM => Ok(Mnemonic::RSM),
            ffi::_ND_INS_CLASS::ND_INS_RSQRTPS => Ok(Mnemonic::RSQRTPS),
            ffi::_ND_INS_CLASS::ND_INS_RSQRTSS => Ok(Mnemonic::RSQRTSS),
            ffi::_ND_INS_CLASS::ND_INS_RSSSP => Ok(Mnemonic::RSSSP),
            ffi::_ND_INS_CLASS::ND_INS_RSTORSSP => Ok(Mnemonic::RSTORSSP),
            ffi::_ND_INS_CLASS::ND_INS_SAHF => Ok(Mnemonic::SAHF),
            ffi::_ND_INS_CLASS::ND_INS_SAL => Ok(Mnemonic::SAL),
            ffi::_ND_INS_CLASS::ND_INS_SALC => Ok(Mnemonic::SALC),
            ffi::_ND_INS_CLASS::ND_INS_SAR => Ok(Mnemonic::SAR),
            ffi::_ND_INS_CLASS::ND_INS_SARX => Ok(Mnemonic::SARX),
            ffi::_ND_INS_CLASS::ND_INS_SAVEPREVSSP => Ok(Mnemonic::SAVEPREVSSP),
            ffi::_ND_INS_CLASS::ND_INS_SBB => Ok(Mnemonic::SBB),
            ffi::_ND_INS_CLASS::ND_INS_SCAS => Ok(Mnemonic::SCAS),
            ffi::_ND_INS_CLASS::ND_INS_SEAMCALL => Ok(Mnemonic::SEAMCALL),
            ffi::_ND_INS_CLASS::ND_INS_SEAMOPS => Ok(Mnemonic::SEAMOPS),
            ffi::_ND_INS_CLASS::ND_INS_SEAMRET => Ok(Mnemonic::SEAMRET),
            ffi::_ND_INS_CLASS::ND_INS_SENDUIPI => Ok(Mnemonic::SENDUIPI),
            ffi::_ND_INS_CLASS::ND_INS_SERIALIZE => Ok(Mnemonic::SERIALIZE),
            ffi::_ND_INS_CLASS::ND_INS_SETSSBSY => Ok(Mnemonic::SETSSBSY),
            ffi::_ND_INS_CLASS::ND_INS_SETcc => Ok(Mnemonic::SETcc),
            ffi::_ND_INS_CLASS::ND_INS_SFENCE => Ok(Mnemonic::SFENCE),
            ffi::_ND_INS_CLASS::ND_INS_SGDT => Ok(Mnemonic::SGDT),
            ffi::_ND_INS_CLASS::ND_INS_SHA1MSG1 => Ok(Mnemonic::SHA1MSG1),
            ffi::_ND_INS_CLASS::ND_INS_SHA1MSG2 => Ok(Mnemonic::SHA1MSG2),
            ffi::_ND_INS_CLASS::ND_INS_SHA1NEXTE => Ok(Mnemonic::SHA1NEXTE),
            ffi::_ND_INS_CLASS::ND_INS_SHA1RNDS4 => Ok(Mnemonic::SHA1RNDS4),
            ffi::_ND_INS_CLASS::ND_INS_SHA256MSG1 => Ok(Mnemonic::SHA256MSG1),
            ffi::_ND_INS_CLASS::ND_INS_SHA256MSG2 => Ok(Mnemonic::SHA256MSG2),
            ffi::_ND_INS_CLASS::ND_INS_SHA256RNDS2 => Ok(Mnemonic::SHA256RNDS2),
            ffi::_ND_INS_CLASS::ND_INS_SHL => Ok(Mnemonic::SHL),
            ffi::_ND_INS_CLASS::ND_INS_SHLD => Ok(Mnemonic::SHLD),
            ffi::_ND_INS_CLASS::ND_INS_SHLX => Ok(Mnemonic::SHLX),
            ffi::_ND_INS_CLASS::ND_INS_SHR => Ok(Mnemonic::SHR),
            ffi::_ND_INS_CLASS::ND_INS_SHRD => Ok(Mnemonic::SHRD),
            ffi::_ND_INS_CLASS::ND_INS_SHRX => Ok(Mnemonic::SHRX),
            ffi::_ND_INS_CLASS::ND_INS_SHUFPD => Ok(Mnemonic::SHUFPD),
            ffi::_ND_INS_CLASS::ND_INS_SHUFPS => Ok(Mnemonic::SHUFPS),
            ffi::_ND_INS_CLASS::ND_INS_SIDT => Ok(Mnemonic::SIDT),
            ffi::_ND_INS_CLASS::ND_INS_SKINIT => Ok(Mnemonic::SKINIT),
            ffi::_ND_INS_CLASS::ND_INS_SLDT => Ok(Mnemonic::SLDT),
            ffi::_ND_INS_CLASS::ND_INS_SLWPCB => Ok(Mnemonic::SLWPCB),
            ffi::_ND_INS_CLASS::ND_INS_SMSW => Ok(Mnemonic::SMSW),
            ffi::_ND_INS_CLASS::ND_INS_SPFLT => Ok(Mnemonic::SPFLT),
            ffi::_ND_INS_CLASS::ND_INS_SQRTPD => Ok(Mnemonic::SQRTPD),
            ffi::_ND_INS_CLASS::ND_INS_SQRTPS => Ok(Mnemonic::SQRTPS),
            ffi::_ND_INS_CLASS::ND_INS_SQRTSD => Ok(Mnemonic::SQRTSD),
            ffi::_ND_INS_CLASS::ND_INS_SQRTSS => Ok(Mnemonic::SQRTSS),
            ffi::_ND_INS_CLASS::ND_INS_STAC => Ok(Mnemonic::STAC),
            ffi::_ND_INS_CLASS::ND_INS_STC => Ok(Mnemonic::STC),
            ffi::_ND_INS_CLASS::ND_INS_STD => Ok(Mnemonic::STD),
            ffi::_ND_INS_CLASS::ND_INS_STGI => Ok(Mnemonic::STGI),
            ffi::_ND_INS_CLASS::ND_INS_STI => Ok(Mnemonic::STI),
            ffi::_ND_INS_CLASS::ND_INS_STMXCSR => Ok(Mnemonic::STMXCSR),
            ffi::_ND_INS_CLASS::ND_INS_STOS => Ok(Mnemonic::STOS),
            ffi::_ND_INS_CLASS::ND_INS_STR => Ok(Mnemonic::STR),
            ffi::_ND_INS_CLASS::ND_INS_STTILECFG => Ok(Mnemonic::STTILECFG),
            ffi::_ND_INS_CLASS::ND_INS_STUI => Ok(Mnemonic::STUI),
            ffi::_ND_INS_CLASS::ND_INS_SUB => Ok(Mnemonic::SUB),
            ffi::_ND_INS_CLASS::ND_INS_SUBPD => Ok(Mnemonic::SUBPD),
            ffi::_ND_INS_CLASS::ND_INS_SUBPS => Ok(Mnemonic::SUBPS),
            ffi::_ND_INS_CLASS::ND_INS_SUBSD => Ok(Mnemonic::SUBSD),
            ffi::_ND_INS_CLASS::ND_INS_SUBSS => Ok(Mnemonic::SUBSS),
            ffi::_ND_INS_CLASS::ND_INS_SWAPGS => Ok(Mnemonic::SWAPGS),
            ffi::_ND_INS_CLASS::ND_INS_SYSCALL => Ok(Mnemonic::SYSCALL),
            ffi::_ND_INS_CLASS::ND_INS_SYSENTER => Ok(Mnemonic::SYSENTER),
            ffi::_ND_INS_CLASS::ND_INS_SYSEXIT => Ok(Mnemonic::SYSEXIT),
            ffi::_ND_INS_CLASS::ND_INS_SYSRET => Ok(Mnemonic::SYSRET),
            ffi::_ND_INS_CLASS::ND_INS_T1MSKC => Ok(Mnemonic::T1MSKC),
            ffi::_ND_INS_CLASS::ND_INS_TCMMIMFP16PS => Ok(Mnemonic::TCMMIMFP16PS),
            ffi::_ND_INS_CLASS::ND_INS_TCMMRLFP16PS => Ok(Mnemonic::TCMMRLFP16PS),
            ffi::_ND_INS_CLASS::ND_INS_TDCALL => Ok(Mnemonic::TDCALL),
            ffi::_ND_INS_CLASS::ND_INS_TDPBF16PS => Ok(Mnemonic::TDPBF16PS),
            ffi::_ND_INS_CLASS::ND_INS_TDPBSSD => Ok(Mnemonic::TDPBSSD),
            ffi::_ND_INS_CLASS::ND_INS_TDPBSUD => Ok(Mnemonic::TDPBSUD),
            ffi::_ND_INS_CLASS::ND_INS_TDPBUSD => Ok(Mnemonic::TDPBUSD),
            ffi::_ND_INS_CLASS::ND_INS_TDPBUUD => Ok(Mnemonic::TDPBUUD),
            ffi::_ND_INS_CLASS::ND_INS_TDPFP16PS => Ok(Mnemonic::TDPFP16PS),
            ffi::_ND_INS_CLASS::ND_INS_TEST => Ok(Mnemonic::TEST),
            ffi::_ND_INS_CLASS::ND_INS_TESTUI => Ok(Mnemonic::TESTUI),
            ffi::_ND_INS_CLASS::ND_INS_TILELOADD => Ok(Mnemonic::TILELOADD),
            ffi::_ND_INS_CLASS::ND_INS_TILELOADDT1 => Ok(Mnemonic::TILELOADDT1),
            ffi::_ND_INS_CLASS::ND_INS_TILERELEASE => Ok(Mnemonic::TILERELEASE),
            ffi::_ND_INS_CLASS::ND_INS_TILESTORED => Ok(Mnemonic::TILESTORED),
            ffi::_ND_INS_CLASS::ND_INS_TILEZERO => Ok(Mnemonic::TILEZERO),
            ffi::_ND_INS_CLASS::ND_INS_TLBSYNC => Ok(Mnemonic::TLBSYNC),
            ffi::_ND_INS_CLASS::ND_INS_TPAUSE => Ok(Mnemonic::TPAUSE),
            ffi::_ND_INS_CLASS::ND_INS_TZCNT => Ok(Mnemonic::TZCNT),
            ffi::_ND_INS_CLASS::ND_INS_TZMSK => Ok(Mnemonic::TZMSK),
            ffi::_ND_INS_CLASS::ND_INS_UCOMISD => Ok(Mnemonic::UCOMISD),
            ffi::_ND_INS_CLASS::ND_INS_UCOMISS => Ok(Mnemonic::UCOMISS),
            ffi::_ND_INS_CLASS::ND_INS_UD0 => Ok(Mnemonic::UD0),
            ffi::_ND_INS_CLASS::ND_INS_UD1 => Ok(Mnemonic::UD1),
            ffi::_ND_INS_CLASS::ND_INS_UD2 => Ok(Mnemonic::UD2),
            ffi::_ND_INS_CLASS::ND_INS_UIRET => Ok(Mnemonic::UIRET),
            ffi::_ND_INS_CLASS::ND_INS_UMONITOR => Ok(Mnemonic::UMONITOR),
            ffi::_ND_INS_CLASS::ND_INS_UMWAIT => Ok(Mnemonic::UMWAIT),
            ffi::_ND_INS_CLASS::ND_INS_UNPCKHPD => Ok(Mnemonic::UNPCKHPD),
            ffi::_ND_INS_CLASS::ND_INS_UNPCKHPS => Ok(Mnemonic::UNPCKHPS),
            ffi::_ND_INS_CLASS::ND_INS_UNPCKLPD => Ok(Mnemonic::UNPCKLPD),
            ffi::_ND_INS_CLASS::ND_INS_UNPCKLPS => Ok(Mnemonic::UNPCKLPS),
            ffi::_ND_INS_CLASS::ND_INS_URDMSR => Ok(Mnemonic::URDMSR),
            ffi::_ND_INS_CLASS::ND_INS_UWRMSR => Ok(Mnemonic::UWRMSR),
            ffi::_ND_INS_CLASS::ND_INS_V4FMADDPS => Ok(Mnemonic::V4FMADDPS),
            ffi::_ND_INS_CLASS::ND_INS_V4FMADDSS => Ok(Mnemonic::V4FMADDSS),
            ffi::_ND_INS_CLASS::ND_INS_V4FNMADDPS => Ok(Mnemonic::V4FNMADDPS),
            ffi::_ND_INS_CLASS::ND_INS_V4FNMADDSS => Ok(Mnemonic::V4FNMADDSS),
            ffi::_ND_INS_CLASS::ND_INS_VADDNEPBF16 => Ok(Mnemonic::VADDNEPBF16),
            ffi::_ND_INS_CLASS::ND_INS_VADDPD => Ok(Mnemonic::VADDPD),
            ffi::_ND_INS_CLASS::ND_INS_VADDPH => Ok(Mnemonic::VADDPH),
            ffi::_ND_INS_CLASS::ND_INS_VADDPS => Ok(Mnemonic::VADDPS),
            ffi::_ND_INS_CLASS::ND_INS_VADDSD => Ok(Mnemonic::VADDSD),
            ffi::_ND_INS_CLASS::ND_INS_VADDSH => Ok(Mnemonic::VADDSH),
            ffi::_ND_INS_CLASS::ND_INS_VADDSS => Ok(Mnemonic::VADDSS),
            ffi::_ND_INS_CLASS::ND_INS_VADDSUBPD => Ok(Mnemonic::VADDSUBPD),
            ffi::_ND_INS_CLASS::ND_INS_VADDSUBPS => Ok(Mnemonic::VADDSUBPS),
            ffi::_ND_INS_CLASS::ND_INS_VAESDEC => Ok(Mnemonic::VAESDEC),
            ffi::_ND_INS_CLASS::ND_INS_VAESDECLAST => Ok(Mnemonic::VAESDECLAST),
            ffi::_ND_INS_CLASS::ND_INS_VAESENC => Ok(Mnemonic::VAESENC),
            ffi::_ND_INS_CLASS::ND_INS_VAESENCLAST => Ok(Mnemonic::VAESENCLAST),
            ffi::_ND_INS_CLASS::ND_INS_VAESIMC => Ok(Mnemonic::VAESIMC),
            ffi::_ND_INS_CLASS::ND_INS_VAESKEYGENASSIST => Ok(Mnemonic::VAESKEYGENASSIST),
            ffi::_ND_INS_CLASS::ND_INS_VALIGND => Ok(Mnemonic::VALIGND),
            ffi::_ND_INS_CLASS::ND_INS_VALIGNQ => Ok(Mnemonic::VALIGNQ),
            ffi::_ND_INS_CLASS::ND_INS_VANDNPD => Ok(Mnemonic::VANDNPD),
            ffi::_ND_INS_CLASS::ND_INS_VANDNPS => Ok(Mnemonic::VANDNPS),
            ffi::_ND_INS_CLASS::ND_INS_VANDPD => Ok(Mnemonic::VANDPD),
            ffi::_ND_INS_CLASS::ND_INS_VANDPS => Ok(Mnemonic::VANDPS),
            ffi::_ND_INS_CLASS::ND_INS_VBCSTNEBF162PS => Ok(Mnemonic::VBCSTNEBF162PS),
            ffi::_ND_INS_CLASS::ND_INS_VBCSTNESH2PS => Ok(Mnemonic::VBCSTNESH2PS),
            ffi::_ND_INS_CLASS::ND_INS_VBLENDMPD => Ok(Mnemonic::VBLENDMPD),
            ffi::_ND_INS_CLASS::ND_INS_VBLENDMPS => Ok(Mnemonic::VBLENDMPS),
            ffi::_ND_INS_CLASS::ND_INS_VBLENDPD => Ok(Mnemonic::VBLENDPD),
            ffi::_ND_INS_CLASS::ND_INS_VBLENDPS => Ok(Mnemonic::VBLENDPS),
            ffi::_ND_INS_CLASS::ND_INS_VBLENDVPD => Ok(Mnemonic::VBLENDVPD),
            ffi::_ND_INS_CLASS::ND_INS_VBLENDVPS => Ok(Mnemonic::VBLENDVPS),
            ffi::_ND_INS_CLASS::ND_INS_VBROADCASTF128 => Ok(Mnemonic::VBROADCASTF128),
            ffi::_ND_INS_CLASS::ND_INS_VBROADCASTF32X2 => Ok(Mnemonic::VBROADCASTF32X2),
            ffi::_ND_INS_CLASS::ND_INS_VBROADCASTF32X4 => Ok(Mnemonic::VBROADCASTF32X4),
            ffi::_ND_INS_CLASS::ND_INS_VBROADCASTF32X8 => Ok(Mnemonic::VBROADCASTF32X8),
            ffi::_ND_INS_CLASS::ND_INS_VBROADCASTF64X2 => Ok(Mnemonic::VBROADCASTF64X2),
            ffi::_ND_INS_CLASS::ND_INS_VBROADCASTF64X4 => Ok(Mnemonic::VBROADCASTF64X4),
            ffi::_ND_INS_CLASS::ND_INS_VBROADCASTI128 => Ok(Mnemonic::VBROADCASTI128),
            ffi::_ND_INS_CLASS::ND_INS_VBROADCASTI32X2 => Ok(Mnemonic::VBROADCASTI32X2),
            ffi::_ND_INS_CLASS::ND_INS_VBROADCASTI32X4 => Ok(Mnemonic::VBROADCASTI32X4),
            ffi::_ND_INS_CLASS::ND_INS_VBROADCASTI32X8 => Ok(Mnemonic::VBROADCASTI32X8),
            ffi::_ND_INS_CLASS::ND_INS_VBROADCASTI64X2 => Ok(Mnemonic::VBROADCASTI64X2),
            ffi::_ND_INS_CLASS::ND_INS_VBROADCASTI64X4 => Ok(Mnemonic::VBROADCASTI64X4),
            ffi::_ND_INS_CLASS::ND_INS_VBROADCASTSD => Ok(Mnemonic::VBROADCASTSD),
            ffi::_ND_INS_CLASS::ND_INS_VBROADCASTSS => Ok(Mnemonic::VBROADCASTSS),
            ffi::_ND_INS_CLASS::ND_INS_VCMPPBF16 => Ok(Mnemonic::VCMPPBF16),
            ffi::_ND_INS_CLASS::ND_INS_VCMPPD => Ok(Mnemonic::VCMPPD),
            ffi::_ND_INS_CLASS::ND_INS_VCMPPH => Ok(Mnemonic::VCMPPH),
            ffi::_ND_INS_CLASS::ND_INS_VCMPPS => Ok(Mnemonic::VCMPPS),
            ffi::_ND_INS_CLASS::ND_INS_VCMPSD => Ok(Mnemonic::VCMPSD),
            ffi::_ND_INS_CLASS::ND_INS_VCMPSH => Ok(Mnemonic::VCMPSH),
            ffi::_ND_INS_CLASS::ND_INS_VCMPSS => Ok(Mnemonic::VCMPSS),
            ffi::_ND_INS_CLASS::ND_INS_VCOMISD => Ok(Mnemonic::VCOMISD),
            ffi::_ND_INS_CLASS::ND_INS_VCOMISH => Ok(Mnemonic::VCOMISH),
            ffi::_ND_INS_CLASS::ND_INS_VCOMISS => Ok(Mnemonic::VCOMISS),
            ffi::_ND_INS_CLASS::ND_INS_VCOMPRESSPD => Ok(Mnemonic::VCOMPRESSPD),
            ffi::_ND_INS_CLASS::ND_INS_VCOMPRESSPS => Ok(Mnemonic::VCOMPRESSPS),
            ffi::_ND_INS_CLASS::ND_INS_VCOMSBF16 => Ok(Mnemonic::VCOMSBF16),
            ffi::_ND_INS_CLASS::ND_INS_VCOMXSD => Ok(Mnemonic::VCOMXSD),
            ffi::_ND_INS_CLASS::ND_INS_VCOMXSH => Ok(Mnemonic::VCOMXSH),
            ffi::_ND_INS_CLASS::ND_INS_VCOMXSS => Ok(Mnemonic::VCOMXSS),
            ffi::_ND_INS_CLASS::ND_INS_VCVT2PS2PHX => Ok(Mnemonic::VCVT2PS2PHX),
            ffi::_ND_INS_CLASS::ND_INS_VCVTBIASPH2BF8 => Ok(Mnemonic::VCVTBIASPH2BF8),
            ffi::_ND_INS_CLASS::ND_INS_VCVTBIASPH2BF8S => Ok(Mnemonic::VCVTBIASPH2BF8S),
            ffi::_ND_INS_CLASS::ND_INS_VCVTBIASPH2HF8 => Ok(Mnemonic::VCVTBIASPH2HF8),
            ffi::_ND_INS_CLASS::ND_INS_VCVTBIASPH2HF8S => Ok(Mnemonic::VCVTBIASPH2HF8S),
            ffi::_ND_INS_CLASS::ND_INS_VCVTDQ2PD => Ok(Mnemonic::VCVTDQ2PD),
            ffi::_ND_INS_CLASS::ND_INS_VCVTDQ2PH => Ok(Mnemonic::VCVTDQ2PH),
            ffi::_ND_INS_CLASS::ND_INS_VCVTDQ2PS => Ok(Mnemonic::VCVTDQ2PS),
            ffi::_ND_INS_CLASS::ND_INS_VCVTHF82PH => Ok(Mnemonic::VCVTHF82PH),
            ffi::_ND_INS_CLASS::ND_INS_VCVTNE2PH2BF8 => Ok(Mnemonic::VCVTNE2PH2BF8),
            ffi::_ND_INS_CLASS::ND_INS_VCVTNE2PH2BF8S => Ok(Mnemonic::VCVTNE2PH2BF8S),
            ffi::_ND_INS_CLASS::ND_INS_VCVTNE2PH2HF8 => Ok(Mnemonic::VCVTNE2PH2HF8),
            ffi::_ND_INS_CLASS::ND_INS_VCVTNE2PH2HF8S => Ok(Mnemonic::VCVTNE2PH2HF8S),
            ffi::_ND_INS_CLASS::ND_INS_VCVTNE2PS2BF16 => Ok(Mnemonic::VCVTNE2PS2BF16),
            ffi::_ND_INS_CLASS::ND_INS_VCVTNEBF162IBS => Ok(Mnemonic::VCVTNEBF162IBS),
            ffi::_ND_INS_CLASS::ND_INS_VCVTNEBF162IUBS => Ok(Mnemonic::VCVTNEBF162IUBS),
            ffi::_ND_INS_CLASS::ND_INS_VCVTNEEBF162PS => Ok(Mnemonic::VCVTNEEBF162PS),
            ffi::_ND_INS_CLASS::ND_INS_VCVTNEEPH2PS => Ok(Mnemonic::VCVTNEEPH2PS),
            ffi::_ND_INS_CLASS::ND_INS_VCVTNEOBF162PS => Ok(Mnemonic::VCVTNEOBF162PS),
            ffi::_ND_INS_CLASS::ND_INS_VCVTNEOPH2PS => Ok(Mnemonic::VCVTNEOPH2PS),
            ffi::_ND_INS_CLASS::ND_INS_VCVTNEPH2BF8 => Ok(Mnemonic::VCVTNEPH2BF8),
            ffi::_ND_INS_CLASS::ND_INS_VCVTNEPH2BF8S => Ok(Mnemonic::VCVTNEPH2BF8S),
            ffi::_ND_INS_CLASS::ND_INS_VCVTNEPH2HF8 => Ok(Mnemonic::VCVTNEPH2HF8),
            ffi::_ND_INS_CLASS::ND_INS_VCVTNEPH2HF8S => Ok(Mnemonic::VCVTNEPH2HF8S),
            ffi::_ND_INS_CLASS::ND_INS_VCVTNEPS2BF16 => Ok(Mnemonic::VCVTNEPS2BF16),
            ffi::_ND_INS_CLASS::ND_INS_VCVTPD2DQ => Ok(Mnemonic::VCVTPD2DQ),
            ffi::_ND_INS_CLASS::ND_INS_VCVTPD2PH => Ok(Mnemonic::VCVTPD2PH),
            ffi::_ND_INS_CLASS::ND_INS_VCVTPD2PS => Ok(Mnemonic::VCVTPD2PS),
            ffi::_ND_INS_CLASS::ND_INS_VCVTPD2QQ => Ok(Mnemonic::VCVTPD2QQ),
            ffi::_ND_INS_CLASS::ND_INS_VCVTPD2UDQ => Ok(Mnemonic::VCVTPD2UDQ),
            ffi::_ND_INS_CLASS::ND_INS_VCVTPD2UQQ => Ok(Mnemonic::VCVTPD2UQQ),
            ffi::_ND_INS_CLASS::ND_INS_VCVTPH2DQ => Ok(Mnemonic::VCVTPH2DQ),
            ffi::_ND_INS_CLASS::ND_INS_VCVTPH2IBS => Ok(Mnemonic::VCVTPH2IBS),
            ffi::_ND_INS_CLASS::ND_INS_VCVTPH2IUBS => Ok(Mnemonic::VCVTPH2IUBS),
            ffi::_ND_INS_CLASS::ND_INS_VCVTPH2PD => Ok(Mnemonic::VCVTPH2PD),
            ffi::_ND_INS_CLASS::ND_INS_VCVTPH2PS => Ok(Mnemonic::VCVTPH2PS),
            ffi::_ND_INS_CLASS::ND_INS_VCVTPH2PSX => Ok(Mnemonic::VCVTPH2PSX),
            ffi::_ND_INS_CLASS::ND_INS_VCVTPH2QQ => Ok(Mnemonic::VCVTPH2QQ),
            ffi::_ND_INS_CLASS::ND_INS_VCVTPH2UDQ => Ok(Mnemonic::VCVTPH2UDQ),
            ffi::_ND_INS_CLASS::ND_INS_VCVTPH2UQQ => Ok(Mnemonic::VCVTPH2UQQ),
            ffi::_ND_INS_CLASS::ND_INS_VCVTPH2UW => Ok(Mnemonic::VCVTPH2UW),
            ffi::_ND_INS_CLASS::ND_INS_VCVTPH2W => Ok(Mnemonic::VCVTPH2W),
            ffi::_ND_INS_CLASS::ND_INS_VCVTPS2DQ => Ok(Mnemonic::VCVTPS2DQ),
            ffi::_ND_INS_CLASS::ND_INS_VCVTPS2IBS => Ok(Mnemonic::VCVTPS2IBS),
            ffi::_ND_INS_CLASS::ND_INS_VCVTPS2IUBS => Ok(Mnemonic::VCVTPS2IUBS),
            ffi::_ND_INS_CLASS::ND_INS_VCVTPS2PD => Ok(Mnemonic::VCVTPS2PD),
            ffi::_ND_INS_CLASS::ND_INS_VCVTPS2PH => Ok(Mnemonic::VCVTPS2PH),
            ffi::_ND_INS_CLASS::ND_INS_VCVTPS2PHX => Ok(Mnemonic::VCVTPS2PHX),
            ffi::_ND_INS_CLASS::ND_INS_VCVTPS2QQ => Ok(Mnemonic::VCVTPS2QQ),
            ffi::_ND_INS_CLASS::ND_INS_VCVTPS2UDQ => Ok(Mnemonic::VCVTPS2UDQ),
            ffi::_ND_INS_CLASS::ND_INS_VCVTPS2UQQ => Ok(Mnemonic::VCVTPS2UQQ),
            ffi::_ND_INS_CLASS::ND_INS_VCVTQQ2PD => Ok(Mnemonic::VCVTQQ2PD),
            ffi::_ND_INS_CLASS::ND_INS_VCVTQQ2PH => Ok(Mnemonic::VCVTQQ2PH),
            ffi::_ND_INS_CLASS::ND_INS_VCVTQQ2PS => Ok(Mnemonic::VCVTQQ2PS),
            ffi::_ND_INS_CLASS::ND_INS_VCVTSD2SH => Ok(Mnemonic::VCVTSD2SH),
            ffi::_ND_INS_CLASS::ND_INS_VCVTSD2SI => Ok(Mnemonic::VCVTSD2SI),
            ffi::_ND_INS_CLASS::ND_INS_VCVTSD2SS => Ok(Mnemonic::VCVTSD2SS),
            ffi::_ND_INS_CLASS::ND_INS_VCVTSD2USI => Ok(Mnemonic::VCVTSD2USI),
            ffi::_ND_INS_CLASS::ND_INS_VCVTSH2SD => Ok(Mnemonic::VCVTSH2SD),
            ffi::_ND_INS_CLASS::ND_INS_VCVTSH2SI => Ok(Mnemonic::VCVTSH2SI),
            ffi::_ND_INS_CLASS::ND_INS_VCVTSH2SS => Ok(Mnemonic::VCVTSH2SS),
            ffi::_ND_INS_CLASS::ND_INS_VCVTSH2USI => Ok(Mnemonic::VCVTSH2USI),
            ffi::_ND_INS_CLASS::ND_INS_VCVTSI2SD => Ok(Mnemonic::VCVTSI2SD),
            ffi::_ND_INS_CLASS::ND_INS_VCVTSI2SH => Ok(Mnemonic::VCVTSI2SH),
            ffi::_ND_INS_CLASS::ND_INS_VCVTSI2SS => Ok(Mnemonic::VCVTSI2SS),
            ffi::_ND_INS_CLASS::ND_INS_VCVTSS2SD => Ok(Mnemonic::VCVTSS2SD),
            ffi::_ND_INS_CLASS::ND_INS_VCVTSS2SH => Ok(Mnemonic::VCVTSS2SH),
            ffi::_ND_INS_CLASS::ND_INS_VCVTSS2SI => Ok(Mnemonic::VCVTSS2SI),
            ffi::_ND_INS_CLASS::ND_INS_VCVTSS2USI => Ok(Mnemonic::VCVTSS2USI),
            ffi::_ND_INS_CLASS::ND_INS_VCVTTNEBF162IBS => Ok(Mnemonic::VCVTTNEBF162IBS),
            ffi::_ND_INS_CLASS::ND_INS_VCVTTNEBF162IUBS => Ok(Mnemonic::VCVTTNEBF162IUBS),
            ffi::_ND_INS_CLASS::ND_INS_VCVTTPD2DQ => Ok(Mnemonic::VCVTTPD2DQ),
            ffi::_ND_INS_CLASS::ND_INS_VCVTTPD2DQS => Ok(Mnemonic::VCVTTPD2DQS),
            ffi::_ND_INS_CLASS::ND_INS_VCVTTPD2QQ => Ok(Mnemonic::VCVTTPD2QQ),
            ffi::_ND_INS_CLASS::ND_INS_VCVTTPD2QQS => Ok(Mnemonic::VCVTTPD2QQS),
            ffi::_ND_INS_CLASS::ND_INS_VCVTTPD2UDQ => Ok(Mnemonic::VCVTTPD2UDQ),
            ffi::_ND_INS_CLASS::ND_INS_VCVTTPD2UDQS => Ok(Mnemonic::VCVTTPD2UDQS),
            ffi::_ND_INS_CLASS::ND_INS_VCVTTPD2UQQ => Ok(Mnemonic::VCVTTPD2UQQ),
            ffi::_ND_INS_CLASS::ND_INS_VCVTTPD2UQQS => Ok(Mnemonic::VCVTTPD2UQQS),
            ffi::_ND_INS_CLASS::ND_INS_VCVTTPH2DQ => Ok(Mnemonic::VCVTTPH2DQ),
            ffi::_ND_INS_CLASS::ND_INS_VCVTTPH2IBS => Ok(Mnemonic::VCVTTPH2IBS),
            ffi::_ND_INS_CLASS::ND_INS_VCVTTPH2IUBS => Ok(Mnemonic::VCVTTPH2IUBS),
            ffi::_ND_INS_CLASS::ND_INS_VCVTTPH2QQ => Ok(Mnemonic::VCVTTPH2QQ),
            ffi::_ND_INS_CLASS::ND_INS_VCVTTPH2UDQ => Ok(Mnemonic::VCVTTPH2UDQ),
            ffi::_ND_INS_CLASS::ND_INS_VCVTTPH2UQQ => Ok(Mnemonic::VCVTTPH2UQQ),
            ffi::_ND_INS_CLASS::ND_INS_VCVTTPH2UW => Ok(Mnemonic::VCVTTPH2UW),
            ffi::_ND_INS_CLASS::ND_INS_VCVTTPH2W => Ok(Mnemonic::VCVTTPH2W),
            ffi::_ND_INS_CLASS::ND_INS_VCVTTPS2DQ => Ok(Mnemonic::VCVTTPS2DQ),
            ffi::_ND_INS_CLASS::ND_INS_VCVTTPS2DQS => Ok(Mnemonic::VCVTTPS2DQS),
            ffi::_ND_INS_CLASS::ND_INS_VCVTTPS2IBS => Ok(Mnemonic::VCVTTPS2IBS),
            ffi::_ND_INS_CLASS::ND_INS_VCVTTPS2IUBS => Ok(Mnemonic::VCVTTPS2IUBS),
            ffi::_ND_INS_CLASS::ND_INS_VCVTTPS2QQ => Ok(Mnemonic::VCVTTPS2QQ),
            ffi::_ND_INS_CLASS::ND_INS_VCVTTPS2QQS => Ok(Mnemonic::VCVTTPS2QQS),
            ffi::_ND_INS_CLASS::ND_INS_VCVTTPS2UDQ => Ok(Mnemonic::VCVTTPS2UDQ),
            ffi::_ND_INS_CLASS::ND_INS_VCVTTPS2UDQS => Ok(Mnemonic::VCVTTPS2UDQS),
            ffi::_ND_INS_CLASS::ND_INS_VCVTTPS2UQQ => Ok(Mnemonic::VCVTTPS2UQQ),
            ffi::_ND_INS_CLASS::ND_INS_VCVTTPS2UQQS => Ok(Mnemonic::VCVTTPS2UQQS),
            ffi::_ND_INS_CLASS::ND_INS_VCVTTSD2SI => Ok(Mnemonic::VCVTTSD2SI),
            ffi::_ND_INS_CLASS::ND_INS_VCVTTSD2SIS => Ok(Mnemonic::VCVTTSD2SIS),
            ffi::_ND_INS_CLASS::ND_INS_VCVTTSD2USI => Ok(Mnemonic::VCVTTSD2USI),
            ffi::_ND_INS_CLASS::ND_INS_VCVTTSD2USIS => Ok(Mnemonic::VCVTTSD2USIS),
            ffi::_ND_INS_CLASS::ND_INS_VCVTTSH2SI => Ok(Mnemonic::VCVTTSH2SI),
            ffi::_ND_INS_CLASS::ND_INS_VCVTTSH2USI => Ok(Mnemonic::VCVTTSH2USI),
            ffi::_ND_INS_CLASS::ND_INS_VCVTTSS2SI => Ok(Mnemonic::VCVTTSS2SI),
            ffi::_ND_INS_CLASS::ND_INS_VCVTTSS2SIS => Ok(Mnemonic::VCVTTSS2SIS),
            ffi::_ND_INS_CLASS::ND_INS_VCVTTSS2USI => Ok(Mnemonic::VCVTTSS2USI),
            ffi::_ND_INS_CLASS::ND_INS_VCVTTSS2USIS => Ok(Mnemonic::VCVTTSS2USIS),
            ffi::_ND_INS_CLASS::ND_INS_VCVTUDQ2PD => Ok(Mnemonic::VCVTUDQ2PD),
            ffi::_ND_INS_CLASS::ND_INS_VCVTUDQ2PH => Ok(Mnemonic::VCVTUDQ2PH),
            ffi::_ND_INS_CLASS::ND_INS_VCVTUDQ2PS => Ok(Mnemonic::VCVTUDQ2PS),
            ffi::_ND_INS_CLASS::ND_INS_VCVTUQQ2PD => Ok(Mnemonic::VCVTUQQ2PD),
            ffi::_ND_INS_CLASS::ND_INS_VCVTUQQ2PH => Ok(Mnemonic::VCVTUQQ2PH),
            ffi::_ND_INS_CLASS::ND_INS_VCVTUQQ2PS => Ok(Mnemonic::VCVTUQQ2PS),
            ffi::_ND_INS_CLASS::ND_INS_VCVTUSI2SD => Ok(Mnemonic::VCVTUSI2SD),
            ffi::_ND_INS_CLASS::ND_INS_VCVTUSI2SH => Ok(Mnemonic::VCVTUSI2SH),
            ffi::_ND_INS_CLASS::ND_INS_VCVTUSI2SS => Ok(Mnemonic::VCVTUSI2SS),
            ffi::_ND_INS_CLASS::ND_INS_VCVTUW2PH => Ok(Mnemonic::VCVTUW2PH),
            ffi::_ND_INS_CLASS::ND_INS_VCVTW2PH => Ok(Mnemonic::VCVTW2PH),
            ffi::_ND_INS_CLASS::ND_INS_VDBPSADBW => Ok(Mnemonic::VDBPSADBW),
            ffi::_ND_INS_CLASS::ND_INS_VDIVNEPBF16 => Ok(Mnemonic::VDIVNEPBF16),
            ffi::_ND_INS_CLASS::ND_INS_VDIVPD => Ok(Mnemonic::VDIVPD),
            ffi::_ND_INS_CLASS::ND_INS_VDIVPH => Ok(Mnemonic::VDIVPH),
            ffi::_ND_INS_CLASS::ND_INS_VDIVPS => Ok(Mnemonic::VDIVPS),
            ffi::_ND_INS_CLASS::ND_INS_VDIVSD => Ok(Mnemonic::VDIVSD),
            ffi::_ND_INS_CLASS::ND_INS_VDIVSH => Ok(Mnemonic::VDIVSH),
            ffi::_ND_INS_CLASS::ND_INS_VDIVSS => Ok(Mnemonic::VDIVSS),
            ffi::_ND_INS_CLASS::ND_INS_VDPBF16PS => Ok(Mnemonic::VDPBF16PS),
            ffi::_ND_INS_CLASS::ND_INS_VDPPD => Ok(Mnemonic::VDPPD),
            ffi::_ND_INS_CLASS::ND_INS_VDPPHPS => Ok(Mnemonic::VDPPHPS),
            ffi::_ND_INS_CLASS::ND_INS_VDPPS => Ok(Mnemonic::VDPPS),
            ffi::_ND_INS_CLASS::ND_INS_VERR => Ok(Mnemonic::VERR),
            ffi::_ND_INS_CLASS::ND_INS_VERW => Ok(Mnemonic::VERW),
            ffi::_ND_INS_CLASS::ND_INS_VEXP2PD => Ok(Mnemonic::VEXP2PD),
            ffi::_ND_INS_CLASS::ND_INS_VEXP2PS => Ok(Mnemonic::VEXP2PS),
            ffi::_ND_INS_CLASS::ND_INS_VEXPANDPD => Ok(Mnemonic::VEXPANDPD),
            ffi::_ND_INS_CLASS::ND_INS_VEXPANDPS => Ok(Mnemonic::VEXPANDPS),
            ffi::_ND_INS_CLASS::ND_INS_VEXTRACTF128 => Ok(Mnemonic::VEXTRACTF128),
            ffi::_ND_INS_CLASS::ND_INS_VEXTRACTF32X4 => Ok(Mnemonic::VEXTRACTF32X4),
            ffi::_ND_INS_CLASS::ND_INS_VEXTRACTF32X8 => Ok(Mnemonic::VEXTRACTF32X8),
            ffi::_ND_INS_CLASS::ND_INS_VEXTRACTF64X2 => Ok(Mnemonic::VEXTRACTF64X2),
            ffi::_ND_INS_CLASS::ND_INS_VEXTRACTF64X4 => Ok(Mnemonic::VEXTRACTF64X4),
            ffi::_ND_INS_CLASS::ND_INS_VEXTRACTI128 => Ok(Mnemonic::VEXTRACTI128),
            ffi::_ND_INS_CLASS::ND_INS_VEXTRACTI32X4 => Ok(Mnemonic::VEXTRACTI32X4),
            ffi::_ND_INS_CLASS::ND_INS_VEXTRACTI32X8 => Ok(Mnemonic::VEXTRACTI32X8),
            ffi::_ND_INS_CLASS::ND_INS_VEXTRACTI64X2 => Ok(Mnemonic::VEXTRACTI64X2),
            ffi::_ND_INS_CLASS::ND_INS_VEXTRACTI64X4 => Ok(Mnemonic::VEXTRACTI64X4),
            ffi::_ND_INS_CLASS::ND_INS_VEXTRACTPS => Ok(Mnemonic::VEXTRACTPS),
            ffi::_ND_INS_CLASS::ND_INS_VFCMADDCPH => Ok(Mnemonic::VFCMADDCPH),
            ffi::_ND_INS_CLASS::ND_INS_VFCMADDCSH => Ok(Mnemonic::VFCMADDCSH),
            ffi::_ND_INS_CLASS::ND_INS_VFCMULCPH => Ok(Mnemonic::VFCMULCPH),
            ffi::_ND_INS_CLASS::ND_INS_VFCMULCSH => Ok(Mnemonic::VFCMULCSH),
            ffi::_ND_INS_CLASS::ND_INS_VFIXUPIMMPD => Ok(Mnemonic::VFIXUPIMMPD),
            ffi::_ND_INS_CLASS::ND_INS_VFIXUPIMMPS => Ok(Mnemonic::VFIXUPIMMPS),
            ffi::_ND_INS_CLASS::ND_INS_VFIXUPIMMSD => Ok(Mnemonic::VFIXUPIMMSD),
            ffi::_ND_INS_CLASS::ND_INS_VFIXUPIMMSS => Ok(Mnemonic::VFIXUPIMMSS),
            ffi::_ND_INS_CLASS::ND_INS_VFMADD132NEPBF16 => Ok(Mnemonic::VFMADD132NEPBF16),
            ffi::_ND_INS_CLASS::ND_INS_VFMADD132PD => Ok(Mnemonic::VFMADD132PD),
            ffi::_ND_INS_CLASS::ND_INS_VFMADD132PH => Ok(Mnemonic::VFMADD132PH),
            ffi::_ND_INS_CLASS::ND_INS_VFMADD132PS => Ok(Mnemonic::VFMADD132PS),
            ffi::_ND_INS_CLASS::ND_INS_VFMADD132SD => Ok(Mnemonic::VFMADD132SD),
            ffi::_ND_INS_CLASS::ND_INS_VFMADD132SH => Ok(Mnemonic::VFMADD132SH),
            ffi::_ND_INS_CLASS::ND_INS_VFMADD132SS => Ok(Mnemonic::VFMADD132SS),
            ffi::_ND_INS_CLASS::ND_INS_VFMADD213NEPBF16 => Ok(Mnemonic::VFMADD213NEPBF16),
            ffi::_ND_INS_CLASS::ND_INS_VFMADD213PD => Ok(Mnemonic::VFMADD213PD),
            ffi::_ND_INS_CLASS::ND_INS_VFMADD213PH => Ok(Mnemonic::VFMADD213PH),
            ffi::_ND_INS_CLASS::ND_INS_VFMADD213PS => Ok(Mnemonic::VFMADD213PS),
            ffi::_ND_INS_CLASS::ND_INS_VFMADD213SD => Ok(Mnemonic::VFMADD213SD),
            ffi::_ND_INS_CLASS::ND_INS_VFMADD213SH => Ok(Mnemonic::VFMADD213SH),
            ffi::_ND_INS_CLASS::ND_INS_VFMADD213SS => Ok(Mnemonic::VFMADD213SS),
            ffi::_ND_INS_CLASS::ND_INS_VFMADD231NEPBF16 => Ok(Mnemonic::VFMADD231NEPBF16),
            ffi::_ND_INS_CLASS::ND_INS_VFMADD231PD => Ok(Mnemonic::VFMADD231PD),
            ffi::_ND_INS_CLASS::ND_INS_VFMADD231PH => Ok(Mnemonic::VFMADD231PH),
            ffi::_ND_INS_CLASS::ND_INS_VFMADD231PS => Ok(Mnemonic::VFMADD231PS),
            ffi::_ND_INS_CLASS::ND_INS_VFMADD231SD => Ok(Mnemonic::VFMADD231SD),
            ffi::_ND_INS_CLASS::ND_INS_VFMADD231SH => Ok(Mnemonic::VFMADD231SH),
            ffi::_ND_INS_CLASS::ND_INS_VFMADD231SS => Ok(Mnemonic::VFMADD231SS),
            ffi::_ND_INS_CLASS::ND_INS_VFMADDCPH => Ok(Mnemonic::VFMADDCPH),
            ffi::_ND_INS_CLASS::ND_INS_VFMADDCSH => Ok(Mnemonic::VFMADDCSH),
            ffi::_ND_INS_CLASS::ND_INS_VFMADDPD => Ok(Mnemonic::VFMADDPD),
            ffi::_ND_INS_CLASS::ND_INS_VFMADDPS => Ok(Mnemonic::VFMADDPS),
            ffi::_ND_INS_CLASS::ND_INS_VFMADDSD => Ok(Mnemonic::VFMADDSD),
            ffi::_ND_INS_CLASS::ND_INS_VFMADDSS => Ok(Mnemonic::VFMADDSS),
            ffi::_ND_INS_CLASS::ND_INS_VFMADDSUB132PD => Ok(Mnemonic::VFMADDSUB132PD),
            ffi::_ND_INS_CLASS::ND_INS_VFMADDSUB132PH => Ok(Mnemonic::VFMADDSUB132PH),
            ffi::_ND_INS_CLASS::ND_INS_VFMADDSUB132PS => Ok(Mnemonic::VFMADDSUB132PS),
            ffi::_ND_INS_CLASS::ND_INS_VFMADDSUB213PD => Ok(Mnemonic::VFMADDSUB213PD),
            ffi::_ND_INS_CLASS::ND_INS_VFMADDSUB213PH => Ok(Mnemonic::VFMADDSUB213PH),
            ffi::_ND_INS_CLASS::ND_INS_VFMADDSUB213PS => Ok(Mnemonic::VFMADDSUB213PS),
            ffi::_ND_INS_CLASS::ND_INS_VFMADDSUB231PD => Ok(Mnemonic::VFMADDSUB231PD),
            ffi::_ND_INS_CLASS::ND_INS_VFMADDSUB231PH => Ok(Mnemonic::VFMADDSUB231PH),
            ffi::_ND_INS_CLASS::ND_INS_VFMADDSUB231PS => Ok(Mnemonic::VFMADDSUB231PS),
            ffi::_ND_INS_CLASS::ND_INS_VFMADDSUBPD => Ok(Mnemonic::VFMADDSUBPD),
            ffi::_ND_INS_CLASS::ND_INS_VFMADDSUBPS => Ok(Mnemonic::VFMADDSUBPS),
            ffi::_ND_INS_CLASS::ND_INS_VFMSUB132NEPBF16 => Ok(Mnemonic::VFMSUB132NEPBF16),
            ffi::_ND_INS_CLASS::ND_INS_VFMSUB132PD => Ok(Mnemonic::VFMSUB132PD),
            ffi::_ND_INS_CLASS::ND_INS_VFMSUB132PH => Ok(Mnemonic::VFMSUB132PH),
            ffi::_ND_INS_CLASS::ND_INS_VFMSUB132PS => Ok(Mnemonic::VFMSUB132PS),
            ffi::_ND_INS_CLASS::ND_INS_VFMSUB132SD => Ok(Mnemonic::VFMSUB132SD),
            ffi::_ND_INS_CLASS::ND_INS_VFMSUB132SH => Ok(Mnemonic::VFMSUB132SH),
            ffi::_ND_INS_CLASS::ND_INS_VFMSUB132SS => Ok(Mnemonic::VFMSUB132SS),
            ffi::_ND_INS_CLASS::ND_INS_VFMSUB213NEPBF16 => Ok(Mnemonic::VFMSUB213NEPBF16),
            ffi::_ND_INS_CLASS::ND_INS_VFMSUB213PD => Ok(Mnemonic::VFMSUB213PD),
            ffi::_ND_INS_CLASS::ND_INS_VFMSUB213PH => Ok(Mnemonic::VFMSUB213PH),
            ffi::_ND_INS_CLASS::ND_INS_VFMSUB213PS => Ok(Mnemonic::VFMSUB213PS),
            ffi::_ND_INS_CLASS::ND_INS_VFMSUB213SD => Ok(Mnemonic::VFMSUB213SD),
            ffi::_ND_INS_CLASS::ND_INS_VFMSUB213SH => Ok(Mnemonic::VFMSUB213SH),
            ffi::_ND_INS_CLASS::ND_INS_VFMSUB213SS => Ok(Mnemonic::VFMSUB213SS),
            ffi::_ND_INS_CLASS::ND_INS_VFMSUB231NEPBF16 => Ok(Mnemonic::VFMSUB231NEPBF16),
            ffi::_ND_INS_CLASS::ND_INS_VFMSUB231PD => Ok(Mnemonic::VFMSUB231PD),
            ffi::_ND_INS_CLASS::ND_INS_VFMSUB231PH => Ok(Mnemonic::VFMSUB231PH),
            ffi::_ND_INS_CLASS::ND_INS_VFMSUB231PS => Ok(Mnemonic::VFMSUB231PS),
            ffi::_ND_INS_CLASS::ND_INS_VFMSUB231SD => Ok(Mnemonic::VFMSUB231SD),
            ffi::_ND_INS_CLASS::ND_INS_VFMSUB231SH => Ok(Mnemonic::VFMSUB231SH),
            ffi::_ND_INS_CLASS::ND_INS_VFMSUB231SS => Ok(Mnemonic::VFMSUB231SS),
            ffi::_ND_INS_CLASS::ND_INS_VFMSUBADD132PD => Ok(Mnemonic::VFMSUBADD132PD),
            ffi::_ND_INS_CLASS::ND_INS_VFMSUBADD132PH => Ok(Mnemonic::VFMSUBADD132PH),
            ffi::_ND_INS_CLASS::ND_INS_VFMSUBADD132PS => Ok(Mnemonic::VFMSUBADD132PS),
            ffi::_ND_INS_CLASS::ND_INS_VFMSUBADD213PD => Ok(Mnemonic::VFMSUBADD213PD),
            ffi::_ND_INS_CLASS::ND_INS_VFMSUBADD213PH => Ok(Mnemonic::VFMSUBADD213PH),
            ffi::_ND_INS_CLASS::ND_INS_VFMSUBADD213PS => Ok(Mnemonic::VFMSUBADD213PS),
            ffi::_ND_INS_CLASS::ND_INS_VFMSUBADD231PD => Ok(Mnemonic::VFMSUBADD231PD),
            ffi::_ND_INS_CLASS::ND_INS_VFMSUBADD231PH => Ok(Mnemonic::VFMSUBADD231PH),
            ffi::_ND_INS_CLASS::ND_INS_VFMSUBADD231PS => Ok(Mnemonic::VFMSUBADD231PS),
            ffi::_ND_INS_CLASS::ND_INS_VFMSUBADDPD => Ok(Mnemonic::VFMSUBADDPD),
            ffi::_ND_INS_CLASS::ND_INS_VFMSUBADDPS => Ok(Mnemonic::VFMSUBADDPS),
            ffi::_ND_INS_CLASS::ND_INS_VFMSUBPD => Ok(Mnemonic::VFMSUBPD),
            ffi::_ND_INS_CLASS::ND_INS_VFMSUBPS => Ok(Mnemonic::VFMSUBPS),
            ffi::_ND_INS_CLASS::ND_INS_VFMSUBSD => Ok(Mnemonic::VFMSUBSD),
            ffi::_ND_INS_CLASS::ND_INS_VFMSUBSS => Ok(Mnemonic::VFMSUBSS),
            ffi::_ND_INS_CLASS::ND_INS_VFMULCPH => Ok(Mnemonic::VFMULCPH),
            ffi::_ND_INS_CLASS::ND_INS_VFMULCSH => Ok(Mnemonic::VFMULCSH),
            ffi::_ND_INS_CLASS::ND_INS_VFNMADD132NEPBF16 => Ok(Mnemonic::VFNMADD132NEPBF16),
            ffi::_ND_INS_CLASS::ND_INS_VFNMADD132PD => Ok(Mnemonic::VFNMADD132PD),
            ffi::_ND_INS_CLASS::ND_INS_VFNMADD132PH => Ok(Mnemonic::VFNMADD132PH),
            ffi::_ND_INS_CLASS::ND_INS_VFNMADD132PS => Ok(Mnemonic::VFNMADD132PS),
            ffi::_ND_INS_CLASS::ND_INS_VFNMADD132SD => Ok(Mnemonic::VFNMADD132SD),
            ffi::_ND_INS_CLASS::ND_INS_VFNMADD132SH => Ok(Mnemonic::VFNMADD132SH),
            ffi::_ND_INS_CLASS::ND_INS_VFNMADD132SS => Ok(Mnemonic::VFNMADD132SS),
            ffi::_ND_INS_CLASS::ND_INS_VFNMADD213NEPBF16 => Ok(Mnemonic::VFNMADD213NEPBF16),
            ffi::_ND_INS_CLASS::ND_INS_VFNMADD213PD => Ok(Mnemonic::VFNMADD213PD),
            ffi::_ND_INS_CLASS::ND_INS_VFNMADD213PH => Ok(Mnemonic::VFNMADD213PH),
            ffi::_ND_INS_CLASS::ND_INS_VFNMADD213PS => Ok(Mnemonic::VFNMADD213PS),
            ffi::_ND_INS_CLASS::ND_INS_VFNMADD213SD => Ok(Mnemonic::VFNMADD213SD),
            ffi::_ND_INS_CLASS::ND_INS_VFNMADD213SH => Ok(Mnemonic::VFNMADD213SH),
            ffi::_ND_INS_CLASS::ND_INS_VFNMADD213SS => Ok(Mnemonic::VFNMADD213SS),
            ffi::_ND_INS_CLASS::ND_INS_VFNMADD231NEPBF16 => Ok(Mnemonic::VFNMADD231NEPBF16),
            ffi::_ND_INS_CLASS::ND_INS_VFNMADD231PD => Ok(Mnemonic::VFNMADD231PD),
            ffi::_ND_INS_CLASS::ND_INS_VFNMADD231PH => Ok(Mnemonic::VFNMADD231PH),
            ffi::_ND_INS_CLASS::ND_INS_VFNMADD231PS => Ok(Mnemonic::VFNMADD231PS),
            ffi::_ND_INS_CLASS::ND_INS_VFNMADD231SD => Ok(Mnemonic::VFNMADD231SD),
            ffi::_ND_INS_CLASS::ND_INS_VFNMADD231SH => Ok(Mnemonic::VFNMADD231SH),
            ffi::_ND_INS_CLASS::ND_INS_VFNMADD231SS => Ok(Mnemonic::VFNMADD231SS),
            ffi::_ND_INS_CLASS::ND_INS_VFNMADDPD => Ok(Mnemonic::VFNMADDPD),
            ffi::_ND_INS_CLASS::ND_INS_VFNMADDPS => Ok(Mnemonic::VFNMADDPS),
            ffi::_ND_INS_CLASS::ND_INS_VFNMADDSD => Ok(Mnemonic::VFNMADDSD),
            ffi::_ND_INS_CLASS::ND_INS_VFNMADDSS => Ok(Mnemonic::VFNMADDSS),
            ffi::_ND_INS_CLASS::ND_INS_VFNMSUB132NEPBF16 => Ok(Mnemonic::VFNMSUB132NEPBF16),
            ffi::_ND_INS_CLASS::ND_INS_VFNMSUB132PD => Ok(Mnemonic::VFNMSUB132PD),
            ffi::_ND_INS_CLASS::ND_INS_VFNMSUB132PH => Ok(Mnemonic::VFNMSUB132PH),
            ffi::_ND_INS_CLASS::ND_INS_VFNMSUB132PS => Ok(Mnemonic::VFNMSUB132PS),
            ffi::_ND_INS_CLASS::ND_INS_VFNMSUB132SD => Ok(Mnemonic::VFNMSUB132SD),
            ffi::_ND_INS_CLASS::ND_INS_VFNMSUB132SH => Ok(Mnemonic::VFNMSUB132SH),
            ffi::_ND_INS_CLASS::ND_INS_VFNMSUB132SS => Ok(Mnemonic::VFNMSUB132SS),
            ffi::_ND_INS_CLASS::ND_INS_VFNMSUB213NEPBF16 => Ok(Mnemonic::VFNMSUB213NEPBF16),
            ffi::_ND_INS_CLASS::ND_INS_VFNMSUB213PD => Ok(Mnemonic::VFNMSUB213PD),
            ffi::_ND_INS_CLASS::ND_INS_VFNMSUB213PH => Ok(Mnemonic::VFNMSUB213PH),
            ffi::_ND_INS_CLASS::ND_INS_VFNMSUB213PS => Ok(Mnemonic::VFNMSUB213PS),
            ffi::_ND_INS_CLASS::ND_INS_VFNMSUB213SD => Ok(Mnemonic::VFNMSUB213SD),
            ffi::_ND_INS_CLASS::ND_INS_VFNMSUB213SH => Ok(Mnemonic::VFNMSUB213SH),
            ffi::_ND_INS_CLASS::ND_INS_VFNMSUB213SS => Ok(Mnemonic::VFNMSUB213SS),
            ffi::_ND_INS_CLASS::ND_INS_VFNMSUB231NEPBF16 => Ok(Mnemonic::VFNMSUB231NEPBF16),
            ffi::_ND_INS_CLASS::ND_INS_VFNMSUB231PD => Ok(Mnemonic::VFNMSUB231PD),
            ffi::_ND_INS_CLASS::ND_INS_VFNMSUB231PH => Ok(Mnemonic::VFNMSUB231PH),
            ffi::_ND_INS_CLASS::ND_INS_VFNMSUB231PS => Ok(Mnemonic::VFNMSUB231PS),
            ffi::_ND_INS_CLASS::ND_INS_VFNMSUB231SD => Ok(Mnemonic::VFNMSUB231SD),
            ffi::_ND_INS_CLASS::ND_INS_VFNMSUB231SH => Ok(Mnemonic::VFNMSUB231SH),
            ffi::_ND_INS_CLASS::ND_INS_VFNMSUB231SS => Ok(Mnemonic::VFNMSUB231SS),
            ffi::_ND_INS_CLASS::ND_INS_VFNMSUBPD => Ok(Mnemonic::VFNMSUBPD),
            ffi::_ND_INS_CLASS::ND_INS_VFNMSUBPS => Ok(Mnemonic::VFNMSUBPS),
            ffi::_ND_INS_CLASS::ND_INS_VFNMSUBSD => Ok(Mnemonic::VFNMSUBSD),
            ffi::_ND_INS_CLASS::ND_INS_VFNMSUBSS => Ok(Mnemonic::VFNMSUBSS),
            ffi::_ND_INS_CLASS::ND_INS_VFPCLASSPBF16 => Ok(Mnemonic::VFPCLASSPBF16),
            ffi::_ND_INS_CLASS::ND_INS_VFPCLASSPD => Ok(Mnemonic::VFPCLASSPD),
            ffi::_ND_INS_CLASS::ND_INS_VFPCLASSPH => Ok(Mnemonic::VFPCLASSPH),
            ffi::_ND_INS_CLASS::ND_INS_VFPCLASSPS => Ok(Mnemonic::VFPCLASSPS),
            ffi::_ND_INS_CLASS::ND_INS_VFPCLASSSD => Ok(Mnemonic::VFPCLASSSD),
            ffi::_ND_INS_CLASS::ND_INS_VFPCLASSSH => Ok(Mnemonic::VFPCLASSSH),
            ffi::_ND_INS_CLASS::ND_INS_VFPCLASSSS => Ok(Mnemonic::VFPCLASSSS),
            ffi::_ND_INS_CLASS::ND_INS_VFRCZPD => Ok(Mnemonic::VFRCZPD),
            ffi::_ND_INS_CLASS::ND_INS_VFRCZPS => Ok(Mnemonic::VFRCZPS),
            ffi::_ND_INS_CLASS::ND_INS_VFRCZSD => Ok(Mnemonic::VFRCZSD),
            ffi::_ND_INS_CLASS::ND_INS_VFRCZSS => Ok(Mnemonic::VFRCZSS),
            ffi::_ND_INS_CLASS::ND_INS_VGATHERDPD => Ok(Mnemonic::VGATHERDPD),
            ffi::_ND_INS_CLASS::ND_INS_VGATHERDPS => Ok(Mnemonic::VGATHERDPS),
            ffi::_ND_INS_CLASS::ND_INS_VGATHERPF0DPD => Ok(Mnemonic::VGATHERPF0DPD),
            ffi::_ND_INS_CLASS::ND_INS_VGATHERPF0DPS => Ok(Mnemonic::VGATHERPF0DPS),
            ffi::_ND_INS_CLASS::ND_INS_VGATHERPF0QPD => Ok(Mnemonic::VGATHERPF0QPD),
            ffi::_ND_INS_CLASS::ND_INS_VGATHERPF0QPS => Ok(Mnemonic::VGATHERPF0QPS),
            ffi::_ND_INS_CLASS::ND_INS_VGATHERPF1DPD => Ok(Mnemonic::VGATHERPF1DPD),
            ffi::_ND_INS_CLASS::ND_INS_VGATHERPF1DPS => Ok(Mnemonic::VGATHERPF1DPS),
            ffi::_ND_INS_CLASS::ND_INS_VGATHERPF1QPD => Ok(Mnemonic::VGATHERPF1QPD),
            ffi::_ND_INS_CLASS::ND_INS_VGATHERPF1QPS => Ok(Mnemonic::VGATHERPF1QPS),
            ffi::_ND_INS_CLASS::ND_INS_VGATHERQPD => Ok(Mnemonic::VGATHERQPD),
            ffi::_ND_INS_CLASS::ND_INS_VGATHERQPS => Ok(Mnemonic::VGATHERQPS),
            ffi::_ND_INS_CLASS::ND_INS_VGETEXPPBF16 => Ok(Mnemonic::VGETEXPPBF16),
            ffi::_ND_INS_CLASS::ND_INS_VGETEXPPD => Ok(Mnemonic::VGETEXPPD),
            ffi::_ND_INS_CLASS::ND_INS_VGETEXPPH => Ok(Mnemonic::VGETEXPPH),
            ffi::_ND_INS_CLASS::ND_INS_VGETEXPPS => Ok(Mnemonic::VGETEXPPS),
            ffi::_ND_INS_CLASS::ND_INS_VGETEXPSD => Ok(Mnemonic::VGETEXPSD),
            ffi::_ND_INS_CLASS::ND_INS_VGETEXPSH => Ok(Mnemonic::VGETEXPSH),
            ffi::_ND_INS_CLASS::ND_INS_VGETEXPSS => Ok(Mnemonic::VGETEXPSS),
            ffi::_ND_INS_CLASS::ND_INS_VGETMANTPBF16 => Ok(Mnemonic::VGETMANTPBF16),
            ffi::_ND_INS_CLASS::ND_INS_VGETMANTPD => Ok(Mnemonic::VGETMANTPD),
            ffi::_ND_INS_CLASS::ND_INS_VGETMANTPH => Ok(Mnemonic::VGETMANTPH),
            ffi::_ND_INS_CLASS::ND_INS_VGETMANTPS => Ok(Mnemonic::VGETMANTPS),
            ffi::_ND_INS_CLASS::ND_INS_VGETMANTSD => Ok(Mnemonic::VGETMANTSD),
            ffi::_ND_INS_CLASS::ND_INS_VGETMANTSH => Ok(Mnemonic::VGETMANTSH),
            ffi::_ND_INS_CLASS::ND_INS_VGETMANTSS => Ok(Mnemonic::VGETMANTSS),
            ffi::_ND_INS_CLASS::ND_INS_VGF2P8AFFINEINVQB => Ok(Mnemonic::VGF2P8AFFINEINVQB),
            ffi::_ND_INS_CLASS::ND_INS_VGF2P8AFFINEQB => Ok(Mnemonic::VGF2P8AFFINEQB),
            ffi::_ND_INS_CLASS::ND_INS_VGF2P8MULB => Ok(Mnemonic::VGF2P8MULB),
            ffi::_ND_INS_CLASS::ND_INS_VHADDPD => Ok(Mnemonic::VHADDPD),
            ffi::_ND_INS_CLASS::ND_INS_VHADDPS => Ok(Mnemonic::VHADDPS),
            ffi::_ND_INS_CLASS::ND_INS_VHSUBPD => Ok(Mnemonic::VHSUBPD),
            ffi::_ND_INS_CLASS::ND_INS_VHSUBPS => Ok(Mnemonic::VHSUBPS),
            ffi::_ND_INS_CLASS::ND_INS_VINSERTF128 => Ok(Mnemonic::VINSERTF128),
            ffi::_ND_INS_CLASS::ND_INS_VINSERTF32X4 => Ok(Mnemonic::VINSERTF32X4),
            ffi::_ND_INS_CLASS::ND_INS_VINSERTF32X8 => Ok(Mnemonic::VINSERTF32X8),
            ffi::_ND_INS_CLASS::ND_INS_VINSERTF64X2 => Ok(Mnemonic::VINSERTF64X2),
            ffi::_ND_INS_CLASS::ND_INS_VINSERTF64X4 => Ok(Mnemonic::VINSERTF64X4),
            ffi::_ND_INS_CLASS::ND_INS_VINSERTI128 => Ok(Mnemonic::VINSERTI128),
            ffi::_ND_INS_CLASS::ND_INS_VINSERTI32X4 => Ok(Mnemonic::VINSERTI32X4),
            ffi::_ND_INS_CLASS::ND_INS_VINSERTI32X8 => Ok(Mnemonic::VINSERTI32X8),
            ffi::_ND_INS_CLASS::ND_INS_VINSERTI64X2 => Ok(Mnemonic::VINSERTI64X2),
            ffi::_ND_INS_CLASS::ND_INS_VINSERTI64X4 => Ok(Mnemonic::VINSERTI64X4),
            ffi::_ND_INS_CLASS::ND_INS_VINSERTPS => Ok(Mnemonic::VINSERTPS),
            ffi::_ND_INS_CLASS::ND_INS_VLDDQU => Ok(Mnemonic::VLDDQU),
            ffi::_ND_INS_CLASS::ND_INS_VLDMXCSR => Ok(Mnemonic::VLDMXCSR),
            ffi::_ND_INS_CLASS::ND_INS_VMASKMOVDQU => Ok(Mnemonic::VMASKMOVDQU),
            ffi::_ND_INS_CLASS::ND_INS_VMASKMOVPD => Ok(Mnemonic::VMASKMOVPD),
            ffi::_ND_INS_CLASS::ND_INS_VMASKMOVPS => Ok(Mnemonic::VMASKMOVPS),
            ffi::_ND_INS_CLASS::ND_INS_VMAXPBF16 => Ok(Mnemonic::VMAXPBF16),
            ffi::_ND_INS_CLASS::ND_INS_VMAXPD => Ok(Mnemonic::VMAXPD),
            ffi::_ND_INS_CLASS::ND_INS_VMAXPH => Ok(Mnemonic::VMAXPH),
            ffi::_ND_INS_CLASS::ND_INS_VMAXPS => Ok(Mnemonic::VMAXPS),
            ffi::_ND_INS_CLASS::ND_INS_VMAXSD => Ok(Mnemonic::VMAXSD),
            ffi::_ND_INS_CLASS::ND_INS_VMAXSH => Ok(Mnemonic::VMAXSH),
            ffi::_ND_INS_CLASS::ND_INS_VMAXSS => Ok(Mnemonic::VMAXSS),
            ffi::_ND_INS_CLASS::ND_INS_VMCALL => Ok(Mnemonic::VMCALL),
            ffi::_ND_INS_CLASS::ND_INS_VMCLEAR => Ok(Mnemonic::VMCLEAR),
            ffi::_ND_INS_CLASS::ND_INS_VMFUNC => Ok(Mnemonic::VMFUNC),
            ffi::_ND_INS_CLASS::ND_INS_VMGEXIT => Ok(Mnemonic::VMGEXIT),
            ffi::_ND_INS_CLASS::ND_INS_VMINMAXNEPBF16 => Ok(Mnemonic::VMINMAXNEPBF16),
            ffi::_ND_INS_CLASS::ND_INS_VMINMAXPD => Ok(Mnemonic::VMINMAXPD),
            ffi::_ND_INS_CLASS::ND_INS_VMINMAXPH => Ok(Mnemonic::VMINMAXPH),
            ffi::_ND_INS_CLASS::ND_INS_VMINMAXPS => Ok(Mnemonic::VMINMAXPS),
            ffi::_ND_INS_CLASS::ND_INS_VMINMAXSD => Ok(Mnemonic::VMINMAXSD),
            ffi::_ND_INS_CLASS::ND_INS_VMINMAXSH => Ok(Mnemonic::VMINMAXSH),
            ffi::_ND_INS_CLASS::ND_INS_VMINMAXSS => Ok(Mnemonic::VMINMAXSS),
            ffi::_ND_INS_CLASS::ND_INS_VMINPBF16 => Ok(Mnemonic::VMINPBF16),
            ffi::_ND_INS_CLASS::ND_INS_VMINPD => Ok(Mnemonic::VMINPD),
            ffi::_ND_INS_CLASS::ND_INS_VMINPH => Ok(Mnemonic::VMINPH),
            ffi::_ND_INS_CLASS::ND_INS_VMINPS => Ok(Mnemonic::VMINPS),
            ffi::_ND_INS_CLASS::ND_INS_VMINSD => Ok(Mnemonic::VMINSD),
            ffi::_ND_INS_CLASS::ND_INS_VMINSH => Ok(Mnemonic::VMINSH),
            ffi::_ND_INS_CLASS::ND_INS_VMINSS => Ok(Mnemonic::VMINSS),
            ffi::_ND_INS_CLASS::ND_INS_VMLAUNCH => Ok(Mnemonic::VMLAUNCH),
            ffi::_ND_INS_CLASS::ND_INS_VMLOAD => Ok(Mnemonic::VMLOAD),
            ffi::_ND_INS_CLASS::ND_INS_VMMCALL => Ok(Mnemonic::VMMCALL),
            ffi::_ND_INS_CLASS::ND_INS_VMOVAPD => Ok(Mnemonic::VMOVAPD),
            ffi::_ND_INS_CLASS::ND_INS_VMOVAPS => Ok(Mnemonic::VMOVAPS),
            ffi::_ND_INS_CLASS::ND_INS_VMOVD => Ok(Mnemonic::VMOVD),
            ffi::_ND_INS_CLASS::ND_INS_VMOVDDUP => Ok(Mnemonic::VMOVDDUP),
            ffi::_ND_INS_CLASS::ND_INS_VMOVDQA => Ok(Mnemonic::VMOVDQA),
            ffi::_ND_INS_CLASS::ND_INS_VMOVDQA32 => Ok(Mnemonic::VMOVDQA32),
            ffi::_ND_INS_CLASS::ND_INS_VMOVDQA64 => Ok(Mnemonic::VMOVDQA64),
            ffi::_ND_INS_CLASS::ND_INS_VMOVDQU => Ok(Mnemonic::VMOVDQU),
            ffi::_ND_INS_CLASS::ND_INS_VMOVDQU16 => Ok(Mnemonic::VMOVDQU16),
            ffi::_ND_INS_CLASS::ND_INS_VMOVDQU32 => Ok(Mnemonic::VMOVDQU32),
            ffi::_ND_INS_CLASS::ND_INS_VMOVDQU64 => Ok(Mnemonic::VMOVDQU64),
            ffi::_ND_INS_CLASS::ND_INS_VMOVDQU8 => Ok(Mnemonic::VMOVDQU8),
            ffi::_ND_INS_CLASS::ND_INS_VMOVHLPS => Ok(Mnemonic::VMOVHLPS),
            ffi::_ND_INS_CLASS::ND_INS_VMOVHPD => Ok(Mnemonic::VMOVHPD),
            ffi::_ND_INS_CLASS::ND_INS_VMOVHPS => Ok(Mnemonic::VMOVHPS),
            ffi::_ND_INS_CLASS::ND_INS_VMOVLHPS => Ok(Mnemonic::VMOVLHPS),
            ffi::_ND_INS_CLASS::ND_INS_VMOVLPD => Ok(Mnemonic::VMOVLPD),
            ffi::_ND_INS_CLASS::ND_INS_VMOVLPS => Ok(Mnemonic::VMOVLPS),
            ffi::_ND_INS_CLASS::ND_INS_VMOVMSKPD => Ok(Mnemonic::VMOVMSKPD),
            ffi::_ND_INS_CLASS::ND_INS_VMOVMSKPS => Ok(Mnemonic::VMOVMSKPS),
            ffi::_ND_INS_CLASS::ND_INS_VMOVNTDQ => Ok(Mnemonic::VMOVNTDQ),
            ffi::_ND_INS_CLASS::ND_INS_VMOVNTDQA => Ok(Mnemonic::VMOVNTDQA),
            ffi::_ND_INS_CLASS::ND_INS_VMOVNTPD => Ok(Mnemonic::VMOVNTPD),
            ffi::_ND_INS_CLASS::ND_INS_VMOVNTPS => Ok(Mnemonic::VMOVNTPS),
            ffi::_ND_INS_CLASS::ND_INS_VMOVQ => Ok(Mnemonic::VMOVQ),
            ffi::_ND_INS_CLASS::ND_INS_VMOVSD => Ok(Mnemonic::VMOVSD),
            ffi::_ND_INS_CLASS::ND_INS_VMOVSH => Ok(Mnemonic::VMOVSH),
            ffi::_ND_INS_CLASS::ND_INS_VMOVSHDUP => Ok(Mnemonic::VMOVSHDUP),
            ffi::_ND_INS_CLASS::ND_INS_VMOVSLDUP => Ok(Mnemonic::VMOVSLDUP),
            ffi::_ND_INS_CLASS::ND_INS_VMOVSS => Ok(Mnemonic::VMOVSS),
            ffi::_ND_INS_CLASS::ND_INS_VMOVUPD => Ok(Mnemonic::VMOVUPD),
            ffi::_ND_INS_CLASS::ND_INS_VMOVUPS => Ok(Mnemonic::VMOVUPS),
            ffi::_ND_INS_CLASS::ND_INS_VMOVW => Ok(Mnemonic::VMOVW),
            ffi::_ND_INS_CLASS::ND_INS_VMPSADBW => Ok(Mnemonic::VMPSADBW),
            ffi::_ND_INS_CLASS::ND_INS_VMPTRLD => Ok(Mnemonic::VMPTRLD),
            ffi::_ND_INS_CLASS::ND_INS_VMPTRST => Ok(Mnemonic::VMPTRST),
            ffi::_ND_INS_CLASS::ND_INS_VMREAD => Ok(Mnemonic::VMREAD),
            ffi::_ND_INS_CLASS::ND_INS_VMRESUME => Ok(Mnemonic::VMRESUME),
            ffi::_ND_INS_CLASS::ND_INS_VMRUN => Ok(Mnemonic::VMRUN),
            ffi::_ND_INS_CLASS::ND_INS_VMSAVE => Ok(Mnemonic::VMSAVE),
            ffi::_ND_INS_CLASS::ND_INS_VMULNEPBF16 => Ok(Mnemonic::VMULNEPBF16),
            ffi::_ND_INS_CLASS::ND_INS_VMULPD => Ok(Mnemonic::VMULPD),
            ffi::_ND_INS_CLASS::ND_INS_VMULPH => Ok(Mnemonic::VMULPH),
            ffi::_ND_INS_CLASS::ND_INS_VMULPS => Ok(Mnemonic::VMULPS),
            ffi::_ND_INS_CLASS::ND_INS_VMULSD => Ok(Mnemonic::VMULSD),
            ffi::_ND_INS_CLASS::ND_INS_VMULSH => Ok(Mnemonic::VMULSH),
            ffi::_ND_INS_CLASS::ND_INS_VMULSS => Ok(Mnemonic::VMULSS),
            ffi::_ND_INS_CLASS::ND_INS_VMWRITE => Ok(Mnemonic::VMWRITE),
            ffi::_ND_INS_CLASS::ND_INS_VMXOFF => Ok(Mnemonic::VMXOFF),
            ffi::_ND_INS_CLASS::ND_INS_VMXON => Ok(Mnemonic::VMXON),
            ffi::_ND_INS_CLASS::ND_INS_VORPD => Ok(Mnemonic::VORPD),
            ffi::_ND_INS_CLASS::ND_INS_VORPS => Ok(Mnemonic::VORPS),
            ffi::_ND_INS_CLASS::ND_INS_VP2INTERSECTD => Ok(Mnemonic::VP2INTERSECTD),
            ffi::_ND_INS_CLASS::ND_INS_VP2INTERSECTQ => Ok(Mnemonic::VP2INTERSECTQ),
            ffi::_ND_INS_CLASS::ND_INS_VP4DPWSSD => Ok(Mnemonic::VP4DPWSSD),
            ffi::_ND_INS_CLASS::ND_INS_VP4DPWSSDS => Ok(Mnemonic::VP4DPWSSDS),
            ffi::_ND_INS_CLASS::ND_INS_VPABSB => Ok(Mnemonic::VPABSB),
            ffi::_ND_INS_CLASS::ND_INS_VPABSD => Ok(Mnemonic::VPABSD),
            ffi::_ND_INS_CLASS::ND_INS_VPABSQ => Ok(Mnemonic::VPABSQ),
            ffi::_ND_INS_CLASS::ND_INS_VPABSW => Ok(Mnemonic::VPABSW),
            ffi::_ND_INS_CLASS::ND_INS_VPACKSSDW => Ok(Mnemonic::VPACKSSDW),
            ffi::_ND_INS_CLASS::ND_INS_VPACKSSWB => Ok(Mnemonic::VPACKSSWB),
            ffi::_ND_INS_CLASS::ND_INS_VPACKUSDW => Ok(Mnemonic::VPACKUSDW),
            ffi::_ND_INS_CLASS::ND_INS_VPACKUSWB => Ok(Mnemonic::VPACKUSWB),
            ffi::_ND_INS_CLASS::ND_INS_VPADDB => Ok(Mnemonic::VPADDB),
            ffi::_ND_INS_CLASS::ND_INS_VPADDD => Ok(Mnemonic::VPADDD),
            ffi::_ND_INS_CLASS::ND_INS_VPADDQ => Ok(Mnemonic::VPADDQ),
            ffi::_ND_INS_CLASS::ND_INS_VPADDSB => Ok(Mnemonic::VPADDSB),
            ffi::_ND_INS_CLASS::ND_INS_VPADDSW => Ok(Mnemonic::VPADDSW),
            ffi::_ND_INS_CLASS::ND_INS_VPADDUSB => Ok(Mnemonic::VPADDUSB),
            ffi::_ND_INS_CLASS::ND_INS_VPADDUSW => Ok(Mnemonic::VPADDUSW),
            ffi::_ND_INS_CLASS::ND_INS_VPADDW => Ok(Mnemonic::VPADDW),
            ffi::_ND_INS_CLASS::ND_INS_VPALIGNR => Ok(Mnemonic::VPALIGNR),
            ffi::_ND_INS_CLASS::ND_INS_VPAND => Ok(Mnemonic::VPAND),
            ffi::_ND_INS_CLASS::ND_INS_VPANDD => Ok(Mnemonic::VPANDD),
            ffi::_ND_INS_CLASS::ND_INS_VPANDN => Ok(Mnemonic::VPANDN),
            ffi::_ND_INS_CLASS::ND_INS_VPANDND => Ok(Mnemonic::VPANDND),
            ffi::_ND_INS_CLASS::ND_INS_VPANDNQ => Ok(Mnemonic::VPANDNQ),
            ffi::_ND_INS_CLASS::ND_INS_VPANDQ => Ok(Mnemonic::VPANDQ),
            ffi::_ND_INS_CLASS::ND_INS_VPAVGB => Ok(Mnemonic::VPAVGB),
            ffi::_ND_INS_CLASS::ND_INS_VPAVGW => Ok(Mnemonic::VPAVGW),
            ffi::_ND_INS_CLASS::ND_INS_VPBLENDD => Ok(Mnemonic::VPBLENDD),
            ffi::_ND_INS_CLASS::ND_INS_VPBLENDMB => Ok(Mnemonic::VPBLENDMB),
            ffi::_ND_INS_CLASS::ND_INS_VPBLENDMD => Ok(Mnemonic::VPBLENDMD),
            ffi::_ND_INS_CLASS::ND_INS_VPBLENDMQ => Ok(Mnemonic::VPBLENDMQ),
            ffi::_ND_INS_CLASS::ND_INS_VPBLENDMW => Ok(Mnemonic::VPBLENDMW),
            ffi::_ND_INS_CLASS::ND_INS_VPBLENDVB => Ok(Mnemonic::VPBLENDVB),
            ffi::_ND_INS_CLASS::ND_INS_VPBLENDW => Ok(Mnemonic::VPBLENDW),
            ffi::_ND_INS_CLASS::ND_INS_VPBROADCASTB => Ok(Mnemonic::VPBROADCASTB),
            ffi::_ND_INS_CLASS::ND_INS_VPBROADCASTD => Ok(Mnemonic::VPBROADCASTD),
            ffi::_ND_INS_CLASS::ND_INS_VPBROADCASTMB2Q => Ok(Mnemonic::VPBROADCASTMB2Q),
            ffi::_ND_INS_CLASS::ND_INS_VPBROADCASTMW2D => Ok(Mnemonic::VPBROADCASTMW2D),
            ffi::_ND_INS_CLASS::ND_INS_VPBROADCASTQ => Ok(Mnemonic::VPBROADCASTQ),
            ffi::_ND_INS_CLASS::ND_INS_VPBROADCASTW => Ok(Mnemonic::VPBROADCASTW),
            ffi::_ND_INS_CLASS::ND_INS_VPCLMULQDQ => Ok(Mnemonic::VPCLMULQDQ),
            ffi::_ND_INS_CLASS::ND_INS_VPCMOV => Ok(Mnemonic::VPCMOV),
            ffi::_ND_INS_CLASS::ND_INS_VPCMPB => Ok(Mnemonic::VPCMPB),
            ffi::_ND_INS_CLASS::ND_INS_VPCMPD => Ok(Mnemonic::VPCMPD),
            ffi::_ND_INS_CLASS::ND_INS_VPCMPEQB => Ok(Mnemonic::VPCMPEQB),
            ffi::_ND_INS_CLASS::ND_INS_VPCMPEQD => Ok(Mnemonic::VPCMPEQD),
            ffi::_ND_INS_CLASS::ND_INS_VPCMPEQQ => Ok(Mnemonic::VPCMPEQQ),
            ffi::_ND_INS_CLASS::ND_INS_VPCMPEQW => Ok(Mnemonic::VPCMPEQW),
            ffi::_ND_INS_CLASS::ND_INS_VPCMPESTRI => Ok(Mnemonic::VPCMPESTRI),
            ffi::_ND_INS_CLASS::ND_INS_VPCMPESTRM => Ok(Mnemonic::VPCMPESTRM),
            ffi::_ND_INS_CLASS::ND_INS_VPCMPGTB => Ok(Mnemonic::VPCMPGTB),
            ffi::_ND_INS_CLASS::ND_INS_VPCMPGTD => Ok(Mnemonic::VPCMPGTD),
            ffi::_ND_INS_CLASS::ND_INS_VPCMPGTQ => Ok(Mnemonic::VPCMPGTQ),
            ffi::_ND_INS_CLASS::ND_INS_VPCMPGTW => Ok(Mnemonic::VPCMPGTW),
            ffi::_ND_INS_CLASS::ND_INS_VPCMPISTRI => Ok(Mnemonic::VPCMPISTRI),
            ffi::_ND_INS_CLASS::ND_INS_VPCMPISTRM => Ok(Mnemonic::VPCMPISTRM),
            ffi::_ND_INS_CLASS::ND_INS_VPCMPQ => Ok(Mnemonic::VPCMPQ),
            ffi::_ND_INS_CLASS::ND_INS_VPCMPUB => Ok(Mnemonic::VPCMPUB),
            ffi::_ND_INS_CLASS::ND_INS_VPCMPUD => Ok(Mnemonic::VPCMPUD),
            ffi::_ND_INS_CLASS::ND_INS_VPCMPUQ => Ok(Mnemonic::VPCMPUQ),
            ffi::_ND_INS_CLASS::ND_INS_VPCMPUW => Ok(Mnemonic::VPCMPUW),
            ffi::_ND_INS_CLASS::ND_INS_VPCMPW => Ok(Mnemonic::VPCMPW),
            ffi::_ND_INS_CLASS::ND_INS_VPCOMB => Ok(Mnemonic::VPCOMB),
            ffi::_ND_INS_CLASS::ND_INS_VPCOMD => Ok(Mnemonic::VPCOMD),
            ffi::_ND_INS_CLASS::ND_INS_VPCOMPRESSB => Ok(Mnemonic::VPCOMPRESSB),
            ffi::_ND_INS_CLASS::ND_INS_VPCOMPRESSD => Ok(Mnemonic::VPCOMPRESSD),
            ffi::_ND_INS_CLASS::ND_INS_VPCOMPRESSQ => Ok(Mnemonic::VPCOMPRESSQ),
            ffi::_ND_INS_CLASS::ND_INS_VPCOMPRESSW => Ok(Mnemonic::VPCOMPRESSW),
            ffi::_ND_INS_CLASS::ND_INS_VPCOMQ => Ok(Mnemonic::VPCOMQ),
            ffi::_ND_INS_CLASS::ND_INS_VPCOMUB => Ok(Mnemonic::VPCOMUB),
            ffi::_ND_INS_CLASS::ND_INS_VPCOMUD => Ok(Mnemonic::VPCOMUD),
            ffi::_ND_INS_CLASS::ND_INS_VPCOMUQ => Ok(Mnemonic::VPCOMUQ),
            ffi::_ND_INS_CLASS::ND_INS_VPCOMUW => Ok(Mnemonic::VPCOMUW),
            ffi::_ND_INS_CLASS::ND_INS_VPCOMW => Ok(Mnemonic::VPCOMW),
            ffi::_ND_INS_CLASS::ND_INS_VPCONFLICTD => Ok(Mnemonic::VPCONFLICTD),
            ffi::_ND_INS_CLASS::ND_INS_VPCONFLICTQ => Ok(Mnemonic::VPCONFLICTQ),
            ffi::_ND_INS_CLASS::ND_INS_VPDPBSSD => Ok(Mnemonic::VPDPBSSD),
            ffi::_ND_INS_CLASS::ND_INS_VPDPBSSDS => Ok(Mnemonic::VPDPBSSDS),
            ffi::_ND_INS_CLASS::ND_INS_VPDPBSUD => Ok(Mnemonic::VPDPBSUD),
            ffi::_ND_INS_CLASS::ND_INS_VPDPBSUDS => Ok(Mnemonic::VPDPBSUDS),
            ffi::_ND_INS_CLASS::ND_INS_VPDPBUSD => Ok(Mnemonic::VPDPBUSD),
            ffi::_ND_INS_CLASS::ND_INS_VPDPBUSDS => Ok(Mnemonic::VPDPBUSDS),
            ffi::_ND_INS_CLASS::ND_INS_VPDPBUUD => Ok(Mnemonic::VPDPBUUD),
            ffi::_ND_INS_CLASS::ND_INS_VPDPBUUDS => Ok(Mnemonic::VPDPBUUDS),
            ffi::_ND_INS_CLASS::ND_INS_VPDPWSSD => Ok(Mnemonic::VPDPWSSD),
            ffi::_ND_INS_CLASS::ND_INS_VPDPWSSDS => Ok(Mnemonic::VPDPWSSDS),
            ffi::_ND_INS_CLASS::ND_INS_VPDPWSUD => Ok(Mnemonic::VPDPWSUD),
            ffi::_ND_INS_CLASS::ND_INS_VPDPWSUDS => Ok(Mnemonic::VPDPWSUDS),
            ffi::_ND_INS_CLASS::ND_INS_VPDPWUSD => Ok(Mnemonic::VPDPWUSD),
            ffi::_ND_INS_CLASS::ND_INS_VPDPWUSDS => Ok(Mnemonic::VPDPWUSDS),
            ffi::_ND_INS_CLASS::ND_INS_VPDPWUUD => Ok(Mnemonic::VPDPWUUD),
            ffi::_ND_INS_CLASS::ND_INS_VPDPWUUDS => Ok(Mnemonic::VPDPWUUDS),
            ffi::_ND_INS_CLASS::ND_INS_VPERM2F128 => Ok(Mnemonic::VPERM2F128),
            ffi::_ND_INS_CLASS::ND_INS_VPERM2I128 => Ok(Mnemonic::VPERM2I128),
            ffi::_ND_INS_CLASS::ND_INS_VPERMB => Ok(Mnemonic::VPERMB),
            ffi::_ND_INS_CLASS::ND_INS_VPERMD => Ok(Mnemonic::VPERMD),
            ffi::_ND_INS_CLASS::ND_INS_VPERMI2B => Ok(Mnemonic::VPERMI2B),
            ffi::_ND_INS_CLASS::ND_INS_VPERMI2D => Ok(Mnemonic::VPERMI2D),
            ffi::_ND_INS_CLASS::ND_INS_VPERMI2PD => Ok(Mnemonic::VPERMI2PD),
            ffi::_ND_INS_CLASS::ND_INS_VPERMI2PS => Ok(Mnemonic::VPERMI2PS),
            ffi::_ND_INS_CLASS::ND_INS_VPERMI2Q => Ok(Mnemonic::VPERMI2Q),
            ffi::_ND_INS_CLASS::ND_INS_VPERMI2W => Ok(Mnemonic::VPERMI2W),
            ffi::_ND_INS_CLASS::ND_INS_VPERMIL2PD => Ok(Mnemonic::VPERMIL2PD),
            ffi::_ND_INS_CLASS::ND_INS_VPERMIL2PS => Ok(Mnemonic::VPERMIL2PS),
            ffi::_ND_INS_CLASS::ND_INS_VPERMILPD => Ok(Mnemonic::VPERMILPD),
            ffi::_ND_INS_CLASS::ND_INS_VPERMILPS => Ok(Mnemonic::VPERMILPS),
            ffi::_ND_INS_CLASS::ND_INS_VPERMPD => Ok(Mnemonic::VPERMPD),
            ffi::_ND_INS_CLASS::ND_INS_VPERMPS => Ok(Mnemonic::VPERMPS),
            ffi::_ND_INS_CLASS::ND_INS_VPERMQ => Ok(Mnemonic::VPERMQ),
            ffi::_ND_INS_CLASS::ND_INS_VPERMT2B => Ok(Mnemonic::VPERMT2B),
            ffi::_ND_INS_CLASS::ND_INS_VPERMT2D => Ok(Mnemonic::VPERMT2D),
            ffi::_ND_INS_CLASS::ND_INS_VPERMT2PD => Ok(Mnemonic::VPERMT2PD),
            ffi::_ND_INS_CLASS::ND_INS_VPERMT2PS => Ok(Mnemonic::VPERMT2PS),
            ffi::_ND_INS_CLASS::ND_INS_VPERMT2Q => Ok(Mnemonic::VPERMT2Q),
            ffi::_ND_INS_CLASS::ND_INS_VPERMT2W => Ok(Mnemonic::VPERMT2W),
            ffi::_ND_INS_CLASS::ND_INS_VPERMW => Ok(Mnemonic::VPERMW),
            ffi::_ND_INS_CLASS::ND_INS_VPEXPANDB => Ok(Mnemonic::VPEXPANDB),
            ffi::_ND_INS_CLASS::ND_INS_VPEXPANDD => Ok(Mnemonic::VPEXPANDD),
            ffi::_ND_INS_CLASS::ND_INS_VPEXPANDQ => Ok(Mnemonic::VPEXPANDQ),
            ffi::_ND_INS_CLASS::ND_INS_VPEXPANDW => Ok(Mnemonic::VPEXPANDW),
            ffi::_ND_INS_CLASS::ND_INS_VPEXTRB => Ok(Mnemonic::VPEXTRB),
            ffi::_ND_INS_CLASS::ND_INS_VPEXTRD => Ok(Mnemonic::VPEXTRD),
            ffi::_ND_INS_CLASS::ND_INS_VPEXTRQ => Ok(Mnemonic::VPEXTRQ),
            ffi::_ND_INS_CLASS::ND_INS_VPEXTRW => Ok(Mnemonic::VPEXTRW),
            ffi::_ND_INS_CLASS::ND_INS_VPGATHERDD => Ok(Mnemonic::VPGATHERDD),
            ffi::_ND_INS_CLASS::ND_INS_VPGATHERDQ => Ok(Mnemonic::VPGATHERDQ),
            ffi::_ND_INS_CLASS::ND_INS_VPGATHERQD => Ok(Mnemonic::VPGATHERQD),
            ffi::_ND_INS_CLASS::ND_INS_VPGATHERQQ => Ok(Mnemonic::VPGATHERQQ),
            ffi::_ND_INS_CLASS::ND_INS_VPHADDBD => Ok(Mnemonic::VPHADDBD),
            ffi::_ND_INS_CLASS::ND_INS_VPHADDBQ => Ok(Mnemonic::VPHADDBQ),
            ffi::_ND_INS_CLASS::ND_INS_VPHADDBW => Ok(Mnemonic::VPHADDBW),
            ffi::_ND_INS_CLASS::ND_INS_VPHADDD => Ok(Mnemonic::VPHADDD),
            ffi::_ND_INS_CLASS::ND_INS_VPHADDDQ => Ok(Mnemonic::VPHADDDQ),
            ffi::_ND_INS_CLASS::ND_INS_VPHADDSW => Ok(Mnemonic::VPHADDSW),
            ffi::_ND_INS_CLASS::ND_INS_VPHADDUBD => Ok(Mnemonic::VPHADDUBD),
            ffi::_ND_INS_CLASS::ND_INS_VPHADDUBQ => Ok(Mnemonic::VPHADDUBQ),
            ffi::_ND_INS_CLASS::ND_INS_VPHADDUBW => Ok(Mnemonic::VPHADDUBW),
            ffi::_ND_INS_CLASS::ND_INS_VPHADDUDQ => Ok(Mnemonic::VPHADDUDQ),
            ffi::_ND_INS_CLASS::ND_INS_VPHADDUWD => Ok(Mnemonic::VPHADDUWD),
            ffi::_ND_INS_CLASS::ND_INS_VPHADDUWQ => Ok(Mnemonic::VPHADDUWQ),
            ffi::_ND_INS_CLASS::ND_INS_VPHADDW => Ok(Mnemonic::VPHADDW),
            ffi::_ND_INS_CLASS::ND_INS_VPHADDWD => Ok(Mnemonic::VPHADDWD),
            ffi::_ND_INS_CLASS::ND_INS_VPHADDWQ => Ok(Mnemonic::VPHADDWQ),
            ffi::_ND_INS_CLASS::ND_INS_VPHMINPOSUW => Ok(Mnemonic::VPHMINPOSUW),
            ffi::_ND_INS_CLASS::ND_INS_VPHSUBBW => Ok(Mnemonic::VPHSUBBW),
            ffi::_ND_INS_CLASS::ND_INS_VPHSUBD => Ok(Mnemonic::VPHSUBD),
            ffi::_ND_INS_CLASS::ND_INS_VPHSUBDQ => Ok(Mnemonic::VPHSUBDQ),
            ffi::_ND_INS_CLASS::ND_INS_VPHSUBSW => Ok(Mnemonic::VPHSUBSW),
            ffi::_ND_INS_CLASS::ND_INS_VPHSUBW => Ok(Mnemonic::VPHSUBW),
            ffi::_ND_INS_CLASS::ND_INS_VPHSUBWD => Ok(Mnemonic::VPHSUBWD),
            ffi::_ND_INS_CLASS::ND_INS_VPINSRB => Ok(Mnemonic::VPINSRB),
            ffi::_ND_INS_CLASS::ND_INS_VPINSRD => Ok(Mnemonic::VPINSRD),
            ffi::_ND_INS_CLASS::ND_INS_VPINSRQ => Ok(Mnemonic::VPINSRQ),
            ffi::_ND_INS_CLASS::ND_INS_VPINSRW => Ok(Mnemonic::VPINSRW),
            ffi::_ND_INS_CLASS::ND_INS_VPLZCNTD => Ok(Mnemonic::VPLZCNTD),
            ffi::_ND_INS_CLASS::ND_INS_VPLZCNTQ => Ok(Mnemonic::VPLZCNTQ),
            ffi::_ND_INS_CLASS::ND_INS_VPMACSDD => Ok(Mnemonic::VPMACSDD),
            ffi::_ND_INS_CLASS::ND_INS_VPMACSDQH => Ok(Mnemonic::VPMACSDQH),
            ffi::_ND_INS_CLASS::ND_INS_VPMACSDQL => Ok(Mnemonic::VPMACSDQL),
            ffi::_ND_INS_CLASS::ND_INS_VPMACSSDD => Ok(Mnemonic::VPMACSSDD),
            ffi::_ND_INS_CLASS::ND_INS_VPMACSSDQH => Ok(Mnemonic::VPMACSSDQH),
            ffi::_ND_INS_CLASS::ND_INS_VPMACSSDQL => Ok(Mnemonic::VPMACSSDQL),
            ffi::_ND_INS_CLASS::ND_INS_VPMACSSWD => Ok(Mnemonic::VPMACSSWD),
            ffi::_ND_INS_CLASS::ND_INS_VPMACSSWW => Ok(Mnemonic::VPMACSSWW),
            ffi::_ND_INS_CLASS::ND_INS_VPMACSWD => Ok(Mnemonic::VPMACSWD),
            ffi::_ND_INS_CLASS::ND_INS_VPMACSWW => Ok(Mnemonic::VPMACSWW),
            ffi::_ND_INS_CLASS::ND_INS_VPMADCSSWD => Ok(Mnemonic::VPMADCSSWD),
            ffi::_ND_INS_CLASS::ND_INS_VPMADCSWD => Ok(Mnemonic::VPMADCSWD),
            ffi::_ND_INS_CLASS::ND_INS_VPMADD52HUQ => Ok(Mnemonic::VPMADD52HUQ),
            ffi::_ND_INS_CLASS::ND_INS_VPMADD52LUQ => Ok(Mnemonic::VPMADD52LUQ),
            ffi::_ND_INS_CLASS::ND_INS_VPMADDUBSW => Ok(Mnemonic::VPMADDUBSW),
            ffi::_ND_INS_CLASS::ND_INS_VPMADDWD => Ok(Mnemonic::VPMADDWD),
            ffi::_ND_INS_CLASS::ND_INS_VPMASKMOVD => Ok(Mnemonic::VPMASKMOVD),
            ffi::_ND_INS_CLASS::ND_INS_VPMASKMOVQ => Ok(Mnemonic::VPMASKMOVQ),
            ffi::_ND_INS_CLASS::ND_INS_VPMAXSB => Ok(Mnemonic::VPMAXSB),
            ffi::_ND_INS_CLASS::ND_INS_VPMAXSD => Ok(Mnemonic::VPMAXSD),
            ffi::_ND_INS_CLASS::ND_INS_VPMAXSQ => Ok(Mnemonic::VPMAXSQ),
            ffi::_ND_INS_CLASS::ND_INS_VPMAXSW => Ok(Mnemonic::VPMAXSW),
            ffi::_ND_INS_CLASS::ND_INS_VPMAXUB => Ok(Mnemonic::VPMAXUB),
            ffi::_ND_INS_CLASS::ND_INS_VPMAXUD => Ok(Mnemonic::VPMAXUD),
            ffi::_ND_INS_CLASS::ND_INS_VPMAXUQ => Ok(Mnemonic::VPMAXUQ),
            ffi::_ND_INS_CLASS::ND_INS_VPMAXUW => Ok(Mnemonic::VPMAXUW),
            ffi::_ND_INS_CLASS::ND_INS_VPMINSB => Ok(Mnemonic::VPMINSB),
            ffi::_ND_INS_CLASS::ND_INS_VPMINSD => Ok(Mnemonic::VPMINSD),
            ffi::_ND_INS_CLASS::ND_INS_VPMINSQ => Ok(Mnemonic::VPMINSQ),
            ffi::_ND_INS_CLASS::ND_INS_VPMINSW => Ok(Mnemonic::VPMINSW),
            ffi::_ND_INS_CLASS::ND_INS_VPMINUB => Ok(Mnemonic::VPMINUB),
            ffi::_ND_INS_CLASS::ND_INS_VPMINUD => Ok(Mnemonic::VPMINUD),
            ffi::_ND_INS_CLASS::ND_INS_VPMINUQ => Ok(Mnemonic::VPMINUQ),
            ffi::_ND_INS_CLASS::ND_INS_VPMINUW => Ok(Mnemonic::VPMINUW),
            ffi::_ND_INS_CLASS::ND_INS_VPMOVB2M => Ok(Mnemonic::VPMOVB2M),
            ffi::_ND_INS_CLASS::ND_INS_VPMOVD2M => Ok(Mnemonic::VPMOVD2M),
            ffi::_ND_INS_CLASS::ND_INS_VPMOVDB => Ok(Mnemonic::VPMOVDB),
            ffi::_ND_INS_CLASS::ND_INS_VPMOVDW => Ok(Mnemonic::VPMOVDW),
            ffi::_ND_INS_CLASS::ND_INS_VPMOVM2B => Ok(Mnemonic::VPMOVM2B),
            ffi::_ND_INS_CLASS::ND_INS_VPMOVM2D => Ok(Mnemonic::VPMOVM2D),
            ffi::_ND_INS_CLASS::ND_INS_VPMOVM2Q => Ok(Mnemonic::VPMOVM2Q),
            ffi::_ND_INS_CLASS::ND_INS_VPMOVM2W => Ok(Mnemonic::VPMOVM2W),
            ffi::_ND_INS_CLASS::ND_INS_VPMOVMSKB => Ok(Mnemonic::VPMOVMSKB),
            ffi::_ND_INS_CLASS::ND_INS_VPMOVQ2M => Ok(Mnemonic::VPMOVQ2M),
            ffi::_ND_INS_CLASS::ND_INS_VPMOVQB => Ok(Mnemonic::VPMOVQB),
            ffi::_ND_INS_CLASS::ND_INS_VPMOVQD => Ok(Mnemonic::VPMOVQD),
            ffi::_ND_INS_CLASS::ND_INS_VPMOVQW => Ok(Mnemonic::VPMOVQW),
            ffi::_ND_INS_CLASS::ND_INS_VPMOVSDB => Ok(Mnemonic::VPMOVSDB),
            ffi::_ND_INS_CLASS::ND_INS_VPMOVSDW => Ok(Mnemonic::VPMOVSDW),
            ffi::_ND_INS_CLASS::ND_INS_VPMOVSQB => Ok(Mnemonic::VPMOVSQB),
            ffi::_ND_INS_CLASS::ND_INS_VPMOVSQD => Ok(Mnemonic::VPMOVSQD),
            ffi::_ND_INS_CLASS::ND_INS_VPMOVSQW => Ok(Mnemonic::VPMOVSQW),
            ffi::_ND_INS_CLASS::ND_INS_VPMOVSWB => Ok(Mnemonic::VPMOVSWB),
            ffi::_ND_INS_CLASS::ND_INS_VPMOVSXBD => Ok(Mnemonic::VPMOVSXBD),
            ffi::_ND_INS_CLASS::ND_INS_VPMOVSXBQ => Ok(Mnemonic::VPMOVSXBQ),
            ffi::_ND_INS_CLASS::ND_INS_VPMOVSXBW => Ok(Mnemonic::VPMOVSXBW),
            ffi::_ND_INS_CLASS::ND_INS_VPMOVSXDQ => Ok(Mnemonic::VPMOVSXDQ),
            ffi::_ND_INS_CLASS::ND_INS_VPMOVSXWD => Ok(Mnemonic::VPMOVSXWD),
            ffi::_ND_INS_CLASS::ND_INS_VPMOVSXWQ => Ok(Mnemonic::VPMOVSXWQ),
            ffi::_ND_INS_CLASS::ND_INS_VPMOVUSDB => Ok(Mnemonic::VPMOVUSDB),
            ffi::_ND_INS_CLASS::ND_INS_VPMOVUSDW => Ok(Mnemonic::VPMOVUSDW),
            ffi::_ND_INS_CLASS::ND_INS_VPMOVUSQB => Ok(Mnemonic::VPMOVUSQB),
            ffi::_ND_INS_CLASS::ND_INS_VPMOVUSQD => Ok(Mnemonic::VPMOVUSQD),
            ffi::_ND_INS_CLASS::ND_INS_VPMOVUSQW => Ok(Mnemonic::VPMOVUSQW),
            ffi::_ND_INS_CLASS::ND_INS_VPMOVUSWB => Ok(Mnemonic::VPMOVUSWB),
            ffi::_ND_INS_CLASS::ND_INS_VPMOVW2M => Ok(Mnemonic::VPMOVW2M),
            ffi::_ND_INS_CLASS::ND_INS_VPMOVWB => Ok(Mnemonic::VPMOVWB),
            ffi::_ND_INS_CLASS::ND_INS_VPMOVZXBD => Ok(Mnemonic::VPMOVZXBD),
            ffi::_ND_INS_CLASS::ND_INS_VPMOVZXBQ => Ok(Mnemonic::VPMOVZXBQ),
            ffi::_ND_INS_CLASS::ND_INS_VPMOVZXBW => Ok(Mnemonic::VPMOVZXBW),
            ffi::_ND_INS_CLASS::ND_INS_VPMOVZXDQ => Ok(Mnemonic::VPMOVZXDQ),
            ffi::_ND_INS_CLASS::ND_INS_VPMOVZXWD => Ok(Mnemonic::VPMOVZXWD),
            ffi::_ND_INS_CLASS::ND_INS_VPMOVZXWQ => Ok(Mnemonic::VPMOVZXWQ),
            ffi::_ND_INS_CLASS::ND_INS_VPMULDQ => Ok(Mnemonic::VPMULDQ),
            ffi::_ND_INS_CLASS::ND_INS_VPMULHRSW => Ok(Mnemonic::VPMULHRSW),
            ffi::_ND_INS_CLASS::ND_INS_VPMULHUW => Ok(Mnemonic::VPMULHUW),
            ffi::_ND_INS_CLASS::ND_INS_VPMULHW => Ok(Mnemonic::VPMULHW),
            ffi::_ND_INS_CLASS::ND_INS_VPMULLD => Ok(Mnemonic::VPMULLD),
            ffi::_ND_INS_CLASS::ND_INS_VPMULLQ => Ok(Mnemonic::VPMULLQ),
            ffi::_ND_INS_CLASS::ND_INS_VPMULLW => Ok(Mnemonic::VPMULLW),
            ffi::_ND_INS_CLASS::ND_INS_VPMULTISHIFTQB => Ok(Mnemonic::VPMULTISHIFTQB),
            ffi::_ND_INS_CLASS::ND_INS_VPMULUDQ => Ok(Mnemonic::VPMULUDQ),
            ffi::_ND_INS_CLASS::ND_INS_VPOPCNTB => Ok(Mnemonic::VPOPCNTB),
            ffi::_ND_INS_CLASS::ND_INS_VPOPCNTD => Ok(Mnemonic::VPOPCNTD),
            ffi::_ND_INS_CLASS::ND_INS_VPOPCNTQ => Ok(Mnemonic::VPOPCNTQ),
            ffi::_ND_INS_CLASS::ND_INS_VPOPCNTW => Ok(Mnemonic::VPOPCNTW),
            ffi::_ND_INS_CLASS::ND_INS_VPOR => Ok(Mnemonic::VPOR),
            ffi::_ND_INS_CLASS::ND_INS_VPORD => Ok(Mnemonic::VPORD),
            ffi::_ND_INS_CLASS::ND_INS_VPORQ => Ok(Mnemonic::VPORQ),
            ffi::_ND_INS_CLASS::ND_INS_VPPERM => Ok(Mnemonic::VPPERM),
            ffi::_ND_INS_CLASS::ND_INS_VPROLD => Ok(Mnemonic::VPROLD),
            ffi::_ND_INS_CLASS::ND_INS_VPROLQ => Ok(Mnemonic::VPROLQ),
            ffi::_ND_INS_CLASS::ND_INS_VPROLVD => Ok(Mnemonic::VPROLVD),
            ffi::_ND_INS_CLASS::ND_INS_VPROLVQ => Ok(Mnemonic::VPROLVQ),
            ffi::_ND_INS_CLASS::ND_INS_VPRORD => Ok(Mnemonic::VPRORD),
            ffi::_ND_INS_CLASS::ND_INS_VPRORQ => Ok(Mnemonic::VPRORQ),
            ffi::_ND_INS_CLASS::ND_INS_VPRORVD => Ok(Mnemonic::VPRORVD),
            ffi::_ND_INS_CLASS::ND_INS_VPRORVQ => Ok(Mnemonic::VPRORVQ),
            ffi::_ND_INS_CLASS::ND_INS_VPROTB => Ok(Mnemonic::VPROTB),
            ffi::_ND_INS_CLASS::ND_INS_VPROTD => Ok(Mnemonic::VPROTD),
            ffi::_ND_INS_CLASS::ND_INS_VPROTQ => Ok(Mnemonic::VPROTQ),
            ffi::_ND_INS_CLASS::ND_INS_VPROTW => Ok(Mnemonic::VPROTW),
            ffi::_ND_INS_CLASS::ND_INS_VPSADBW => Ok(Mnemonic::VPSADBW),
            ffi::_ND_INS_CLASS::ND_INS_VPSCATTERDD => Ok(Mnemonic::VPSCATTERDD),
            ffi::_ND_INS_CLASS::ND_INS_VPSCATTERDQ => Ok(Mnemonic::VPSCATTERDQ),
            ffi::_ND_INS_CLASS::ND_INS_VPSCATTERQD => Ok(Mnemonic::VPSCATTERQD),
            ffi::_ND_INS_CLASS::ND_INS_VPSCATTERQQ => Ok(Mnemonic::VPSCATTERQQ),
            ffi::_ND_INS_CLASS::ND_INS_VPSHAB => Ok(Mnemonic::VPSHAB),
            ffi::_ND_INS_CLASS::ND_INS_VPSHAD => Ok(Mnemonic::VPSHAD),
            ffi::_ND_INS_CLASS::ND_INS_VPSHAQ => Ok(Mnemonic::VPSHAQ),
            ffi::_ND_INS_CLASS::ND_INS_VPSHAW => Ok(Mnemonic::VPSHAW),
            ffi::_ND_INS_CLASS::ND_INS_VPSHLB => Ok(Mnemonic::VPSHLB),
            ffi::_ND_INS_CLASS::ND_INS_VPSHLD => Ok(Mnemonic::VPSHLD),
            ffi::_ND_INS_CLASS::ND_INS_VPSHLDD => Ok(Mnemonic::VPSHLDD),
            ffi::_ND_INS_CLASS::ND_INS_VPSHLDQ => Ok(Mnemonic::VPSHLDQ),
            ffi::_ND_INS_CLASS::ND_INS_VPSHLDVD => Ok(Mnemonic::VPSHLDVD),
            ffi::_ND_INS_CLASS::ND_INS_VPSHLDVQ => Ok(Mnemonic::VPSHLDVQ),
            ffi::_ND_INS_CLASS::ND_INS_VPSHLDVW => Ok(Mnemonic::VPSHLDVW),
            ffi::_ND_INS_CLASS::ND_INS_VPSHLDW => Ok(Mnemonic::VPSHLDW),
            ffi::_ND_INS_CLASS::ND_INS_VPSHLQ => Ok(Mnemonic::VPSHLQ),
            ffi::_ND_INS_CLASS::ND_INS_VPSHLW => Ok(Mnemonic::VPSHLW),
            ffi::_ND_INS_CLASS::ND_INS_VPSHRDD => Ok(Mnemonic::VPSHRDD),
            ffi::_ND_INS_CLASS::ND_INS_VPSHRDQ => Ok(Mnemonic::VPSHRDQ),
            ffi::_ND_INS_CLASS::ND_INS_VPSHRDVD => Ok(Mnemonic::VPSHRDVD),
            ffi::_ND_INS_CLASS::ND_INS_VPSHRDVQ => Ok(Mnemonic::VPSHRDVQ),
            ffi::_ND_INS_CLASS::ND_INS_VPSHRDVW => Ok(Mnemonic::VPSHRDVW),
            ffi::_ND_INS_CLASS::ND_INS_VPSHRDW => Ok(Mnemonic::VPSHRDW),
            ffi::_ND_INS_CLASS::ND_INS_VPSHUFB => Ok(Mnemonic::VPSHUFB),
            ffi::_ND_INS_CLASS::ND_INS_VPSHUFBITQMB => Ok(Mnemonic::VPSHUFBITQMB),
            ffi::_ND_INS_CLASS::ND_INS_VPSHUFD => Ok(Mnemonic::VPSHUFD),
            ffi::_ND_INS_CLASS::ND_INS_VPSHUFHW => Ok(Mnemonic::VPSHUFHW),
            ffi::_ND_INS_CLASS::ND_INS_VPSHUFLW => Ok(Mnemonic::VPSHUFLW),
            ffi::_ND_INS_CLASS::ND_INS_VPSIGNB => Ok(Mnemonic::VPSIGNB),
            ffi::_ND_INS_CLASS::ND_INS_VPSIGND => Ok(Mnemonic::VPSIGND),
            ffi::_ND_INS_CLASS::ND_INS_VPSIGNW => Ok(Mnemonic::VPSIGNW),
            ffi::_ND_INS_CLASS::ND_INS_VPSLLD => Ok(Mnemonic::VPSLLD),
            ffi::_ND_INS_CLASS::ND_INS_VPSLLDQ => Ok(Mnemonic::VPSLLDQ),
            ffi::_ND_INS_CLASS::ND_INS_VPSLLQ => Ok(Mnemonic::VPSLLQ),
            ffi::_ND_INS_CLASS::ND_INS_VPSLLVD => Ok(Mnemonic::VPSLLVD),
            ffi::_ND_INS_CLASS::ND_INS_VPSLLVQ => Ok(Mnemonic::VPSLLVQ),
            ffi::_ND_INS_CLASS::ND_INS_VPSLLVW => Ok(Mnemonic::VPSLLVW),
            ffi::_ND_INS_CLASS::ND_INS_VPSLLW => Ok(Mnemonic::VPSLLW),
            ffi::_ND_INS_CLASS::ND_INS_VPSRAD => Ok(Mnemonic::VPSRAD),
            ffi::_ND_INS_CLASS::ND_INS_VPSRAQ => Ok(Mnemonic::VPSRAQ),
            ffi::_ND_INS_CLASS::ND_INS_VPSRAVD => Ok(Mnemonic::VPSRAVD),
            ffi::_ND_INS_CLASS::ND_INS_VPSRAVQ => Ok(Mnemonic::VPSRAVQ),
            ffi::_ND_INS_CLASS::ND_INS_VPSRAVW => Ok(Mnemonic::VPSRAVW),
            ffi::_ND_INS_CLASS::ND_INS_VPSRAW => Ok(Mnemonic::VPSRAW),
            ffi::_ND_INS_CLASS::ND_INS_VPSRLD => Ok(Mnemonic::VPSRLD),
            ffi::_ND_INS_CLASS::ND_INS_VPSRLDQ => Ok(Mnemonic::VPSRLDQ),
            ffi::_ND_INS_CLASS::ND_INS_VPSRLQ => Ok(Mnemonic::VPSRLQ),
            ffi::_ND_INS_CLASS::ND_INS_VPSRLVD => Ok(Mnemonic::VPSRLVD),
            ffi::_ND_INS_CLASS::ND_INS_VPSRLVQ => Ok(Mnemonic::VPSRLVQ),
            ffi::_ND_INS_CLASS::ND_INS_VPSRLVW => Ok(Mnemonic::VPSRLVW),
            ffi::_ND_INS_CLASS::ND_INS_VPSRLW => Ok(Mnemonic::VPSRLW),
            ffi::_ND_INS_CLASS::ND_INS_VPSUBB => Ok(Mnemonic::VPSUBB),
            ffi::_ND_INS_CLASS::ND_INS_VPSUBD => Ok(Mnemonic::VPSUBD),
            ffi::_ND_INS_CLASS::ND_INS_VPSUBQ => Ok(Mnemonic::VPSUBQ),
            ffi::_ND_INS_CLASS::ND_INS_VPSUBSB => Ok(Mnemonic::VPSUBSB),
            ffi::_ND_INS_CLASS::ND_INS_VPSUBSW => Ok(Mnemonic::VPSUBSW),
            ffi::_ND_INS_CLASS::ND_INS_VPSUBUSB => Ok(Mnemonic::VPSUBUSB),
            ffi::_ND_INS_CLASS::ND_INS_VPSUBUSW => Ok(Mnemonic::VPSUBUSW),
            ffi::_ND_INS_CLASS::ND_INS_VPSUBW => Ok(Mnemonic::VPSUBW),
            ffi::_ND_INS_CLASS::ND_INS_VPTERNLOGD => Ok(Mnemonic::VPTERNLOGD),
            ffi::_ND_INS_CLASS::ND_INS_VPTERNLOGQ => Ok(Mnemonic::VPTERNLOGQ),
            ffi::_ND_INS_CLASS::ND_INS_VPTEST => Ok(Mnemonic::VPTEST),
            ffi::_ND_INS_CLASS::ND_INS_VPTESTMB => Ok(Mnemonic::VPTESTMB),
            ffi::_ND_INS_CLASS::ND_INS_VPTESTMD => Ok(Mnemonic::VPTESTMD),
            ffi::_ND_INS_CLASS::ND_INS_VPTESTMQ => Ok(Mnemonic::VPTESTMQ),
            ffi::_ND_INS_CLASS::ND_INS_VPTESTMW => Ok(Mnemonic::VPTESTMW),
            ffi::_ND_INS_CLASS::ND_INS_VPTESTNMB => Ok(Mnemonic::VPTESTNMB),
            ffi::_ND_INS_CLASS::ND_INS_VPTESTNMD => Ok(Mnemonic::VPTESTNMD),
            ffi::_ND_INS_CLASS::ND_INS_VPTESTNMQ => Ok(Mnemonic::VPTESTNMQ),
            ffi::_ND_INS_CLASS::ND_INS_VPTESTNMW => Ok(Mnemonic::VPTESTNMW),
            ffi::_ND_INS_CLASS::ND_INS_VPUNPCKHBW => Ok(Mnemonic::VPUNPCKHBW),
            ffi::_ND_INS_CLASS::ND_INS_VPUNPCKHDQ => Ok(Mnemonic::VPUNPCKHDQ),
            ffi::_ND_INS_CLASS::ND_INS_VPUNPCKHQDQ => Ok(Mnemonic::VPUNPCKHQDQ),
            ffi::_ND_INS_CLASS::ND_INS_VPUNPCKHWD => Ok(Mnemonic::VPUNPCKHWD),
            ffi::_ND_INS_CLASS::ND_INS_VPUNPCKLBW => Ok(Mnemonic::VPUNPCKLBW),
            ffi::_ND_INS_CLASS::ND_INS_VPUNPCKLDQ => Ok(Mnemonic::VPUNPCKLDQ),
            ffi::_ND_INS_CLASS::ND_INS_VPUNPCKLQDQ => Ok(Mnemonic::VPUNPCKLQDQ),
            ffi::_ND_INS_CLASS::ND_INS_VPUNPCKLWD => Ok(Mnemonic::VPUNPCKLWD),
            ffi::_ND_INS_CLASS::ND_INS_VPXOR => Ok(Mnemonic::VPXOR),
            ffi::_ND_INS_CLASS::ND_INS_VPXORD => Ok(Mnemonic::VPXORD),
            ffi::_ND_INS_CLASS::ND_INS_VPXORQ => Ok(Mnemonic::VPXORQ),
            ffi::_ND_INS_CLASS::ND_INS_VRANGEPD => Ok(Mnemonic::VRANGEPD),
            ffi::_ND_INS_CLASS::ND_INS_VRANGEPS => Ok(Mnemonic::VRANGEPS),
            ffi::_ND_INS_CLASS::ND_INS_VRANGESD => Ok(Mnemonic::VRANGESD),
            ffi::_ND_INS_CLASS::ND_INS_VRANGESS => Ok(Mnemonic::VRANGESS),
            ffi::_ND_INS_CLASS::ND_INS_VRCP14PD => Ok(Mnemonic::VRCP14PD),
            ffi::_ND_INS_CLASS::ND_INS_VRCP14PS => Ok(Mnemonic::VRCP14PS),
            ffi::_ND_INS_CLASS::ND_INS_VRCP14SD => Ok(Mnemonic::VRCP14SD),
            ffi::_ND_INS_CLASS::ND_INS_VRCP14SS => Ok(Mnemonic::VRCP14SS),
            ffi::_ND_INS_CLASS::ND_INS_VRCP28PD => Ok(Mnemonic::VRCP28PD),
            ffi::_ND_INS_CLASS::ND_INS_VRCP28PS => Ok(Mnemonic::VRCP28PS),
            ffi::_ND_INS_CLASS::ND_INS_VRCP28SD => Ok(Mnemonic::VRCP28SD),
            ffi::_ND_INS_CLASS::ND_INS_VRCP28SS => Ok(Mnemonic::VRCP28SS),
            ffi::_ND_INS_CLASS::ND_INS_VRCPPBF16 => Ok(Mnemonic::VRCPPBF16),
            ffi::_ND_INS_CLASS::ND_INS_VRCPPH => Ok(Mnemonic::VRCPPH),
            ffi::_ND_INS_CLASS::ND_INS_VRCPPS => Ok(Mnemonic::VRCPPS),
            ffi::_ND_INS_CLASS::ND_INS_VRCPSH => Ok(Mnemonic::VRCPSH),
            ffi::_ND_INS_CLASS::ND_INS_VRCPSS => Ok(Mnemonic::VRCPSS),
            ffi::_ND_INS_CLASS::ND_INS_VREDUCENEPBF16 => Ok(Mnemonic::VREDUCENEPBF16),
            ffi::_ND_INS_CLASS::ND_INS_VREDUCEPD => Ok(Mnemonic::VREDUCEPD),
            ffi::_ND_INS_CLASS::ND_INS_VREDUCEPH => Ok(Mnemonic::VREDUCEPH),
            ffi::_ND_INS_CLASS::ND_INS_VREDUCEPS => Ok(Mnemonic::VREDUCEPS),
            ffi::_ND_INS_CLASS::ND_INS_VREDUCESD => Ok(Mnemonic::VREDUCESD),
            ffi::_ND_INS_CLASS::ND_INS_VREDUCESH => Ok(Mnemonic::VREDUCESH),
            ffi::_ND_INS_CLASS::ND_INS_VREDUCESS => Ok(Mnemonic::VREDUCESS),
            ffi::_ND_INS_CLASS::ND_INS_VRNDSCALENEPBF16 => Ok(Mnemonic::VRNDSCALENEPBF16),
            ffi::_ND_INS_CLASS::ND_INS_VRNDSCALEPD => Ok(Mnemonic::VRNDSCALEPD),
            ffi::_ND_INS_CLASS::ND_INS_VRNDSCALEPH => Ok(Mnemonic::VRNDSCALEPH),
            ffi::_ND_INS_CLASS::ND_INS_VRNDSCALEPS => Ok(Mnemonic::VRNDSCALEPS),
            ffi::_ND_INS_CLASS::ND_INS_VRNDSCALESD => Ok(Mnemonic::VRNDSCALESD),
            ffi::_ND_INS_CLASS::ND_INS_VRNDSCALESH => Ok(Mnemonic::VRNDSCALESH),
            ffi::_ND_INS_CLASS::ND_INS_VRNDSCALESS => Ok(Mnemonic::VRNDSCALESS),
            ffi::_ND_INS_CLASS::ND_INS_VROUNDPD => Ok(Mnemonic::VROUNDPD),
            ffi::_ND_INS_CLASS::ND_INS_VROUNDPS => Ok(Mnemonic::VROUNDPS),
            ffi::_ND_INS_CLASS::ND_INS_VROUNDSD => Ok(Mnemonic::VROUNDSD),
            ffi::_ND_INS_CLASS::ND_INS_VROUNDSS => Ok(Mnemonic::VROUNDSS),
            ffi::_ND_INS_CLASS::ND_INS_VRSQRT14PD => Ok(Mnemonic::VRSQRT14PD),
            ffi::_ND_INS_CLASS::ND_INS_VRSQRT14PS => Ok(Mnemonic::VRSQRT14PS),
            ffi::_ND_INS_CLASS::ND_INS_VRSQRT14SD => Ok(Mnemonic::VRSQRT14SD),
            ffi::_ND_INS_CLASS::ND_INS_VRSQRT14SS => Ok(Mnemonic::VRSQRT14SS),
            ffi::_ND_INS_CLASS::ND_INS_VRSQRT28PD => Ok(Mnemonic::VRSQRT28PD),
            ffi::_ND_INS_CLASS::ND_INS_VRSQRT28PS => Ok(Mnemonic::VRSQRT28PS),
            ffi::_ND_INS_CLASS::ND_INS_VRSQRT28SD => Ok(Mnemonic::VRSQRT28SD),
            ffi::_ND_INS_CLASS::ND_INS_VRSQRT28SS => Ok(Mnemonic::VRSQRT28SS),
            ffi::_ND_INS_CLASS::ND_INS_VRSQRTPBF16 => Ok(Mnemonic::VRSQRTPBF16),
            ffi::_ND_INS_CLASS::ND_INS_VRSQRTPH => Ok(Mnemonic::VRSQRTPH),
            ffi::_ND_INS_CLASS::ND_INS_VRSQRTPS => Ok(Mnemonic::VRSQRTPS),
            ffi::_ND_INS_CLASS::ND_INS_VRSQRTSH => Ok(Mnemonic::VRSQRTSH),
            ffi::_ND_INS_CLASS::ND_INS_VRSQRTSS => Ok(Mnemonic::VRSQRTSS),
            ffi::_ND_INS_CLASS::ND_INS_VSCALEFPBF16 => Ok(Mnemonic::VSCALEFPBF16),
            ffi::_ND_INS_CLASS::ND_INS_VSCALEFPD => Ok(Mnemonic::VSCALEFPD),
            ffi::_ND_INS_CLASS::ND_INS_VSCALEFPH => Ok(Mnemonic::VSCALEFPH),
            ffi::_ND_INS_CLASS::ND_INS_VSCALEFPS => Ok(Mnemonic::VSCALEFPS),
            ffi::_ND_INS_CLASS::ND_INS_VSCALEFSD => Ok(Mnemonic::VSCALEFSD),
            ffi::_ND_INS_CLASS::ND_INS_VSCALEFSH => Ok(Mnemonic::VSCALEFSH),
            ffi::_ND_INS_CLASS::ND_INS_VSCALEFSS => Ok(Mnemonic::VSCALEFSS),
            ffi::_ND_INS_CLASS::ND_INS_VSCATTERDPD => Ok(Mnemonic::VSCATTERDPD),
            ffi::_ND_INS_CLASS::ND_INS_VSCATTERDPS => Ok(Mnemonic::VSCATTERDPS),
            ffi::_ND_INS_CLASS::ND_INS_VSCATTERPF0DPD => Ok(Mnemonic::VSCATTERPF0DPD),
            ffi::_ND_INS_CLASS::ND_INS_VSCATTERPF0DPS => Ok(Mnemonic::VSCATTERPF0DPS),
            ffi::_ND_INS_CLASS::ND_INS_VSCATTERPF0QPD => Ok(Mnemonic::VSCATTERPF0QPD),
            ffi::_ND_INS_CLASS::ND_INS_VSCATTERPF0QPS => Ok(Mnemonic::VSCATTERPF0QPS),
            ffi::_ND_INS_CLASS::ND_INS_VSCATTERPF1DPD => Ok(Mnemonic::VSCATTERPF1DPD),
            ffi::_ND_INS_CLASS::ND_INS_VSCATTERPF1DPS => Ok(Mnemonic::VSCATTERPF1DPS),
            ffi::_ND_INS_CLASS::ND_INS_VSCATTERPF1QPD => Ok(Mnemonic::VSCATTERPF1QPD),
            ffi::_ND_INS_CLASS::ND_INS_VSCATTERPF1QPS => Ok(Mnemonic::VSCATTERPF1QPS),
            ffi::_ND_INS_CLASS::ND_INS_VSCATTERQPD => Ok(Mnemonic::VSCATTERQPD),
            ffi::_ND_INS_CLASS::ND_INS_VSCATTERQPS => Ok(Mnemonic::VSCATTERQPS),
            ffi::_ND_INS_CLASS::ND_INS_VSHA512MSG1 => Ok(Mnemonic::VSHA512MSG1),
            ffi::_ND_INS_CLASS::ND_INS_VSHA512MSG2 => Ok(Mnemonic::VSHA512MSG2),
            ffi::_ND_INS_CLASS::ND_INS_VSHA512RNDS2 => Ok(Mnemonic::VSHA512RNDS2),
            ffi::_ND_INS_CLASS::ND_INS_VSHUFF32X4 => Ok(Mnemonic::VSHUFF32X4),
            ffi::_ND_INS_CLASS::ND_INS_VSHUFF64X2 => Ok(Mnemonic::VSHUFF64X2),
            ffi::_ND_INS_CLASS::ND_INS_VSHUFI32X4 => Ok(Mnemonic::VSHUFI32X4),
            ffi::_ND_INS_CLASS::ND_INS_VSHUFI64X2 => Ok(Mnemonic::VSHUFI64X2),
            ffi::_ND_INS_CLASS::ND_INS_VSHUFPD => Ok(Mnemonic::VSHUFPD),
            ffi::_ND_INS_CLASS::ND_INS_VSHUFPS => Ok(Mnemonic::VSHUFPS),
            ffi::_ND_INS_CLASS::ND_INS_VSM3MSG1 => Ok(Mnemonic::VSM3MSG1),
            ffi::_ND_INS_CLASS::ND_INS_VSM3MSG2 => Ok(Mnemonic::VSM3MSG2),
            ffi::_ND_INS_CLASS::ND_INS_VSM3RNDS2 => Ok(Mnemonic::VSM3RNDS2),
            ffi::_ND_INS_CLASS::ND_INS_VSM4KEY4 => Ok(Mnemonic::VSM4KEY4),
            ffi::_ND_INS_CLASS::ND_INS_VSM4RNDS4 => Ok(Mnemonic::VSM4RNDS4),
            ffi::_ND_INS_CLASS::ND_INS_VSQRTNEPBF16 => Ok(Mnemonic::VSQRTNEPBF16),
            ffi::_ND_INS_CLASS::ND_INS_VSQRTPD => Ok(Mnemonic::VSQRTPD),
            ffi::_ND_INS_CLASS::ND_INS_VSQRTPH => Ok(Mnemonic::VSQRTPH),
            ffi::_ND_INS_CLASS::ND_INS_VSQRTPS => Ok(Mnemonic::VSQRTPS),
            ffi::_ND_INS_CLASS::ND_INS_VSQRTSD => Ok(Mnemonic::VSQRTSD),
            ffi::_ND_INS_CLASS::ND_INS_VSQRTSH => Ok(Mnemonic::VSQRTSH),
            ffi::_ND_INS_CLASS::ND_INS_VSQRTSS => Ok(Mnemonic::VSQRTSS),
            ffi::_ND_INS_CLASS::ND_INS_VSTMXCSR => Ok(Mnemonic::VSTMXCSR),
            ffi::_ND_INS_CLASS::ND_INS_VSUBNEPBF16 => Ok(Mnemonic::VSUBNEPBF16),
            ffi::_ND_INS_CLASS::ND_INS_VSUBPD => Ok(Mnemonic::VSUBPD),
            ffi::_ND_INS_CLASS::ND_INS_VSUBPH => Ok(Mnemonic::VSUBPH),
            ffi::_ND_INS_CLASS::ND_INS_VSUBPS => Ok(Mnemonic::VSUBPS),
            ffi::_ND_INS_CLASS::ND_INS_VSUBSD => Ok(Mnemonic::VSUBSD),
            ffi::_ND_INS_CLASS::ND_INS_VSUBSH => Ok(Mnemonic::VSUBSH),
            ffi::_ND_INS_CLASS::ND_INS_VSUBSS => Ok(Mnemonic::VSUBSS),
            ffi::_ND_INS_CLASS::ND_INS_VTESTPD => Ok(Mnemonic::VTESTPD),
            ffi::_ND_INS_CLASS::ND_INS_VTESTPS => Ok(Mnemonic::VTESTPS),
            ffi::_ND_INS_CLASS::ND_INS_VUCOMISD => Ok(Mnemonic::VUCOMISD),
            ffi::_ND_INS_CLASS::ND_INS_VUCOMISH => Ok(Mnemonic::VUCOMISH),
            ffi::_ND_INS_CLASS::ND_INS_VUCOMISS => Ok(Mnemonic::VUCOMISS),
            ffi::_ND_INS_CLASS::ND_INS_VUCOMXSD => Ok(Mnemonic::VUCOMXSD),
            ffi::_ND_INS_CLASS::ND_INS_VUCOMXSH => Ok(Mnemonic::VUCOMXSH),
            ffi::_ND_INS_CLASS::ND_INS_VUCOMXSS => Ok(Mnemonic::VUCOMXSS),
            ffi::_ND_INS_CLASS::ND_INS_VUNPCKHPD => Ok(Mnemonic::VUNPCKHPD),
            ffi::_ND_INS_CLASS::ND_INS_VUNPCKHPS => Ok(Mnemonic::VUNPCKHPS),
            ffi::_ND_INS_CLASS::ND_INS_VUNPCKLPD => Ok(Mnemonic::VUNPCKLPD),
            ffi::_ND_INS_CLASS::ND_INS_VUNPCKLPS => Ok(Mnemonic::VUNPCKLPS),
            ffi::_ND_INS_CLASS::ND_INS_VXORPD => Ok(Mnemonic::VXORPD),
            ffi::_ND_INS_CLASS::ND_INS_VXORPS => Ok(Mnemonic::VXORPS),
            ffi::_ND_INS_CLASS::ND_INS_VZEROALL => Ok(Mnemonic::VZEROALL),
            ffi::_ND_INS_CLASS::ND_INS_VZEROUPPER => Ok(Mnemonic::VZEROUPPER),
            ffi::_ND_INS_CLASS::ND_INS_WAIT => Ok(Mnemonic::WAIT),
            ffi::_ND_INS_CLASS::ND_INS_WBINVD => Ok(Mnemonic::WBINVD),
            ffi::_ND_INS_CLASS::ND_INS_WBNOINVD => Ok(Mnemonic::WBNOINVD),
            ffi::_ND_INS_CLASS::ND_INS_WRFSBASE => Ok(Mnemonic::WRFSBASE),
            ffi::_ND_INS_CLASS::ND_INS_WRGSBASE => Ok(Mnemonic::WRGSBASE),
            ffi::_ND_INS_CLASS::ND_INS_WRMSR => Ok(Mnemonic::WRMSR),
            ffi::_ND_INS_CLASS::ND_INS_WRMSRLIST => Ok(Mnemonic::WRMSRLIST),
            ffi::_ND_INS_CLASS::ND_INS_WRMSRNS => Ok(Mnemonic::WRMSRNS),
            ffi::_ND_INS_CLASS::ND_INS_WRPKRU => Ok(Mnemonic::WRPKRU),
            ffi::_ND_INS_CLASS::ND_INS_WRSS => Ok(Mnemonic::WRSS),
            ffi::_ND_INS_CLASS::ND_INS_WRUSS => Ok(Mnemonic::WRUSS),
            ffi::_ND_INS_CLASS::ND_INS_XABORT => Ok(Mnemonic::XABORT),
            ffi::_ND_INS_CLASS::ND_INS_XADD => Ok(Mnemonic::XADD),
            ffi::_ND_INS_CLASS::ND_INS_XBEGIN => Ok(Mnemonic::XBEGIN),
            ffi::_ND_INS_CLASS::ND_INS_XCHG => Ok(Mnemonic::XCHG),
            ffi::_ND_INS_CLASS::ND_INS_XEND => Ok(Mnemonic::XEND),
            ffi::_ND_INS_CLASS::ND_INS_XGETBV => Ok(Mnemonic::XGETBV),
            ffi::_ND_INS_CLASS::ND_INS_XLATB => Ok(Mnemonic::XLATB),
            ffi::_ND_INS_CLASS::ND_INS_XOR => Ok(Mnemonic::XOR),
            ffi::_ND_INS_CLASS::ND_INS_XORPD => Ok(Mnemonic::XORPD),
            ffi::_ND_INS_CLASS::ND_INS_XORPS => Ok(Mnemonic::XORPS),
            ffi::_ND_INS_CLASS::ND_INS_XRESLDTRK => Ok(Mnemonic::XRESLDTRK),
            ffi::_ND_INS_CLASS::ND_INS_XRSTOR => Ok(Mnemonic::XRSTOR),
            ffi::_ND_INS_CLASS::ND_INS_XRSTORS => Ok(Mnemonic::XRSTORS),
            ffi::_ND_INS_CLASS::ND_INS_XSAVE => Ok(Mnemonic::XSAVE),
            ffi::_ND_INS_CLASS::ND_INS_XSAVEC => Ok(Mnemonic::XSAVEC),
            ffi::_ND_INS_CLASS::ND_INS_XSAVEOPT => Ok(Mnemonic::XSAVEOPT),
            ffi::_ND_INS_CLASS::ND_INS_XSAVES => Ok(Mnemonic::XSAVES),
            ffi::_ND_INS_CLASS::ND_INS_XSETBV => Ok(Mnemonic::XSETBV),
            ffi::_ND_INS_CLASS::ND_INS_XSUSLDTRK => Ok(Mnemonic::XSUSLDTRK),
            ffi::_ND_INS_CLASS::ND_INS_XTEST => Ok(Mnemonic::XTEST),
        }
    }
}
